/********
	    Packet filter by wallrich@loria.fr

	    analyse.c
	    
	    analyse of a packet.
*********/

#include "nicedump.h"


/******************************
 ********* VARIABLE ***********
 ******************************/

variable_value *get_variable_value(variable *, int);
static int construct_global_var_name(char *, const char *, params *);

/******* externs from outside *******/
extern int strncmp(const char *, const char *, size_t);
extern void bzero(void *, int );

/******* externs from packet.c *******/
#ifdef STDC_HEADERS
extern int add_display(params *, const char, int, ...);
#else
extern int add_display(params *, const char, int);
#endif

/******* externs from calcul.c *******/
extern int compare (variable_value *, variable_value *, char);

/*************************************/


/********
	    get_variable_by_name
	    return a ptr to the variable, in func of the name.
	    or return NULL;

	    if the name of the variable start with a !, we look for it in the
	    global chain.


	    get_variable :
	    STRICT_LABEL_PATH : search variable in this protocol.
	    FOLLOW_LABEL_PATH : search variable in this protocol or previous protocols.

	    this can be overwrite by p->get_variable

	    RAMBO_SEARCH      : search variable without protocol considerations.

*********/
variable *get_variable_by_name(const char *name, params *parm, char get_variable) {
  variable *p;
  protocol prot;

  char secname[MAX_LINE];  /* used to add ':' as a prefix */
  int len, l;
  /* add_display(parm,D_DEBUG,-1,"looking for <%s>\n",name);  */
  /* printf("looking for <%s>\n",name); */
  /* looking in the global chain */
  if (name[0]=='!') {
    len=construct_global_var_name(secname,name,parm);
    p=*(parm->global_vstart);
    while (p!=NULL) {
      /*      add_display(parm,D_DEBUG,-1,"testing <%s>(%d) et <%s>(%d)\n",secname,len,p->name,strlen(p->name)); */
      if ((len==strlen(p->name)) && (strncmp(p->name,secname,len)==0)) break;
      p=p->next;
    }
    
    /* if (p==NULL)  add_display(parm,D_DEBUG,-1,"variable <%s> not definied\n",secname); 
       else add_display(parm,D_DEBUG,-1,"variable <%s> definied (=%d)\n",secname,*((UNSIG_DEC *)p->values[0].val));  */
    

    return(p);
  }

  /* search the variable in the list */
  /* the variable name is a FQ name */
  if (name[0]==':') {
    p=parm->vend;
    len=strlen(name);
    while (p!=NULL) {
      if ((len==strlen(p->name)) && (strncmp(p->name,name,len)==0)) break;
      p=p->prev;
    }
    /* if (p==NULL)  add_display(parm,D_DEBUG,-1,"variable <%s> not definied\n",name); 
       else add_display(parm,D_DEBUG,-1,"variable <%s> definied\n",name); 
    */
    return(p);
  }
  
  /* we don't follow the label path */
  if (parm->get_variable==RAMBO_SEARCH) {
    p=parm->vend;
    len=strlen(name);
    while ((p!=NULL) &&
	   (len<=strlen(p->name)) &&
	   (strncmp(p->name+strlen(p->name)-len,name,len)!=0)) {
      p=p->prev;
    }
    /* if (p!=NULL) printf("searching <%s>, found <%s>\n",name, p->name);
	  else printf("nothing found for  <%s>\n",name); */
    /*    if (p==NULL)  add_display(parm,D_DEBUG,-1,"variable <%s> not definied\n",name); 
	  else add_display(parm,D_DEBUG,-1,"variable <%s> definied =<%s>\n",name,p->printvalue); 
    */
    
    return(p);
  }


  /* Not a FQ name, we add the label_path */
  prot=parm->prot[parm->prot_index];
  len=prot.label_path_length;
  p=NULL;
  while((len>=0) && (p==NULL)) {
    memcpy(secname,prot.label_path,len);
    secname[len]=':';
    memcpy(secname+len+1,name,strlen(name));
    l=len+strlen(name)+1;
    secname[l]='\0';
    p=parm->vend;
    while (p!=NULL) {
      /* add_display(parm,D_DEBUG,-1,"testing variable <%s> and <%s> \n",secname,p->name);  */
      if ((l==strlen(p->name) && (strncmp(p->name,secname,l)==0))) break;
      p=p->prev;
    }

    /* not found with this path, decrement the path */
    if (p==NULL) {
	 if (get_variable==STRICT_LABEL_PATH) return(p);
	 len--;
	 while((len>=0) && (prot.label_path[len]!=':')) len--;
    }
  } 

  /*  
  if (p==NULL)  add_display(parm,D_DEBUG,-1,"variable <%s> not definied\n",secname); 
  else add_display(parm,D_DEBUG,-1,"variable <%s> definied\n",secname); 
  */
  return(p);
}

/*****
    construct the global_variable_name.
    example :

    !src!port_src!dst!port_dst!state 

    becomes

    !152.81.1.12!1024!152.81.1.14!23!state
****/
static int construct_global_var_name(char *res, const char *name, params *parm) {
  int start, end;
  char *nn;       /* the copy of name, because we modify it */
  variable_value *varval;
  int lres=1, i, len;
  variable *p;

  start=1;
  len=strlen(name);
  nn=malloc(len+1);  /* nn is a copy of name, because we modify it */
  memcpy(nn,name,len);

  res[0]='!';
  /* replace all variables by there printvalues */
  for(end=1 ; end<len ; end++) {
    if (nn[end]=='!') {
	 nn[end]='\0';
	 p=get_variable_by_name(nn+start,parm,FOLLOW_LABEL_PATH);
	 varval=get_variable_value(p, STRING_FORCE);
	 i=strlen((char *)(varval->val));
	 memcpy(res+lres,(char *)(varval->val),i);
	 lres+=i;
	 res[lres++]='!';
	 end++;
	 start=end;
	 free(varval->val);
	 free(varval);
    }
  }

  /* adding the name of the last variable */
  i=end-start;
  memcpy(res+lres,nn+start,i);
  res[lres+i]='\0';
  free(nn);
  return(lres+i);
}

/********
	    add_new_variable
	    initialise a new variable in the chaine of the packet, or a global
	    variable if its name start with a |
	    name : the name of the variable
         path : the label PATH, (prefix to the variable name)
	    dspfmt : the display format
	    parm: the parm struct
********/
variable *add_new_variable (char *name, const char *dspfmt, params *parm) {
  int len;
  char secname[MAX_LINE];
  char *nn;                 /* the copy of name, in case of modifications */
  int freenn=0;
  variable *vp;
  protocol prot;

  if (!dspfmt) return(NULL);

  nn=name;
  /* bypass the .[0-9] at the end */
  len=strlen(name)-1;  
  if ((len>2) && (nn[len-1]=='.') && (nn[len]>='0')
	 && (nn[len]<='9')) {
    nn=malloc(len);
    memcpy(nn,name,len-2);
    nn[len-1]='\0';    
    freenn=1;
  }


  /* adding a global variable name */
  if (nn[0]=='!') {

    /* if this var exists, return is pointer */
    vp=get_variable_by_name(nn, parm,STRICT_LABEL_PATH);
    if (vp!=NULL) {
	 memcpy(vp->dspfmt,dspfmt,strlen(dspfmt));
	 vp->dspfmt[strlen(dspfmt)]='\0'; 
	 return(vp);
    }
    len=construct_global_var_name(secname,nn,parm);
    /* add_display(parm,D_DEBUG,-1,"adding variable <%s>(%d)",secname,len);   */
   if (*(parm->global_vstart)==NULL) {
	 *(parm->global_vstart)=malloc(sizeof(variable));
	 vp=(*(parm->global_vstart));
	 vp->next=NULL;
    } else {
	 vp=malloc(sizeof(variable));
	 vp->next=(*(parm->global_vstart))->next;
	 (*(parm->global_vstart))->next=vp;
    }
    
    memcpy(vp->name,secname,len);
    vp->name[len]='\0';
    memcpy(vp->dspfmt,dspfmt,strlen(dspfmt));
    vp->dspfmt[strlen(dspfmt)]='\0';
    
    /* initialising */
    vp->num=0;
    for(len=0;len<10;len++) {
      vp->values[len].format=F_UNDEF;
      vp->values[len].length=-1;
      vp->values[len].val=NULL;
    }
    return(vp);

  }

  /* adding a "local" variable */
  prot=parm->prot[parm->prot_index];
  if ((prot.label_path_length+strlen(nn)+2)>MAX_LINE) {
    add_display(parm,DE_MEMORY,-1,"label path %s too long with variable %s.",prot.label_path,nn); 
    return(NULL);
  }

  /* calculate the Full qualified name */
  memcpy(secname,prot.label_path,prot.label_path_length);
  secname[prot.label_path_length]=':';
  memcpy(secname+prot.label_path_length+1,nn,strlen(nn));
  secname[prot.label_path_length+strlen(nn)+1]='\0';
  len=prot.label_path_length+strlen(nn)+2;


  /* modifie variable, and not add the same */
  vp=get_variable_by_name(secname, parm,STRICT_LABEL_PATH);
  if (vp!=NULL) {
    memcpy(vp->dspfmt,dspfmt,strlen(dspfmt));
    vp->dspfmt[strlen(dspfmt)]='\0';
    return(vp);
  }
  /* add_display(parm,D_DEBUG,-1,"Realy adding variable <%s>\n",secname); */

  /* a new entry in the struct */
  if (parm->vstart==NULL) {
    parm->vstart=malloc(sizeof(variable));
    parm->vend=parm->vstart;
    parm->vend->next=NULL;
    parm->vend->prev=NULL;
  } else {
    vp=parm->vend;
    parm->vend=malloc(sizeof(variable));
    vp->next=parm->vend;
    parm->vend->prev=vp;
    parm->vend->next=NULL;
  }


  /* stocking name */
  memcpy(parm->vend->name,secname,len);
  
  memcpy(parm->vend->dspfmt,dspfmt,strlen(dspfmt));
  parm->vend->dspfmt[strlen(dspfmt)]='\0';

  /* initialising */
  parm->vend->num=0;
  for(len=0;len<10;len++) {
    parm->vend->values[len].format=F_UNDEF;
    parm->vend->values[len].val=NULL;
    parm->vend->values[len].length=-1;
  }
  
  if (freenn) free(nn);
  return(parm->vend);
}

/**********
		 modify_display
		 modify the display string (printvalue)
		 in function of dspfmt.
***********/
int modify_display (variable *p) {
  int i,j;       /* i for dspfmt, j for printvalue */
  int doit;      /* to do or not to do a sub in printvalue ? */
  int len;       /* length of dspfmt */
  int to_add;    /* increments of j */
  int rang;

  /* recalculate the printvalue */
  len=strlen(p->dspfmt);
  bzero(p->printvalue,MAX_LINE);
  /* printf("dspfmt(%s) = %s\n",p->name,p->dspfmt); */

  j=0;
  for (i=0;i<len;i++) {
    if (j>=MAX_LINE) break;
    doit=0;    
    if ((p->dspfmt[i]=='%') && (i>0) && (p->dspfmt[i-1]!='\\') && (i<(len-1))) doit=1;
    if ((p->dspfmt[i]=='%') && (i==0) && (i<(len-1))) doit=1;

    if (doit) {
	 i++;
	 rang=0;
	 to_add=0;
	 if ((i<(len-1)) && (p->dspfmt[i+1]>='0') && (p->dspfmt[i+1]<='9')) {
	   rang=p->dspfmt[i+1]-'0';
	   to_add=1;
	 }
	 switch (p->dspfmt[i]) {
	 case 'd' :     /* signed dec */	   
	   if ((p->values[rang].format==F_SIGNED) || (p->values[rang].format==F_UNSIGNED))
		snprintf(p->printvalue+j,MAX_LINE-j,"%ld",*((SIG_DEC *)(p->values[rang].val)));
	   break;
	 case 'u' :     /* unsigned dec */
	   if ((p->values[rang].format==F_SIGNED) || (p->values[rang].format==F_UNSIGNED))
		snprintf(p->printvalue+j,MAX_LINE-j,"%lu",*((UNSIG_DEC *)(p->values[rang].val)));
	   break;
	 case 's' :     /* string */
	   if (p->values[rang].format==F_STRING)
		snprintf(p->printvalue+j,MAX_LINE-j,"%s",(char *)(p->values[rang].val));
	   break;
	 case 'b' :     /* binary */
	   if ((p->values[rang].format==F_SIGNED) || (p->values[rang].format==F_UNSIGNED))
		snprintf(p->printvalue+j,MAX_LINE-j,"%lx",*((UNSIG_DEC *)(p->values[rang].val)));
	   if (p->values[rang].format==F_BINAR)
	       snprintf(p->printvalue+j,MAX_LINE-j,"(binary %d octets)",p->values[rang].length);
	   break;
	 case 'H' :     /* hex 16 */
	   if ((p->values[rang].format==F_SIGNED) || (p->values[rang].format==F_UNSIGNED))
	   snprintf(p->printvalue+j,MAX_LINE-j,"%04lx",*((UNSIG_DEC *)(p->values[rang].val)));
	   break;
	 case 'h' :     /* hex 8 */
	   if ((p->values[rang].format==F_SIGNED) || (p->values[rang].format==F_UNSIGNED))
	   snprintf(p->printvalue+j,MAX_LINE-j,"%02lx",*((UNSIG_DEC *)(p->values[rang].val)));
	   break;
	 }
	 while(p->printvalue[j]!='\0') j++;  /* jump */
	 i+=to_add;
    } else {
	 p->printvalue[j]=p->dspfmt[i];
	 j++;
    }
  }

  return(1);
}

/********** 
		  modifiy variable
		  p: the pointer to variable
		  rang : the index IN the variable list
		  format : the format
		  value : then pointer to the value.
		  length : (optionnal) the length (in octet) of the value
***********/
int modify_variable (variable *p, int rang, const char format, const void *value, int length) {
  variable_value *vv;


  vv=&(p->values[rang]);

  /* freeing the previous value */
  if (vv->format!=F_UNDEF) free(vv->val);

  vv->format=format;


  /* stocking value */
  switch (format) {
  case F_STRING:
    vv->val=malloc(strlen(value)+1);
    memcpy(vv->val,value,strlen(value)+1);
    vv->length=-1;
    break;
  case F_BINAR:
    vv->val=malloc(length);
    memcpy(vv->val,value,length);
    vv->length=length;
    break;
  case F_SIGNED:
    vv->val=malloc(sizeof(SIG_DEC));
    memcpy(vv->val,value,sizeof(SIG_DEC));
    vv->length=-1;
    break;
  case F_UNSIGNED:
    vv->val=malloc(sizeof(UNSIG_DEC));
    memcpy(vv->val,value,sizeof(UNSIG_DEC));
    vv->length=-1;
    break;
  }
  
  if (rang>=p->num)
    p->num=rang+1;

  /*** recalculate the printvalue ***/
  modify_display(p);

  return(1);
} 

/*******
 *******/
int duplicate_value(variable_value *dst, variable_value *src) {

  if (!src) {
    dst=NULL;
    return(1);
  }

  dst->length=src->length;
  dst->format=src->format;
  dst->length=src->length;

  if (dst->val) free(dst->val);

  switch (dst->format) {
  case F_STRING:
    dst->val=malloc(strlen((char *)(src->val))+1);
    memcpy((char *)(dst->val),(char *)(src->val),strlen((char *)(src->val))+1);
    break;
  case F_BINAR:
    dst->val=malloc(dst->length);
    memcpy(dst->val,src->val,dst->length);
    break;
  case F_SIGNED:
    dst->val=malloc(sizeof(SIG_DEC));
    memcpy(dst->val,src->val,sizeof(SIG_DEC));
    break;
  case F_UNSIGNED:
    dst->val=malloc(sizeof(UNSIG_DEC));
    memcpy(dst->val,src->val,sizeof(UNSIG_DEC));
    break;
  case F_UNDEF:
    dst->val=NULL;
    break;
  }
  return(1);
}

/********
	    get_variable_value
	    p: the variable
	    rang: the rang in table.
	          if rang == NO_RANG_DEF : the rang is not defined.
			 so return the string value if the variable is composed 
			 by more than 1 value, otherwise return the value itself.
			if rang == STRING_FORCE : return the string value 
			                          (ie the printvalue).
	    format: a ptr to the return format. (must be allocated)
	    return NULL if pbs, or the pointer.
********/
variable_value *get_variable_value(variable *p, int rang) {
  variable_value *vv;

  if (p==NULL) return(NULL);

  vv=malloc(sizeof(variable_value));
  vv->val=NULL;

  if (((rang==NO_RANG_DEF) && (p->num==1)) || (rang>=0)) {
    if (rang==NO_RANG_DEF) rang=0; 
    duplicate_value(vv,&(p->values[rang]));
    return(vv);
  }


  /* the string by default (STRING_FORCE is the default) */
  vv->length=-1;
  vv->format=F_STRING;
  vv->val=malloc(strlen(p->printvalue)+1);
  memcpy((char *)(vv->val),p->printvalue,strlen(p->printvalue)+1);
  return(vv);

} 

/*******
	duplicate variable src into variable dst...
	return 0;
*******/ 
int duplicate_variable(variable *dst, variable *src) {
  int i;

  memcpy(dst->dspfmt,src->dspfmt,MAX_LINE);
  memcpy(dst->printvalue,src->printvalue,MAX_LINE);

  for (i=0 ; i<dst->num ; i++) free(dst->values[i].val);
  dst->num=src->num;
  for (i=0 ; i<dst->num ; i++) {
    duplicate_value(&dst->values[i],&src->values[i]);
  }

  return(0);
}

/******
       get_file_to_load
       return the file to load
******/

char *get_file_to_load(char *label) {
  int i;
  int len;
  char *p;
  len=strlen(label);

  for (i=0 ; i<len ; i++) {
    if (label[i]=='/') {
      p=malloc(len-i+LENGTH_EXTEND);
      memcpy(p,label+i+1,len-(i+1));
      memcpy(p+len-(i+1),EXTEND,LENGTH_EXTEND);
      p[len-(i+1)+LENGTH_EXTEND]='\0';
      label[i]='\0';   /* lost the label file */
      return(p);
    }
  }

  p=malloc(len+LENGTH_EXTEND+1);
  memcpy(p,label,len);
  memcpy(p+len,EXTEND,LENGTH_EXTEND);
  p[len+LENGTH_EXTEND]='\0';
  return(p);
}


/****************************
 ********** CALCUL **********
 ****************************/
/*******
	   ascii_to_var
	   return a variable_value
	   by default return a SIGNED value.
********/
variable_value *ascii_to_var(const char *value, int len) {

  SIG_DEC *sa=malloc(sizeof(SIG_DEC));
  char *pt;
  int base=10;
  int offset=0;
  variable_value *vv=malloc(sizeof(variable_value));


  if (len==0) return (NULL);

  pt=NULL;
  /*** testing a hex value ***/
  if ((len>2) && (value[0]=='0') && (value[1]=='x')) {
    base=16;
    offset=2;
  }  
  /*** ------------------- ***/

  /*** a numeric value ***/
  *sa=strtol(value+offset,&pt,base);
  if (pt!=value+offset) {
	 vv->format=F_SIGNED;
	 vv->val=(void *)sa;
	 vv->length=-1;
	 return(vv);
  }
  /*** --------------- ***/

  /*** a string value  ***/
  free(sa);  
  MY_STRDUP(((char *)(vv->val)),value,len);
  vv->format=F_STRING;
  vv->length=-1;
  return(vv);
  /*** --------------- ***/
  
}

/*******
	   get format
	   return a inteligible form of the format string
********/
char get_format( char *a) {
  if (strcmp(a,FORMAT_U)==0) return(F_UNSIGNED);
  if (strcmp(a,FORMAT_D)==0) return(F_SIGNED);
  if (strcmp(a,FORMAT_S)==0) return(F_STRING);
  if (strcmp(a,FORMAT_B)==0) return(F_BINAR);
  return(F_ERROR);
}

/*******
	   format_to_str
	   return a inteligible form of the format string
	   ex :
********/
char *format_to_str( char format) {
  if (format==F_UNSIGNED) return(FORMAT_U);
  if (format==F_SIGNED) return(FORMAT_D);
  if (format==F_STRING) return(FORMAT_S);
  if (format==F_BINAR) return(FORMAT_B);
  return(NULL);
}

/***************************
 ********** TABLES **********
 ***************************/
/*******
	   get_label_table
	   return the label name in funct of the variable
	   otherwise, return NULL
 *******/
table_in *get_label_table(char *table_name, char *variable_name, tablestruct *ttfirst, params *parm) {
  
  variable *p;
  tablestruct *tt;
  table_in *tin, *tin_result=NULL;
  char keep='\0';
  variable_value *value;
  variable_value *keyv;
  int rang;
  int len;

  /* look for the rang */
  rang=NO_RANG_DEF;
  len=strlen(variable_name);
  keep=variable_name[len-2];
  if ((variable_name[len-2]=='.') && (variable_name[len-1]>='0') && (variable_name[len-1]<='9')) {
    rang=variable_name[len-1]-'0';
    variable_name[len-2]='\0';
  }

  p=get_variable_by_name(variable_name, parm,FOLLOW_LABEL_PATH);
  variable_name[len-2]=keep;
  if (p==NULL) return(NULL);
  value=get_variable_value(p,rang);

  /* look for the table */
  tt=ttfirst;
  while((tt!=NULL) && (strcmp(tt->name,table_name))) tt=tt->next;
  if (tt==NULL) {
    add_display(parm,DE_PROTO,-1,"Table %s not definied\n",table_name); 
    return(NULL);
  }
  
  /* comp keys of the table */
  tin=tt->in;
  while(tin!=NULL) {
    /* add_display(parm,D_DEBUG,-1,"Table %s key %s (%d)\n",table_name,tin->key, tin);  */

    /**** the key is '*' (the default) ****/ 
    if ((tin->key[0]=='*') && (tin->key[1]=='\0')) {
      tin_result=tin;
      tin=NULL;
    } else {
      keyv=ascii_to_var(tin->key,strlen(tin->key));
      if (compare(value,keyv,F_EQ)) {
	tin_result=tin;
	tin=NULL;
      } else tin=tin->next;
      free(keyv->val);
      free(keyv);
    }
    /**** ---------------------------- ****/ 
    
  }
  
  if (tin_result==NULL) {
    if (value->format==F_UNSIGNED) 
	 add_display(parm,DE_PROTO,-1,"Key %s=%d(%d) not found in table %s.",variable_name,*((UNSIG_DEC *)value->val), value->format, table_name);
    if (value->format==F_SIGNED) 
	 add_display(parm,DE_PROTO,-1,"Key %s=%d(%d) not found in table %s.",variable_name,*((SIG_DEC *)value->val), value->format, table_name); 
    if (value->format==F_STRING) 
	 add_display(parm,DE_PROTO,-1,"Key %s=%s(%d) not found in table %s.",variable_name,(char *)value->val, value->format, table_name); 

  }

  free(value->val);
  free(value);
  return(tin_result);
}

/***************************
 ********** FILES **********
 ***************************/
/********
	    search_in_path
	    cherche le fichier dans le chemin
	    path : le chemin
	    file : le nom de fichier
	    retourne un pointeur sur le fichier, sinon NULL
*********/
char *search_in_path (char *path, char *file) {
  char *ff;
  int deb=0; 
  int fin=0;
  int fd;

  ff=malloc(strlen(path)+strlen(file)+2);

  /* On regarde pour chaque elt du path */
  while(*(path+deb)!='\0') {
    while((*(path+fin)!=':') && (*(path+fin)!='\0')) {
	 fin++;
    }

    /* concatenation */
    memcpy(ff,path+deb,fin-deb);
    memcpy(ff+fin-deb,"/",1);
    memcpy(ff+fin-deb+1,file,strlen(file));
    memcpy(ff+fin-deb+strlen(file)+1,"\0",1);

    /* On essaye de l'ouvrir */
    if ((fd=open(ff,O_RDONLY))>-1) {
	 close(fd);
	 return(ff);
    }

    if (*(path+fin)==':') fin++;

    deb=fin;
  }
  
  free(ff);
  return(NULL);
}



/***********************************
 ********* STOCK / UNSTOCK *********
 ***********************************/

/*****
      add a new stock struct in the chain,
	 or get the stock struct.
	 flag=DONT_ADD -> dont add if not found
	 flag=ADD -> add to the struct if not found
******/ 
static stock_chain *get_or_add_in_stock(params *parm, struct stock_struct stock, int flag) {
   stock_chain *st, *prev;
   int i, right, len;
   variable *vp;
   variable_value *varval[10];

   /* adding to varval tab */
   for (i=0 ; i<stock.num ; i++) {
	vp=get_variable_by_name(stock.name[i],parm,FOLLOW_LABEL_PATH);
	varval[i]=get_variable_value(vp, STRING_FORCE);
   }   

   /* retreiving the good stock struct */
   st=(*parm->stock_start);

   prev=st;
   while (st!=NULL) {
	right=0;
	if (stock.num==st->num) {
	  right=1;
	  for (i=0 ; i<stock.num ; i++) {
	    if (strcmp(st->key[i],(char *)(varval[i]->val))!=0) {
		 right=0;
		 i=stock.num;
	    }
	  } /* end for */
	}
	if (right) {
	  for (i=0 ; i<stock.num ; i++) {
	    free(varval[i]->val);
	    free(varval[i]);
	  }
	  return(st);
	}

	prev=st;
	st=st->next;
   } /* end while */

   /*** Adding a new stock ***/
   if ((st==NULL) && (flag==ADD)) {
	st=malloc(sizeof(stock_chain));
	st->next=NULL;
	st->length=0;
	if (*(parm->stock_start)==NULL) *(parm->stock_start)=st;
	else prev->next=st;
	st->num=stock.num;
	st->frag=NULL;     /* initialise frag */
	st->length=-1;

	/* adding keys */
	for (i=0 ; i<stock.num ; i++) {
	  len=strlen((char *)(varval[i]->val));
	  st->key[i]=malloc(len+1);
	  memcpy(st->key[i],(char *)(varval[i]->val),len);
	  st->key[i][len]='\0';
	}
   }
   /*** ------------------ ***/

   /*** freing varval ***/
   for (i=0 ; i<stock.num ; i++) {
     free(varval[i]->val);
     free(varval[i]);
   }
   /*** ------------- ***/

   return(st);
}

/**********
		 stock
		 stock the packet in a big table, before
		 reading it
		 parm is used to get the stock_start;
		 offset is used to place the packet in the table
		 if offset==-1, place to the end.
		 p is the packet.
***************/
int stock(params *parm, struct stock_struct stock, int offset, int length, int tot_length) {
   stock_chain *st;
   stock_fragment *fr, *fr1, *fr2;

   /* get the right packet */
   st=get_or_add_in_stock(parm,stock,ADD);
   
   /* add the first fragment into the list */
   if (st->frag==NULL) {
	st->frag=malloc(sizeof(stock_fragment));
	st->frag->next=NULL;
	st->frag->length=length;
	st->frag->offset=((offset>-1)?offset:0);
	st->frag->p=malloc(length+1);
	memcpy(st->frag->p,parm->p+(parm->offoct),length);
	if (tot_length>-1) st->length=tot_length;
	return(0);
   }
   
   /* add the fragment into the list, at the right place */
   fr=st->frag;
   fr1=NULL;
   
   while(fr!=NULL) {
	if ((offset>-1)&&(fr->offset>=offset)) break;
	fr1=fr;
	fr=fr->next;
   }
   /* must be inserted between fr1 and fr */
   fr2=malloc(sizeof(stock_fragment));
   if (fr1!=NULL) fr1->next=fr2;
   else st->frag=fr2;
   fr2->next=fr;
   fr2->length=length;
   fr2->offset=((offset>-1)?offset:0);
   fr2->p=malloc(length+1);
   memcpy(fr2->p,parm->p+(parm->offoct),length);
   
   /* fr=st->frag;
   while(fr) {
	add_display(parm,D_DEBUG,-1,"debug list: offset=%d, len=%d\n",fr->offset,fr->length);
	fr=fr->next;
	} */
   if (tot_length>-1) st->length=tot_length;

   return(0);
}


/**********
		 unstock
		 get the packet from the stock_struct.
		 return -1 if pb.
		 return the length added in the packet.
***************/
int unstock(params *parm, struct stock_struct stock) {
   stock_chain *st, *prev;
   stock_fragment *fr, *fr1;
   int old_packet_size;
   int len=0;
   u_char *p;

   st=get_or_add_in_stock(parm,stock,DONT_ADD);   
   if (st==NULL) {
	add_display(parm,D_WARNING,-1,"Unable to find the previous packet in stock."); 
	return(-1);
   }

   if (st->frag==NULL) {
	add_display(parm,D_WARNING,-1,"packet contain no fragments."); 
	return(-1);
   }

   /* all fragments are presents ? */
   fr=st->frag;
   while(fr!=NULL) {
	/* add_display(parm,D_DEBUG,-1,"debug : offset=%d, len=%d\n",fr->offset,fr->length); */
	if (fr->offset!=len) break;
	len+=fr->length;
	fr=fr->next;
   }
   
   /* all frags are not presents */
   if (fr!=NULL) {
	add_display(parm,D_INFO,-1,"all frags are not here."); 
	return(-1);
   }
   
   /* if the length OK ? */
   if ((st->length==-1)||(st->length!=len)) {	/* add_display(parm,DE_PROTO,-1,"length not OK.(%d sur %d) \n",st->length,len);  */
	return(-1);
   }

   /* OK, the packet is OK !!! */
   len=0;
   p=malloc(st->length+parm->packet_size/8); /* the packet + the new one */
   memcpy(p,parm->p,parm->offoct);	
   /* copy all frags */
   fr=st->frag;
   while(fr) {
	memcpy(p+parm->offoct+len,fr->p,fr->length);
	len+=fr->length;
	fr=fr->next;
   }
   parm->p=p;
   parm->free_p=1;
   old_packet_size=parm->packet_size;
   parm->packet_size=(parm->offoct+st->length)*8;
   
   /* freeing */
   fr=st->frag;
   while(fr) {
	fr1=fr;
	free(fr->p);
	fr=fr->next;
	free(fr1);
   }   
   if (st==(*(parm->stock_start))) {
	(*(parm->stock_start))=st->next;
   } else {
	prev=(*(parm->stock_start));
	while(prev->next!=st) {
	  prev=prev->next;
	}
	prev->next=st->next;
   }
   free(st);

   add_display(parm,D_INFO,-1,"Packet reassembled."); 
   return(parm->packet_size-old_packet_size);
}


/***********************************
 ********* FLUX ********************
 ***********************************/

/******
 flux_exist
 test if a flux exist in the flux chain .
 if yes, return the ptr to the flux, otherwise, return NULL. 
 ******/

flux_chain *flux_exist (struct flux_struct flux, params *parm) {
  flux_chain *fl;
  int i, right;
  variable *vp;
  variable_value *varval[10];
  
  /*** adding to varval tab ***/
  for (i=0 ; i<flux.num ; i++) {
    vp=get_variable_by_name(flux.name[i],parm,FOLLOW_LABEL_PATH);
    varval[i]=get_variable_value(vp, STRING_FORCE);
  }   
  /*** -------------------- ***/
  
  /*** searching in the flux list ***/
  fl=(*(parm->flux_start));
  while (fl!=NULL) {
    right=0;
    if (flux.num==fl->num) {
	 right=1;
	 for (i=0 ; i<flux.num ; i++) {
	   if (strcmp(fl->key[i],(char *)(varval[i]->val))!=0) {
		right=0;
		i=flux.num;
	   }
	 } /* end for */
    }
  /*** -------------------------- ***/
    
    if (right) {

      /*** freing varval ***/
      for (i=0 ; i<flux.num ; i++) {
	free(varval[i]->val);
	free(varval[i]);
      }
      /*** ------------- ***/
      return(fl);
    }

    fl=fl->next;
  } /* end while */

   /*** freing varval ***/
   for (i=0 ; i<flux.num ; i++) {
     free(varval[i]->val);
     free(varval[i]);
   }
   /*** ------------- ***/
  return(NULL);
}

/********
   add_flux
   create a flux into the flux chain 
   and return a ptr tu the flux.
 *******/

flux_chain *add_flux (struct flux_struct flux, params *parm) {
  flux_chain *fl;
  variable *vp;
  int i, len;
  variable_value *varval[10];

  fl=(*(parm->flux_start));
  if (fl==NULL) {
    fl=malloc(sizeof(flux_chain));
    (*(parm->flux_start))=fl;
    fl->next=NULL;
  } else {
    fl=malloc(sizeof(flux_chain));
    fl->next=(*(parm->flux_start))->next;
    (*(parm->flux_start))->next=fl;
  }


  /* adding to varval tab */
  for (i=0 ; i<flux.num ; i++) {
    vp=get_variable_by_name(flux.name[i],parm,FOLLOW_LABEL_PATH);
    varval[i]=get_variable_value(vp, STRING_FORCE);
  }   

  fl->num=flux.num;
  fl->prots_index=0;
  /*** initialise all protocols ***/
  bzero(fl->prots,sizeof(protocol)*100);
  /*** ------------------------ ***/

  /* adding keys */
  for (i=0 ; i<flux.num ; i++) {
    len=strlen((char *)(varval[i]->val));
    fl->key[i]=malloc(len+1);
    memcpy(fl->key[i],varval[i]->val,len);
    fl->key[i][len]='\0';
  }
 
  /*** freing varval ***/
  for (i=0 ; i<flux.num ; i++) {
    free(varval[i]->val);
    free(varval[i]);
  }
  /*** ------------- ***/
  
  /* remember the flux struct */
  return(fl);
}


/*******
	get_max_size
	return the maximum size in the packet
	(in bits).
********/
int get_max_size(params *parm, protocol *prot) {
  int phy_size;        /* the physical size */
  int prot_size;       /* the size in the protocol */
  
  phy_size=parm->packet_size-parm->offoct*8-parm->offbit;
  if (prot->user_length_in_bits>-1)
    prot_size=prot->user_length_in_bits-(*(prot->nb_bits_read));
  else
    prot_size=prot->max_length_in_bits-(*(prot->nb_bits_read));
    
  return(MY_MIN(phy_size,prot_size));
}
