#include "nicedump.h"

#ifdef ENABLE_REGEX
#include<regex.h>
#endif

/* func declarations */
static int evaluation_stage2 (compiled_str *, params *);

/******* externs from analyse.c ********/
extern variable *get_variable_by_name(char *, params *, char);
extern variable_value *get_variable_value(variable *, int);
extern int compare(variable_value *,variable_value *,char );
extern variable_value *ascii_to_var(char *, int);
extern char *format_to_str( char);
extern variable *add_new_variable (char *, const char *, params *);
extern int duplicate_value(variable_value *, variable_value *);
extern int modify_variable (variable *, int , const char , const void *, int );

/******* externs from packet.c *******/

extern confvalue *get_label_ptr(char *, confvalue *);
#ifdef STDC_HEADERS
extern int add_display(params *, const char, int, ...);
#else
extern int add_display(params *, const char, int);
#endif

/******* externs from outside *******/
extern void bzero(void *, int );

/*************************************/


/**********************************************************/
/************ Comparison, calculation *********************/
/**********************************************************/

#define SA (*((SIG_DEC *)a->val))
#define UA (*((UNSIG_DEC *)a->val))
#define STA ((char *)a->val)
#define BA (a->val)

#define SB (*((SIG_DEC *)b->val))
#define UB (*((UNSIG_DEC *)b->val))
#define STB ((char *)b->val)
#define BB (b->val)

#define SC (*((SIG_DEC *)c->val))
#define UC (*((UNSIG_DEC *)c->val))
#define STC ((char *)c->val)
#define BC (c->val)

/*****
      This func make somme context swithing if possible
      first, it convert from b->format to string, 
      next, convert from string to dest_format
 *****/
static variable_value *convert_value(char dest_format, variable_value *b) {
  variable_value *c=malloc(sizeof(variable_value));
  char str[MAX_LINE];
  char *pt=NULL;
  int len;
  int base=10;        /* in hex or in decimal ? */
  int offset=0;       /* tu switch the "0x" in case of hex */

  c->format=dest_format;
  c->length=-1;
  c->val=NULL;

  /*** Convert to string ***/
  bzero(str,MAX_LINE);
  
  switch (b->format) {
  case F_STRING:
    snprintf(str,MAX_LINE,"%s",STB);
    break;
  case F_SIGNED:
    snprintf(str,MAX_LINE,"%ld",SB);
    break;
  case F_UNSIGNED:
    snprintf(str,MAX_LINE,"%lu",UB);
    break;
  default:
    c->format=F_ERROR;
    return(c);
    break;
  }  
  /*** ----------------- ***/

  /*** Convert from string to ... ***/
  len=strlen(str);
  if ((len>2) && (str[0]=='0') && (str[1]=='x')) {
    base=16;
    offset=2;
  }

  switch (dest_format) {
  case F_STRING:
    c->val=malloc(MAX_LINE);
    memcpy(STC,str,MAX_LINE);
    break;
  case F_SIGNED:
    c->val=malloc(sizeof(SIG_DEC));
    SC=strtol(str+offset,&pt,base);
    break;
  case F_UNSIGNED:
    c->val=malloc(sizeof(UNSIG_DEC));
    UC=strtol(str+offset,&pt,base);
    break;
  default:
    c->format=F_ERROR;
    break;
  }  
  /*** ----------------- ***/

  return(c);
}

/****
	do the calcul of "a = a (ope) b"
	make some context switching if
	necessary
****/
static int calculate (variable_value *a, variable_value *b, char operator) {
  int isok=1;
  variable_value *c;
  int free_c=NO;

  /*** context switching ***/
  if (a->format==b->format) c=b;
  else {
    c=convert_value(a->format,b);
    free_c=YES;
    if (c->format==F_ERROR) {
      free(c->val);
      free(c);
      return(-1);
    }
  }
  /*** ----------------- ***/

  switch (operator) {
  case F_MOD:
    switch(a->format) {
    case F_SIGNED:
      SA-=(SA/SC)*SC;
      break;
    case F_UNSIGNED:
      UA-=(UA/UC)*UC;
      break;
    default:
      isok=-1;
    }
    break;
  case F_PLUS:
    switch(a->format) {
    case F_SIGNED:
      SA+=SC;
      break;
    case F_UNSIGNED:
      UA+=UC;
      break;
    case F_STRING:
      isok=-1;
      break;
    default:
      isok=-1;
    }
    break;
  case F_MIN:
    switch(a->format) {
    case F_SIGNED:
      SA-=SC;
      break;
    case F_UNSIGNED:
      UA-=SC;
      break;
    default:
      isok=-1;
    }
    break;
  case F_MUL:
    switch(a->format) {
    case F_SIGNED:
      SA*=SC;
      break;
    case F_UNSIGNED:
      UA*=SC;
      break;
    default:
      isok=-1;
    }
    break;
  case F_DIV:
    switch(a->format) {
    case F_SIGNED:
      SA/=SC;
      break;
    case F_UNSIGNED:
      UA/=SC;
      break;
    default:
      isok=-1;
    }
    break;
  default:
    isok=-1;
  }

  /* printf("calculate a=%d b=%d ope=%c\n",(*((SIG_DEC *)a->val)),(*((SIG_DEC *)b->val)),operator); */
 
  if ((free_c)&&(c)) {
    free(c->val);
    free(c);
  }

  /*   if (isok>0) printf("calculate is goot =%d\n",SA); */
  return(isok);
}


/*******
	   compare
	   compare two value. in function of
	   the type of comparison (F_EQ, F_SUP, etc...)
	   return 1 of comparison is OK,
	   otherwise return 0.
	   return -1 if error.

********/
int compare (variable_value *a, variable_value *b, char type) {

#ifdef ENABLE_REGEX
  regex_t preg;
#endif

  variable_value *c;
  int free_c=NO;

  int resp=0;                         /* the response       */
  int rep;                            /* temporary response */

  /*** context switching ***/
  if (a->format==b->format) c=b;
  else {
    c=convert_value(a->format,b);
    free_c=YES;
    if (c->format==F_ERROR) {
      free(c->val);
      free(c);
      return(-1);
    }
  }
  /*** ----------------- ***/

  
  switch (type) {
  case F_SUP:  /* > */
    switch(a->format) {
    case F_SIGNED:
      if (SA>SC) resp=1;
      break;
    case F_UNSIGNED:
      if (UA>UC) resp=1;
      break;
    case F_STRING:
      if (strcmp(STA,STC)>0) resp=1;
      break;
    default:
      resp=-1;
    }
    break;
  case F_INF:  /* < */
    switch(a->format) {
    case F_SIGNED:
      if (SA<SC) resp=1;
      break;
    case F_UNSIGNED:
      if (UA<UC) resp=1;
      break;
    case F_STRING:
      if (strcmp(STA,STC)<0) resp=1;
      break;
    default:
      resp=-1;
    }
   break;
  case F_EQ:  /* == */
    switch(a->format) {
    case F_SIGNED:
      if (SA==SC) resp=1;
      break;
    case F_UNSIGNED:
      if (UA==UC) resp=1;
      break;
    case F_STRING:
      if (strcmp(STA,STC)==0) resp=1;
      break;
    default:
      resp=-1;
    }
    break;
  case F_NE:  /* != */
    switch(a->format) {
    case F_SIGNED:
      if (SA!=SC) resp=1;
      break;
    case F_UNSIGNED:
      if (UA!=UC) resp=1;
      break;
    case F_STRING:
      if (strcmp(STA,STC)!=0) resp=1;
      break;
    default:
      resp=-1;
    }
    break;
  case F_SE:  /* >= */
    switch(a->format) {
    case F_SIGNED:
      if (SA>=SC) resp=1;
      break;
    case F_UNSIGNED:
      if (UA>=UC) resp=1;
      break;
    case F_STRING:
      if (strcmp(STA,STC)>=0) resp=1;
      break;
    default:
      resp=-1;
    }
    break;
  case F_IE:  /* <= */
    switch(a->format) {
    case F_SIGNED:
      if (SA<=SC) resp=1;
      break;
    case F_UNSIGNED:
      if (UA<=UC) resp=1;
      break;
    case F_STRING:
      if (strcmp(STA,STC)<=0) resp=1;
      break;
    default:
      resp=-1;
    }
    break;
  case F_IN:  /* =~ */
#ifdef ENABLE_REGEX
    if ((a->format==F_STRING) && (!regcomp(&preg,STB,REG_NOSUB))){
      rep=regexec(&preg,STA,0,NULL,0);
      resp=((rep==0)?1:0);
      regfree(&preg);
    } else return(-1);
    break;
#else
    break;
    return(-1);
#endif
    break;
 case F_NI:  /* !~ */
#ifdef ENABLE_REGEX
    if ((a->format==F_STRING) && (!regcomp(&preg,STB,REG_NOSUB))){
      /* printf("not regcomp %s et %s\n",sta,stb); */
      rep=regexec(&preg,STA,0,NULL,0);
      resp=((rep==0)?0:1);
      regfree(&preg);
     } else return(-1);
    break;
#else
    break;
    return(-1);
#endif
  }

  if ((free_c)&&(c)) {
    free(c->val);
    free(c);
  }

  return(resp);
}

/**********************************************************/
/************ String analyse          *********************/
/**********************************************************/

/******
       name : is_start_word
       author : BW
       used to know if this char is a beginning of a word
       (ie a variable, a text)
       return 1 if is a start of word
       return 0 if not.
*******/

int is_start_word(char a) {
  if ((a>='0') && (a<='9')) return(1);
  if ((a>='A') && (a<='Z')) return(1);
  if ((a>='a') && (a<='z')) return(1);
  if ((a=='_') || (a=='"') ||(a=='!') || (a=='-') || (a==':')) return(1);
  return(0);
}




/******
       name : new_ptr
       author : BW
       this func is used to had a new compiled_str struct in the chain,
       and initialise it.
       it return the ptr.
       params : the previous ptr, we had the new one after the prev_ptr.
*******/
static compiled_str *new_ptr(compiled_str *prev_ptr) {
  compiled_str *ptr;

  ptr=malloc(sizeof(compiled_str));
  ptr->preset_operator=F_FIN;
  ptr->preset_neg=0;
  ptr->preset_inc=0;
  ptr->preset_test_result=1;
  ptr->preset_var=NULL;
  ptr->preset_value=NULL;
  ptr->preset_down=NULL;

  if (prev_ptr) {
    prev_ptr->preset_next=ptr;
    ptr->preset_prev=prev_ptr;
    ptr->prev=prev_ptr;
  } else {
    ptr->preset_prev=NULL;
    ptr->prev=NULL;
  }
  ptr->preset_next=NULL;
  ptr->next=NULL;

  ptr->operator=F_FIN;
  ptr->neg=0;
  ptr->inc=0;
  ptr->test_result=1;
  ptr->str=NULL;
  ptr->var=NULL;
  ptr->value=NULL;
  ptr->down=NULL;

  return(ptr);
}

/******
       name : registration
       author : BW
       this func is used to add a new element
	  in the chain compiled_str. or modify only the
	  operator.
*******/
static compiled_str *registration(compiled_str *prev_ptr, 
			char *cond,
			int start, 
			int end,
			int *negation,
			int *increment,
			compiled_str **first,
			char ope,
			params *parm) {

  compiled_str *ptr;

  /*** copying string, etc ***/
  if ((start>=0) && (end>=0)) {

    ptr=new_ptr(prev_ptr);
    if ((*first)==NULL) (*first)=ptr;

    MY_STRDUP(ptr->str,cond+start,end-start+1);

    ptr->preset_neg=(*negation);
    ptr->neg=ptr->preset_neg;
    (*negation)=0;

    ptr->preset_inc=(*increment);
    ptr->inc=ptr->preset_inc;
    (*increment)=0;
   
    ptr->preset_operator=ope;
    ptr->operator=ptr->preset_operator;
    return(ptr);
  }
  /*** -------------------------------------------- ***/

  /*** modification of the operator ***/
  if (prev_ptr==NULL) {
    add_display(parm,DE_SYNTAX,-1,"starting \"%s\" with an operator ?",cond+start);
    return(NULL);
  }

  /* printf("re-registre : <%s> ope=%c\n",prev_ptr->str,ope); */

  if (prev_ptr->preset_operator!=F_FIN) {
    add_display(parm,DE_SYNTAX,-1,"I don't know wy an operators is here at \"%s\" ?",cond+start);
  }
  prev_ptr->preset_operator=ope;
  prev_ptr->operator=ope;
  return(prev_ptr);
  /*** ---------------------------- ***/

}



static compiled_str *evaluation_stage1(char *cond,int start,int end, params *parm) {  
  int i;
  int numpar=0;       /* the number of () */
  int inastring=NO;   /* boolean to know if we are in a string " */
  int inacote=NO;     /* boolean to know if we are in a string ' */
  int inaslash=NO;    /* boolean to know if we are in a string / (regex) */
  int back=NO;        /* if the previous char was a \ */

  int st=-1;          /* start of the sub chain */
  int ed=-1;          /* end of the sub chain */
  int negation=NO;
  int increment=NO;
  int waiting_for_slash=NO;  /* use after =~ or !~ operator */

  compiled_str *ptr=NULL;        /* the list of result */
  compiled_str *first_ptr=NULL;  /* the first one of the list */ 
  char *strerror;             /* the substring to display error */

  st=start;
  
  /* printf("Get in %d\n",ptr); */
  for (i=start ; i <=end ; i++) {
    /* ntf("a char cond[%d]=%c (st=%d,ed=%d) par=%d string=%d cote=%d\n",i,cond[i],st,ed,numpar,inastring,inacote); */
   
    /* a space */
    if ((cond[i]==' ') || (cond[i]=='\t')) {
      if ((st==i) && (inastring==NO) && (inacote==NO) && (inaslash==NO)) st++;
      continue;
    }

    /* ignore because it was a \<some char> */
    if (back==YES) {
      back=NO;
      continue;
    }
  
    /* a ( */
    if (cond[i]=='(') {
      if (numpar==0) st=i+1;
      numpar++;
    }

    /* a ) */
    if (cond[i]==')') {
      numpar--;
      if (numpar==0) {
	ed=i-1;
	ptr=registration(ptr,cond,st,ed,&negation,&increment, &first_ptr, F_FIN, parm);
	free(ptr->str);
	ptr->str=NULL;
	if (!first_ptr) first_ptr=ptr;
	/* printf("test1 : A sub= %d-> ???? \n",ptr); */
	ptr->preset_down=evaluation_stage1(cond,st,ed,parm);
	ptr->down=ptr->preset_down;
	/* printf("test1 : A sub= %d->%d\n",ptr, ptr->down); */
	st=i+1;
	ed=-1;
      }
      continue;
    }

     /* \ something */
    if (cond[i]=='\\') {
      i++;
    }

     /* a string */
    if (cond[i]=='"') {
      if (inacote) continue;
      if (inastring) {
	inastring=NO;
	ed=i;
      } else {
	inastring=YES;
	st=i;
      }
    }

    /* a cote */
    if (cond[i]=='\'') {
      if (inastring) continue;
      if (inacote) {
	inacote=NO;
	ed=i;
      } else {
	inacote=YES;
	st=i;
      }
    }
   
    /* the end of a regex */
    if ((numpar==0) && (inastring==NO) && (inacote==NO) 
	&& (inaslash==YES) && (cond[i]=='/')) {
      ed=i-1;
      ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_FIN, parm);
      st=i+1;
      ed=-1;
      inaslash=NO;
      continue;
    }

    /* a normal char, looking for the operator */
    if ((numpar==0) && (inastring==NO) && (inacote==NO) && (inaslash==NO)) {

      switch (cond[i]) {
      case '|' :                    /* OR */
	if ((i>start) && (cond[i+1]=='|')) {
	  ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_OR, parm);
	  i++;
	  st=i+1;
	  ed=-1;
	}
	break;
      case '&' :                    /* AND */
	if ((i>start) && (cond[i+1]=='&')) {
	  ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_AND, parm);
	  i++;
	  st=i+1;
	  ed=-1;
	}
	break;
      case '<' :                    /* <= and < */
	if ((i<end) && (cond[i+1]=='=')) {
	  ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_IE, parm);
	  i++;
	  st=i+1;
	  ed=-1;
	} else {
	  ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_INF, parm);
	  st=i+1;
	  ed=-1;
	}
	break;
      case '>' :                    /* >= and > */
	if ((i<end) && (cond[i+1]=='=')) {
	  ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_SE, parm);
	  i++;
	  st=i+1;
	  ed=-1;
	} else {
	  ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_SUP, parm);
	  st=i+1;
	  ed=-1;
	}
	break;
      case '=' :                    /* == and =? and = */
	if ((i<end) && (cond[i+1]=='=')) {
	  ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_EQ, parm);
	  i++;
	  st=i+1;
	  ed=-1;
	} else {
	  if ((i<end) && (cond[i+1]=='~')) {
	    ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_IN, parm);
	    i++;
	    st=i+1;
	    ed=-1; 
	    waiting_for_slash=YES;
	  } else {
	    ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_AFF, parm);
	    st=i+1;
	    ed=-1;
	  }
	}
      break;
      case '!' :                    /* == and =? and = */
	if ((i<end) && (cond[i+1]=='=')) {
	  ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_NE, parm);
	  i++;
	  st=i+1;
	  ed=-1;
	} else {
	  if ((i<end) && (cond[i+1]=='~')) {
	    ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_NI, parm);
	    i++;
	    st=i+1;
	    ed=-1;
	    waiting_for_slash=YES;
	  } else {
	    if ((i<end) && (!is_start_word(cond[i+1]))) {
	      negation=1;
	      st=i+1;
	      ed=-1;
	    }
	  }
	}
      break;
      case '%' :                    /* % */
	ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_MOD, parm);
	st=i+1;
	ed=-1;
	break;
      case '+' :                    /* + */
	ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_PLUS, parm);
	st=i+1;
	ed=-1;
	break;
      case '-' :                    /* - */
	ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_MIN, parm);
	st=i+1;
	ed=-1;
	break;
      case '*' :                    /* * */
	ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_MUL, parm);
	st=i+1;
	ed=-1;
	break;
      case '/' :                    /* / */
	/* / can be the division (bell), or start of a regex,
	   or end of a regex */
	/*** start of a regex / ***/
	if (waiting_for_slash==YES) {
	  st=i+1;
	  ed=-1;
	  waiting_for_slash=NO;
	  inaslash=YES;
	  break;	  
	}

	ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_DIV, parm);
	st=i+1;
	ed=-1;
	break;
      default :
	if ((cond[i]!='\'') && (cond[i]!='"'))
	  ed=i;
	break;
      } /* end switch */
    }
  } /* end for */

  
  /*** the last world ***/
  if ((st>=0) && (ed>=0)) {
    ptr=registration(ptr,cond,st,ed,&negation, &increment, &first_ptr, F_FIN, parm);
  }
  /*** -------------- ***/

  /*** errors (but just warning) ***/
  if (inacote) {
    MY_STRDUP(strerror,cond+start,end-start);
    add_display(parm,D_WARNING,-1,"a quote never ended in <%s>",strerror);
    free(strerror);
  }
  if (inastring) {
    MY_STRDUP(strerror,cond+start,end-start);
    add_display(parm,D_WARNING,-1,"a double-quote never ended in <%s>",strerror);
    free(strerror);
  }
  if (inaslash) {
    MY_STRDUP(strerror,cond+start,end-start);
    add_display(parm,D_WARNING,-1,"a regex never finish in <%s>",strerror);
    free(strerror);
  }
  if (waiting_for_slash) {
    MY_STRDUP(strerror,cond+start,end-start);
    add_display(parm,D_WARNING,-1,"a regex never start (by a /) in <%s>",strerror);
    free(strerror);
  }
  if (numpar) {
    MY_STRDUP(strerror,cond+start,end-start);
    add_display(parm,D_WARNING,-1,"one or more ) missing in <%s>",strerror);
    free(strerror);
  }
  /*** ------------------------- ***/

  return(first_ptr);
}

/*****
      name : obtain_value
      autor: Bertrand.Wallrich@loria.fr
      parameters :
        pt     : the value to find
	prevop : the previous operator, or F_FIN if no previous op
	nextop : then next operator, or F_FIN if no next operators
	parm   : all params
      description :
         this func take the pt->str or pt->down (in case of parenthesis), and
	 calcul the ptr->value
      return values :
        0 or 1 : depends on tests
	-1 : errors
******/
static int obtain_value (compiled_str *pt, 
					char prevop, char nextop,
					params *parm) {
  int ret_code;
  int rang;
  int len;               

  if (!pt) return(-1);

  if (pt->value) return(pt->test_result);

  
  /*** if pt is a () ***/
  if (pt->down) { 
    ret_code=evaluation_stage2(pt->down,parm);
    if (ret_code==-1) return(-1);

    pt->test_result=pt->down->test_result;

    pt->value=malloc(sizeof(variable_value));
    pt->value->val=NULL;
    duplicate_value(pt->value,pt->down->value);
    free(pt->down->value->val);
    free(pt->down->value);
    pt->down->value=NULL;
    return(pt->test_result);
  }

  /*** ------------- ***/
  /* value already calculate */
  if (!pt->str) {
    return(1);
  }

  /*** if pt is a string with a " ***/
  if ((pt->str[0]=='"') || (pt->str[0]=='\'')) {    /* a real string */
    pt->preset_value=malloc(sizeof(variable_value));
    len=strlen(pt->str);
    if ((pt->str[len-1]=='"') || (pt->str[len-1]=='\'')) {
      MY_STRDUP(((char *)(pt->preset_value->val)),pt->str+1,len-2);
    } else {
      pt->preset_value->val=NULL;
    }
    pt->preset_value->format=F_STRING;
    pt->preset_value->length=-1;
    free(pt->str);
    pt->str=NULL;

    if (pt->value==NULL) {
	 pt->value=malloc(sizeof(variable_value));
	 pt->value->val=NULL;
    }
    duplicate_value(pt->value,pt->preset_value);
    return(1);
  }
  /*** -------------------------- ***/

  /*** if in a regex  ***/
  if ((pt->prev) && ((pt->prev->operator==F_IN) ||(pt->prev->operator==F_NI))) { 
    pt->preset_value=malloc(sizeof(variable_value));
    pt->preset_value->val=(void *)(pt->str);
    pt->preset_value->format=F_STRING;
    pt->preset_value->length=-1;

    free(pt->str);
    pt->str=NULL;

    if (pt->value==NULL) {
	 pt->value=malloc(sizeof(variable_value));
	 pt->value->val=NULL;
    }
    duplicate_value(pt->value,pt->preset_value);
    return(1);
  }
  /*** -------------------------- ***/

  /*** if pt is a variable or a numeric ***/
  pt->preset_value=ascii_to_var(pt->str,strlen(pt->str)); /* is an num or a string */
  
  if (pt->preset_value->format!=F_STRING) { /* a numeric */
    pt->test_result=1;
    
    if (pt->value==NULL) {
	 pt->value=malloc(sizeof(variable_value));
	 pt->value->val=NULL;
    }
    duplicate_value(pt->value,pt->preset_value);

    free(pt->str);
    pt->str=NULL;
    return(1);
  }

  /*** reset pt->value ***/
  free(pt->preset_value->val);
  free(pt->preset_value);

  pt->preset_value=NULL;
  if (pt->value!=NULL) {
    free(pt->value->val);
    free(pt->value);
    pt->value=NULL;
  }
  /*** --------------- ***/

  /* a variable */
  rang=NO_RANG_DEF;
  if ((nextop==F_IN)||(nextop==F_NI))  rang=STRING_FORCE;
  if ((prevop==F_IN)||(prevop==F_NI))  rang=STRING_FORCE;
  if (rang!=STRING_FORCE) {
    len=strlen(pt->str)-1;
    if ((len>1) && (pt->str[len-1]=='.') && (pt->str[len]>='0') &&  (pt->str[len]>='9'))
      rang=pt->str[len]-'0';
  }


  pt->var=get_variable_by_name(pt->str,parm,FOLLOW_LABEL_PATH);

  if (pt->var)
    pt->value=get_variable_value(pt->var,rang);

  pt->test_result=1;
  if (!pt->value) {  /* not a variable set , so = 0 */
    pt->value=malloc(sizeof(variable_value));
    pt->value->length=-1;
    if (rang==STRING_FORCE) {
      pt->value->val=malloc(1);
      ((char *)(pt->value->val))[0]='\0';
      pt->value->format=F_STRING;
    } else {
      pt->value->val=malloc(sizeof(SIG_DEC));
      (*((SIG_DEC *)(pt->value->val)))=0;
      pt->value->format=F_SIGNED;
    }
    pt->test_result=0;
  }
    
  return(pt->test_result);
  /*** -------------------------------- ***/
}


static int evaluation_stage2 (compiled_str *start, params *parm) {  
  compiled_str *pt, *pttempo;
  int rang;
  int len;
  int i=0;
  char order_op[]=ORDER_OPERATORS;
  char ope;
  int ret_code;
  int do_substitute;

  /*** in case of eval(foo) ***/
  if ((start!=NULL) && (start->next==NULL) && (start->operator==F_FIN)) {
    start->test_result=obtain_value(start,F_FIN,start->operator,parm);
    return(start->test_result);
  }
  /*** -------------------- ***/

  ope=order_op[i];
  while(ope!=F_FIN) {  /* for each operator, in right order */
    pt=start;
    while ((pt!=NULL) && (pt->next!=NULL)) {
      if (pt->operator!=ope) {
	   pt=pt->next;
      } else {
	   /*** get the value ***/
	   ret_code=obtain_value(pt,F_FIN,pt->operator,parm);
	   if (ret_code==-1) return(-1);
	   ret_code=obtain_value(pt->next,pt->operator,F_FIN,parm);
	   if (ret_code==-1) return(-1);
	   /*** ------------- ***/
		  
	   /*** doing operation ***/
	   do_substitute=0;
	   switch (ope) {
	   case F_SUP:
	   case F_INF:
	   case F_EQ:
	   case F_NE:
	   case F_SE:
	   case F_IE:
	   case F_IN:
	   case F_NI:
		pt->test_result=compare(pt->value,pt->next->value,ope);
		if (pt->neg)
		  pt->test_result=1-pt->test_result;
		pt->neg=0;
		do_substitute=1;
		break;
	   case F_ER:
		return(-1);
		break;
	   case F_MOD:
	   case F_PLUS:
	   case F_MIN:
	   case F_MUL:
	   case F_DIV:
		ret_code=calculate(pt->value, pt->next->value,ope);
		do_substitute=1;
		pt->test_result=ret_code;
		if (ret_code==-1) return(-1);
		break;
	   case F_AFF:
		if (!pt->next->value) return(-1);
		if (!pt->var) {
		  pt->var=add_new_variable(pt->str,format_to_str(pt->next->value->format),parm);
		}
		if (!pt->var) return(-1);
		/* calc the rang */
		rang=0;
		len=strlen(pt->str)-1;
		if ((len>1) && (pt->str[len-1]=='.') && 
		    (pt->str[len]>='0') &&  (pt->str[len]>='9'))
		  rang=pt->str[len]-'0';
		modify_variable(pt->var,rang,pt->next->value->format,pt->next->value->val,pt->next->value->length);
		pt->value=pt->next->value;
		pt->next->value=NULL;
		pt->test_result=1;
		do_substitute=1;
		break;
	   case F_AND:
		pt->test_result&=pt->next->test_result;
		do_substitute=1;
		break;
	   case F_OR:
		pt->test_result|=pt->next->test_result;
		do_substitute=1;
		break;
	   case F_NOT:
	   }
	   /*** --------------- ***/
	   
	   /*** subtitute the variable by the result ***/
	   if (do_substitute) {
		pttempo=pt->next;
		if (pttempo->value) {
		  free(pttempo->value->val);
		  free(pttempo->value);
		  pttempo->value=NULL;
		}
		pt->operator=pttempo->operator;
		pt->next=pttempo->next;
		if (pt->next) pt->next->prev=pt;
	   } else {  
		pt=pt->next;
	   }
	   /*** ------------------------------------ ***/
	   
	 }
    }
  
    i++;
    ope=order_op[i];
  }
  
 
  /*** doing the last inversion if necessary ***/
  if (start->neg)
    start->test_result=1-start->test_result;
  start->neg=0;
  /*** ------------------------------------- ***/
  return(1);
}

/****
	reinit_for_stage2
	initialise the calul chain (the modify part)
	for used in stage2
****/
int	reinit_for_stage2 (compiled_str *first) {
 compiled_str *ptr;

 ptr=first;
 while(ptr!=NULL) {
   ptr->operator=ptr->preset_operator;
   ptr->neg=ptr->preset_neg;
   ptr->inc=ptr->preset_inc;
   ptr->test_result=ptr->preset_test_result;
   
   ptr->var=ptr->preset_var;
   
   if (ptr->value) {
     free(ptr->value->val);
	free(ptr->value);
	ptr->value=NULL;
   }
   
   if (ptr->preset_value) {
	ptr->value=malloc(sizeof(variable_value));
	ptr->value->val=NULL;
	duplicate_value(ptr->value,ptr->preset_value);
   }
   
   ptr->down=ptr->preset_down;
   ptr->next=ptr->preset_next;
   ptr->prev=ptr->preset_prev;
   
   if (ptr->preset_down) reinit_for_stage2(ptr->preset_down);
   
   ptr=ptr->preset_next;
 }
 return(0);
}



/****
	evaluate a condition.
	return YES or NO
****/
	
int test(char *cond, compiled_str **ptr, params *parm) {  
  int ret;

  if ((*ptr)==NULL) {
    (*ptr)=evaluation_stage1(cond,0,strlen(cond)-1,parm);
    if ((*ptr)==NULL) return(-1);
  }
  
  reinit_for_stage2((*ptr));
 
  ret=evaluation_stage2((*ptr),parm);
  if (ret==-1) return(-1);

  if ((*ptr)->value) {
     free((*ptr)->value->val);
	free((*ptr)->value);
    (*ptr)->value=NULL;
  }
  
  return((*ptr)->test_result);
}

/****
	evaluate an arithmetic or string 
	expression.
	return a variable value, of NULL if pb.
****/
variable_value *eval(char *expr, compiled_str **ptr, params *parm) {
  int ret;

  /* add_display(parm,D_DEBUG,-1,"eval : <%s>",expr); */
  if ((*ptr)==NULL) {
    (*ptr)=evaluation_stage1(expr,0,strlen(expr)-1,parm);
    if ((*ptr)==NULL) return(NULL);
  }
  

  /* (re)initialise the calcul chail */
  reinit_for_stage2((*ptr));

  ret=evaluation_stage2((*ptr),parm);
  if (ret==-1) {
    return(NULL);
  }

  return((*ptr)->value);
}

/****
	evaluate a expression
	and return its value as a int,
	or -1 if pb
	result is a SIG_DEC ptr, already alocated 
****/
int eval_decimal(SIG_DEC *result, char *expr, compiled_str **ptr, params *parm) {
  variable_value *value;

  value=eval(expr,ptr,parm);
  if (value==NULL) return(-1);
  if (value->val==NULL) return(-1);

  if (value->format==F_SIGNED) {
    *result=(*((SIG_DEC *)(value->val)));
    return(1);
  } 
  if (value->format==F_UNSIGNED) {
    *result=(*((SIG_DEC *)(value->val)));  /* dangerous */
    return(1);
  } 

  /* 
	value is never free, because it is keep in mind in
	**ptr->value
   */
  return(-1);
}
