/********
	    Packet filter by wallrich@loria.fr
*********/

#include "nicedump.h"

/********
	    variables globales
	    Car on ne peut pas les passer en parametre a pcap_loop
*********/

confvalue *c_start; /* start of the desc */
confvalue *c_end;

variable *global_variable_start;  /* beginning of the global variable */

stock_chain *stock_start;         /* table of stock */
flux_chain *flux_start;           /* table of flux */

codestruct *codefirst;            /* table of code */
tablestruct *tablefirst;          /* table of tables */

int tab;                          /* num of tabulation, used for the display */
int start_display_packet=YES;     /* case of start_filter and close filter */

/* pcap vars */
pcap_t *pt;                       /* reading struct */
pcap_dumper_t *dump=NULL;         /* dump into a file */
options opt;  /* les options */ 


/******* externs from analyse.c *******/
extern variable *add_new_variable (char *, const char *, params *);
extern int modify_variable (variable *, int, const char, const void *, int);
extern variable *get_variable_by_name(const char *, params *, char);
extern int duplicate_variable(variable *, variable *);
extern table_in *get_label_table(char *, char *, tablestruct *ttfirst, params *);
extern char *search_in_path (char *, char *);
extern char get_format( char *);
extern char *format_to_str( char);
extern char *get_file_to_load(char *);
extern flux_chain *flux_exist (struct flux_struct , params *);
extern flux_chain *add_flux (struct flux_struct , params *);
extern int stock(params *, struct stock_struct, int , int , int);
extern int unstock(params *, struct stock_struct);
extern int get_max_size(params *, protocol *);
extern int modify_display(variable *);

/******* externs from readconf.c *******/

extern int readconf (const char *, confvalue **, codestruct *, tablestruct *);

/******* externs from packet.c *******/

extern confvalue *get_label_ptr(char *, confvalue *);
#ifdef STDC_HEADERS
extern int add_display(params *, const char, int, ...);
#else
extern int add_display(params *, const char, int);
#endif
extern char *display_str(params *, char *);
extern void display_rest(params *, int, int);
extern int get_fixed_num_in_packet(params *, int, int, void *);
extern int get_fixed_char_in_packet(params *, int , int , char **);
extern int get_numbit_to_read(params *, const char *,int , int);

/******* externs from display.c *******/

extern void display(params *, options);
/******* externs from calcul.c *******/

extern variable_value *eval(char *, compiled_str **, params *);
extern int test(char *, compiled_str **, params *);
extern SIG_DEC eval_decimal(SIG_DEC *, char *, compiled_str **, params *);

/****** internal functions ***********/
void cleanup(int );

/********
	    usage function
	    print a help information and exit
*********/
void usage(const char *progname) {
    fprintf(stderr,"\nUsage of %s:\n",progname);
    fprintf(stderr,"\t sniffing options :\n");
    fprintf(stderr,"\t\t [-c <count>]\n");
    fprintf(stderr,"\t\t [-p] not promiscuous\n");
    fprintf(stderr,"\t\t [-i <interface>]\n");
    fprintf(stderr,"\t\t [-s <snaplen>]\n");
    fprintf(stderr,"\t\t [-r <input file>]\n");
    fprintf(stderr,"\t\t [-o <output file>]\n");

    fprintf(stderr,"\t filtering options :\n");
    fprintf(stderr,"\t\t [-pf <pcap filter>]\n");
    fprintf(stderr,"\t\t [-sc <start condition>]\n");
    fprintf(stderr,"\t\t [-st <stop condition>]\n");
    fprintf(stderr,"\t\t [-f <filter>]\n");
    fprintf(stderr,"\t\t [-protocol <label>]\n");

    fprintf(stderr,"\t display options :\n");
    fprintf(stderr,"\t\t [-rest] without the rest of the packet.\n");
    fprintf(stderr,"\t\t [-q ] quiet.\n");
    fprintf(stderr,"\t\t [-2 ] print to stderr (fd=2) (stdout by default).\n");
    fprintf(stderr,"\t\t [-dl <label>]\n");
    fprintf(stderr,"\t\t [-sum] (SUMmary) less verbose\n");

    fprintf(stderr,"\t others options :\n");
    fprintf(stderr,"\t\t [-check <file>]\n");
    fprintf(stderr,"\t\t [-V] print version and exit.\n");

    fprintf(stderr,"\t Combined options :\n");
    fprintf(stderr,"\t\t [-proto <label>] ~ -protocol <label> -dl <label> -sum\n");

    fprintf(stderr,"\nWarning, filter must be between ' and ' to not be interpret by the shell.\n");
   exit(1);
}


/********
	    getopts
	    lit les variables
	  
*********/
void getopts (options *opt, int argc, char **argv) {
  int i;
  int ok;

  /* initialisation */
  opt->r_file=NULL;
  opt->w_file=NULL;
  opt->label=NULL;
  opt->filter=NULL;
  opt->filter_compiled=NULL;
  opt->device=NULL;
  opt->p_filter=NULL;
  opt->disp_label=NULL;
  opt->pflag=YES;  /* in promiscuous */
  opt->quiet=NO;
  opt->count=-1;
  opt->dest=stdout;
  opt->cond_start=NULL;
  opt->cond_start_compiled=NULL;
  opt->cond_start_ok=0;

  opt->cond_stop=NULL;
  opt->cond_stop_compiled=NULL;
  opt->cond_stop_ok=0;

  opt->display_rest=YES;
  opt->sum=NO;
  opt->display_proto=NULL;
  opt->desc_file=NICEDUMP_FIRST_FILE;
  opt->nice_path=getenv("NICEDUMP_PATH");
  if (!opt->nice_path) opt->nice_path=NICEDUMP_PATH;
  opt->snaplen=DEFAULT_SNAPLEN;

  /* reading parms */
  for (i=1;i<argc;i++) {
    ok=0;
    if (!(strcmp(argv[i],"-r"))) { if (i<(argc-1)) { opt->r_file=argv[++i]; ok=1; }} 
    if (!(strcmp(argv[i],"-o"))) { if (i<(argc-1)) { opt->w_file=argv[++i]; ok=1; }} 
    if (!(strcmp(argv[i],"-check"))) { if (i<(argc-1)) { opt->label=argv[++i]; ok=1; }} 
    if (!(strcmp(argv[i],"-dl"))) { if (i<(argc-1)) { opt->disp_label=argv[++i]; ok=1; }} 
    if (!(strcmp(argv[i],"-sc"))) { if (i<(argc-1)) { opt->cond_start=argv[++i]; ok=1; }} 
    if (!(strcmp(argv[i],"-st"))) { if (i<(argc-1)) { opt->cond_stop=argv[++i]; ok=1; }} 
    if (!(strcmp(argv[i],"-f"))) { if (i<(argc-1)) { opt->filter=argv[++i]; ok=1; }} 
    if (!(strcmp(argv[i],"-pf"))) { if (i<(argc-1)) { opt->p_filter=argv[++i]; ok=1; }} 
    if (!(strcmp(argv[i],"-protocol"))) { if (i<(argc-1)) { opt->display_proto=argv[++i]; ok=1; }} 
    if (!(strcmp(argv[i],"-proto"))) { 
      if (i<(argc-1)) { 
	opt->display_proto=argv[++i]; ok=1; 
	opt->disp_label=opt->display_proto;
	opt->sum=YES;
      }
    } 
    if (!(strcmp(argv[i],"-i"))) { if (i<(argc-1)) { opt->device=argv[++i]; ok=1; }} 
    if (!(strcmp(argv[i],"-c"))) { if (i<(argc-1)) { opt->count=atoi(argv[++i]); ok=1; }} 
    if (!(strcmp(argv[i],"-s"))) { if (i<(argc-1)) { opt->snaplen=atoi(argv[++i]); ok=1; }} 
    if (!(strcmp(argv[i],"-p"))) { opt->pflag=NO; ok=1; } 
    if (!(strcmp(argv[i],"-q"))) { opt->quiet=YES; ok=1; } 
    if (!(strcmp(argv[i],"-rest"))) { opt->display_rest=NO; ok=1; } 
    if (!(strcmp(argv[i],"-sum"))) { opt->sum=YES; ok=1; } 
    if (!(strcmp(argv[i],"-2"))) {   opt->dest=stderr; ok=1; } 
    if (!(strcmp(argv[i],"-V"))) { 
      printf("%s version %s\n",argv[0],VERSION);
      printf("NICEDUMP_PATH=%s\n",opt->nice_path);
      exit(0);
    } 
    if (!ok) usage(argv[0]);
  }

  /* unknown args */
  if (i<argc) usage(argv[0]);
  
  /* incompatibles args */
  if (opt->label) {   /* -check and others */
    if (opt->r_file) fprintf(stderr,"options -check and -r are incompatible.\n"); 
    if (opt->w_file) fprintf(stderr,"options -check and -o are incompatible.\n"); 
    if (opt->disp_label)  fprintf(stderr,"options -check and -dl are incompatible.\n"); 
    if (opt->pflag)  fprintf(stderr,"options -check and -p are incompatible.\n"); 
    if (opt->quiet)  fprintf(stderr,"options -check and -q are incompatible.\n"); 
    if (opt->cond_start)  fprintf(stderr,"options -check and -sc are incompatible.\n"); 
    if (opt->cond_stop)  fprintf(stderr,"options -check and -st are incompatible.\n"); 
    if (opt->filter)  fprintf(stderr,"options -check and -f are incompatible.\n"); 
    if (opt->p_filter)  fprintf(stderr,"options -check and -pf are incompatible.\n"); 
    if (opt->device)  fprintf(stderr,"options -check and -i are incompatible.\n"); 
    if (opt->count>0)  fprintf(stderr,"options -check and -c are incompatible.\n"); 
    if (opt->snaplen!=DEFAULT_SNAPLEN)  fprintf(stderr,"options -check and -s are incompatible.\n"); 
  }

}

/**************************************************************************/
/**************************************************************************/

/*******
	block
	block gestion.
	valid : if the test was valid or not
	return gotonext
*******/ 
int block_gestion(params *parm, protocol *prot, int valid) {

  /*** a } ***/
  if (prot->pt->in.block.type_of_block==END_BLOCK) {
    /* add_display(parm,D_DEBUG,-1,"A end block replay=%d, index=%d\n",prot->replay_block[prot->block_index],prot->block_index); */
    if (prot->replay_block[prot->block_index]==1) {    /* in case of a while */
      prot->pt=prot->pt->in.block.corresp;
      prot->block_index--;
      /* add_display(parm,D_DEBUG,-1,"redo the block\n"); */
      return(0);
    } else {                                         /* not a while */
      prot->block_index--;
      return(1);
    }
  }
  /*** --- ***/

  /*** a { ***/
  prot->block_index++;
  if (prot->pt->in.block.type_of_block==START_BLOCK_WHILE) 
    prot->replay_block[prot->block_index]=1;
  else
    prot->replay_block[prot->block_index]=0;

  /* add_display(parm,D_DEBUG,-1,"A start block replay=%d, index=%d\n",prot->replay_block[prot->block_index],prot->block_index); */
  /*** --- ***/
  
  /*** in a start block of a not valid condition or while ***/ 
  if (!valid) {
    prot->pt=prot->pt->in.block.corresp;
    prot->block_index--;
    return(1);
  }
  /*** -------------------------------------------------- ***/
  return(1);
}

/*******
	label
	label gestion.
	return 1 if display, otherwise return 0;
*******/ 
int label_gestion (params *parm, protocol *prot) {
  int len;    
  variable *var;

  /**** add to the label path ****/
  if (prot->add_to_label_path) {
    len=strlen(prot->pt->in.label.label_name);
    prot->label_path[prot->label_path_length++]=':';
    memcpy(prot->label_path+prot->label_path_length,prot->pt->in.label.label_name,len);
    prot->label_path_length=prot->label_path_length+len;
    prot->label_path[prot->label_path_length+1]='\0';
    
    /**** add variable _current_ ****/
    prot->nbbitsread=0;
    var=add_new_variable("_current_","%d",parm);
    modify_variable (var,0,F_SIGNED,(void *)&(prot->nbbitsread),0);  /* to initialise to 0 */
    prot->nb_bits_read=(SIG_DEC *)(var->values[0].val);
    /**** ---------------------- ****/

    /**** add the name of the protocol ***/
    if (prot->label_name==NULL) {
      MY_STRDUP(prot->label_name,
	     prot->pt->in.label.label_name,
	     len);
    }
    if ((prot->label_str==NULL) && (prot->pt->in.label.label_str!=NULL)) {
      MY_STRDUP(prot->label_str,
		prot->pt->in.label.label_str,
		strlen(prot->pt->in.label.label_str));    
    }
    /**** ---------------------------- ***/
  }
  prot->add_to_label_path=1;  /* by default, add to label_path */
  /**** --------------------- ****/


  /**** display the label ***/
  if (prot->pt->in.label.label_str!=NULL) {
    add_display(parm,D_LABEL,prot->indent,"%s",prot->pt->in.label.label_str);
    return(1);
  }
  /**** ----------------- ***/


  return(0);
}


/*******
	call (or loopfor)
	call gestion.
	return end_of_proto;
*******/ 
int call_gestion (params *parm, protocol *prot, int to_loop_next, 
			   char *return_variable_name, 
			   char *return_variable_display,
			   char return_variable_options) {
  char *label=NULL;          /* the label name */

  SIG_DEC user_length_in_bits=-1;    /* the number of the bits in this protocol, ask by user  -1 if not def) */
  int max_length_in_bits;    
  int resp;

  confvalue *start_point_next=NULL;  /* start point of the called protocol  */
  table_in *tin=NULL;
  char *file_to_search;
  char *f_to_read;
  char *to_search=NULL;              /* a ptr to to_search */
  confvalue **quickjump=NULL;        /* a ptr to quickjump */
  int way_to_exit=NOT_END;           /* if a func, return an error. 
								if a call or loopfor, return not_end,
							     to continue the protocol desc. */

  /**** get the label name or the ptr to the next protocol ****/
  switch (prot->pt->in.call.typecall) {
  case CALL_LABEL:
    to_search=&prot->pt->in.call.to_search;
    quickjump=&prot->pt->in.call.quickjump;
    label=prot->pt->in.call.in.label_name;
    break;
  case CALL_TABLE:
    tin=get_label_table(prot->pt->in.call.in.table_name,
			prot->pt->in.call.variable_name,
			tablefirst,
			parm);
    if (tin==NULL) return(1);
    to_search=&tin->to_search;
    quickjump=&tin->quickjump;
    label=tin->label;
    break;
  }
  /**** -------------------------------------------------- ****/

  /*** usefull to search ? ***/
  if (*to_search==NO) {
    add_display(parm,DE_PROTO,prot->indent,"label %s not found.",label);
    return(END_DUE_TO_ERROR);
  }
  /*** ------------------- ***/

  /**** the max num of bits to add to the call ****/
  if (prot->user_length_in_bits>-1)
    max_length_in_bits=prot->user_length_in_bits-(*(prot->nb_bits_read));
  else
    max_length_in_bits=prot->max_length_in_bits-(*(prot->nb_bits_read));
  /**** -------------------------------------- ****/

  /**** the user num of bits to add to the call ****/
  switch (prot->pt->in.call.type) {
  case CALL_NONE:       /* no n um of bits specifieds. calculate it */
    user_length_in_bits=-1;
    break;
  case CALL_FIXED:
    user_length_in_bits=prot->pt->in.call.bits;
    break;
  case CALL_EVAL:
    resp=eval_decimal(&user_length_in_bits,prot->pt->in.call.eval,
				  &prot->pt->in.call.eval_compiled,parm);
    if (resp==-1) return(END_DUE_TO_ERROR);
    break;
  }
  /**** ---------------------------------------- ****/

  /****  whe dont call a protocol with zero bits in it ****/
  if (return_variable_name) way_to_exit=END_DUE_TO_TOO_SMALL;
  if (user_length_in_bits==0) return(way_to_exit);
  if ((user_length_in_bits==-1) && (max_length_in_bits==0)) return(way_to_exit);
  /**** ---------------------------------------------- ****/
  

  /**** looking for the file  ****/
  if ((*quickjump)!=NULL)
    start_point_next=*quickjump;
  else
    start_point_next=get_label_ptr(label,c_start);

  if (start_point_next==NULL) {
    file_to_search=get_file_to_load(label);
    
    /* looking for the label in a new file */
    if ((f_to_read=search_in_path(opt.nice_path,file_to_search))==NULL) {
      add_display(parm,DE_PROTO,prot->indent,"no file %s in %s.",file_to_search,opt.nice_path);
      free(file_to_search);
      *to_search=NO;
      return(END_DUE_TO_ERROR);
    } 
    start_point_next=c_end;
    readconf(f_to_read,&c_end,codefirst,tablefirst);
    free(file_to_search);
    free(f_to_read);
    start_point_next=get_label_ptr(label,start_point_next);
  }

  if (start_point_next==NULL) {
    add_display(parm,DE_PROTO,prot->indent,"no label %s.",label);
    *to_search=NO;
    return(END_DUE_TO_ERROR);
  }
  /**** ------------------------ ****/


  /**** a new protocol ***/
  *quickjump=start_point_next;
  prot->pt=prot->pt->next;                    /* skip the call itself */
  parm->prot_index++;                         /* a new proto */

  parm->prot[parm->prot_index].label_path_length=prot->label_path_length;
  parm->prot[parm->prot_index].add_to_label_path=1;
  parm->prot[parm->prot_index].label_name=NULL;
  parm->prot[parm->prot_index].label_str=NULL;
  MEMCPY_STR(parm->prot[parm->prot_index].label_path,prot->label_path,prot->label_path_length);
  parm->prot[parm->prot_index].offset=(*(prot->nb_bits_read))+prot->offset;
  parm->prot[parm->prot_index].indent=prot->indent;
  parm->prot[parm->prot_index].to_loop=to_loop_next;
  parm->prot[parm->prot_index].user_length_in_bits=user_length_in_bits;
  parm->prot[parm->prot_index].max_length_in_bits=max_length_in_bits;
  parm->prot[parm->prot_index].pt=start_point_next;
  parm->prot[parm->prot_index].start_point=start_point_next;
  parm->prot[parm->prot_index].block_index=0;
  parm->prot[parm->prot_index].gotonext=0;
  parm->prot[parm->prot_index].to_do_or=0;
  parm->prot[parm->prot_index].end_proto=NOT_END;
  parm->prot[parm->prot_index].nb_bits_read=NULL;
  parm->prot[parm->prot_index].flux=NULL;
  parm->prot[parm->prot_index].prev_flux=prot->prev_flux;
  parm->prot[parm->prot_index].return_variable_name=return_variable_name;
  parm->prot[parm->prot_index].return_variable_display=return_variable_display;
  parm->prot[parm->prot_index].return_variable_options=return_variable_options;

  /**** ------------- ****/
  /* add_display(parm,D_DEBUG,-1,"Call %s, max=%d/ps=%d poff=%d off=%d user_length=%d",label,max_length_in_bits,parm->packet_size,parm->prot[parm->prot_index].offset,parm->offoct*8,user_length_in_bits);  */


  return(NOT_END);
}


/*******
	goto
	goto gestion.
	return end_of_proto;
 *******/ 
int goto_gestion (params *parm, protocol *prot) {
  char *label=NULL;             /* the label name                      */
  confvalue *start_point_next;  /* start point of the called protocol  */
  table_in *tin=NULL;
  char *file_to_search;
  char *f_to_read;
  char *to_search=NULL;              /* a ptr to to_search */
  confvalue **quickjump=NULL;        /* a ptr to quickjump */

  /**** get the label name or the ptr to the next protocol ****/
  switch (prot->pt->in.go_to.typegoto) {
  case CALL_LABEL:
    to_search=&prot->pt->in.go_to.to_search;
    quickjump=&prot->pt->in.go_to.quickjump;
    label=prot->pt->in.go_to.in.label_name;
    break;
  case CALL_TABLE:
    tin=get_label_table(prot->pt->in.go_to.in.table_name,
			prot->pt->in.go_to.variable_name,
			tablefirst,
			parm);
    if (tin==NULL) return(1);
    to_search=&tin->to_search;
    quickjump=&tin->quickjump;
    label=tin->label;
    break;
  }
  /**** -------------------------------------------------- ****/

  /*** usefull to search ? ***/
  if (*to_search==NO) {
    add_display(parm,DE_PROTO,prot->indent,"label %s not found.",label);
    return(END_DUE_TO_ERROR);
  }
  /*** ------------------- ***/

		  
  /**** looking for the file  ****/
  if ((*quickjump)!=NULL)
    start_point_next=*quickjump;
  else
    start_point_next=get_label_ptr(label,c_start);

  if (start_point_next==NULL) {

    file_to_search=get_file_to_load(label);
    
    /* looking for the label in a new file */
    if ((f_to_read=search_in_path(opt.nice_path,file_to_search))==NULL) {
      add_display(parm,DE_PROTO,prot->indent,"no file %s in %s.",file_to_search,opt.nice_path);
      free(file_to_search);
      *to_search=NO;
      return(END_DUE_TO_ERROR);
    } 
    start_point_next=c_end;
    readconf(f_to_read,&c_end,codefirst,tablefirst);
    free(file_to_search);
    free(f_to_read);
    start_point_next=get_label_ptr(label,start_point_next);
  }

  if (start_point_next==NULL) {
    add_display(parm,DE_PROTO,prot->indent,"no label %s.",label);
    *to_search=NO;
    return(END_DUE_TO_ERROR);
  }
  /**** ---------------------------------- ****/

  /**** go to the new label ***/
  *quickjump=start_point_next;
  prot->pt=start_point_next;
  prot->gotonext=0;
  prot->add_to_label_path=0;   /* in a goto, the label path is not modified */
  /**** ------------------ ****/


  return(NOT_END);
}


/*******
	fmt
	format_gestion
	return end_of_proto;
*******/ 
int format_gestion (params *parm, protocol *prot) {
  
  int soffbit=0, soffoct=0;                      /* to save offset */
  int localindent;                               /* indentation */
  int i, j;
  int max_len;
  int rang;
  SIG_DEC numbit_to_read=-1;                     /* num of bits to read in the packet */
  int ret_value=0;                               /* to test the return value of read in the packet */
  int resp;
  variable *var=NULL;

  int fluxoffbit, fluxoffoct;                    /* to save offset in case of flux */
  variable *fluxvar;

  int bit_read=0;                                /* num of bits read for this variable */

  char f_of_var;                                 /* format of the variable */

  SIG_DEC svalue;                                /* all the values */
  UNSIG_DEC uvalue;
  char *strvalue;
		  

  /**** In case of flux ****/         /* use in case of variable with rang>1 , and flux stops in the n rang,
	                                 we will reread the entire variable since the beggining */
  fluxoffoct=parm->offoct;
  fluxoffbit=parm->offbit;
  fluxvar=parm->vend;

  /**** --------------- ****/



  /**** the indentation ****/
  localindent=prot->indent;
  if (prot->pt->in.fmt.indent) localindent++;
  /**** --------------- ****/

  /* foreach values*/
  for (rang=0 ; rang<prot->pt->in.fmt.num ; rang++) {

    /*** testing the format ***/
    f_of_var=get_format(prot->pt->in.fmt.format[rang]);
    if (f_of_var==F_ERROR) {
      add_display(parm,DE_PROTO,localindent,"unknown format \"%s\" (variable %s.%d) file %s, line %d.",prot->pt->in.fmt.format[rang],prot->pt->in.fmt.name,rang,prot->pt->file, prot->pt->num_line);
      return(END_DUE_TO_ERROR);
    }
    /*** ------------------ ***/

    /*** determining the max bits to read ***/
    max_len=get_max_size(parm,prot);
    /*** -------------------------------- ***/


    /*** determining the num of bits to read ***/
    switch(prot->pt->in.fmt.type[rang]) {
    case FMT_EVAL:
      numbit_to_read=-1;
      resp=eval_decimal(&numbit_to_read,prot->pt->in.fmt.eval[rang],
				    &prot->pt->in.fmt.eval_compiled[rang],parm);
      break;
    case FMT_FIXED:
      numbit_to_read=prot->pt->in.fmt.length[rang];
      if (numbit_to_read==0) numbit_to_read=max_len;
      break;
    case FMT_UNTIL:
      numbit_to_read=get_numbit_to_read(parm,prot->pt->in.fmt.until[rang],prot->pt->in.fmt.length[rang], max_len);
      break;
    }
    /*** ------------------------------------ ***/

    /*       add_display(parm,D_DEBUG,-1,"%s : to_read=%d max_length=%d, user_length=%d, nb_read = %d , max_len=%d, ps=%d, off=%d, poff=%d",
		  prot->pt->in.fmt.name,
		  numbit_to_read,
		  prot->max_length_in_bits,
		  prot->user_length_in_bits,
		  (*(prot->nb_bits_read)),
		  max_len,
		  parm->packet_size,
		  parm->offoct*8,
		  prot->offset);
    */

    /*** unable to read from the packet ***/
    if ((numbit_to_read==-1) || (numbit_to_read>max_len)) {
      if (parm->flux) {

	   /* add_display(parm,D_DEBUG,-1,"%s numbit_to_read=%d, max_len=%d",prot->pt->in.fmt.name,numbit_to_read,max_len); */


	/*** errors ***/
	if (parm->prot_index<parm->flux->prots_index) {
	  add_display(parm,DE_PROTO,-1,"starting record fo a flux continue in a lower protocol !.");
	  return(END_DUE_TO_ERROR);
	}
	
	if (fluxoffbit) {
	  add_display(parm,DE_CALC,prot->indent,"unable to memorise  packet, i'm not in the beggining of a octet\n");
	  return(END_DUE_TO_ERROR);
	}
	/*** ------ ***/

	parm->free_flux=NO;   /* don't free the flux at the end of the protocol */
	parm->flux->first_time=NO;
	prot->gotonext=NO;
	
	/*** ending all protocols ***/
	/*
	for (i=parm->prot_index-1 ; i>=parm->flux->prots_index ; i--) {
	  add_display(parm,D_DEBUG,-1,"%d, adding %d",i,(*(parm->prot[i+1].nb_bits_read)));
	  (*(parm->prot[i].nb_bits_read))+=(*(parm->prot[i+1].nb_bits_read));
	  add_display(parm,D_DEBUG,-1,"%d, %d/%d",i,(*(parm->prot[i].nb_bits_read)),parm->prot[i].max_length_in_bits);
	}
	*/
	/*** -------------------- ***/

	/*** copying protocols stack ***/
	for (i=parm->flux->prots_index ; i<=parm->prot_index ; i++) {
	  /* add_display(parm,D_DEBUG,-1,"prot = %d numbit_to_read=%d, max_len=%d",prot,numbit_to_read,max_len); */
	  /*** copying protocol ***/
	  memcpy(&(parm->flux->prots[i-parm->flux->prots_index]),
		 &(parm->prot[i]),
		 sizeof(protocol));

	  parm->flux->prots[i-parm->flux->prots_index].nbbitsread=(*(parm->prot[i].nb_bits_read));
	  /*** ---------------- ***/

	  /*** copy label_name and label_str ***/
	  MY_STRDUP(parm->flux->prots[i-parm->flux->prots_index].label_name,
		    parm->prot[i].label_name,
		    strlen(parm->prot[i].label_name));

	  if (parm->prot[i].label_str) {
	    MY_STRDUP(parm->flux->prots[i-parm->flux->prots_index].label_str,
		      parm->prot[i].label_str,
		      strlen(parm->prot[i].label_str));
	  } else parm->flux->prots[i-parm->flux->prots_index].label_str=NULL;
	  /*** ----------------------------- ***/

	  
	  /*** copying blocks ***/
	  parm->flux->prots[i-parm->flux->prots_index].block_index=parm->prot[i].block_index;
	  for (j=0 ; j<=parm->prot[i].block_index ; j++) {
	    parm->flux->prots[i-parm->flux->prots_index].replay_block[j]=parm->prot[i].replay_block[j];
	  }
	  /*** -------------- ***/

	}
	/*** (end) copying protocols stack ***/

	parm->flux->prots_index_length=parm->prot_index-parm->flux->prots_index;
	
	/* the first time, loc_var_flux means the last read value before flux.
	   after, it means the first read value after flux */
	if ((parm->flux->loc_var_flux) && (parm->flux->first_time))
	  parm->flux->loc_var_flux=parm->flux->loc_var_flux->next;
		  
	/*** copying rest of the packet ***/
	parm->flux->length=parm->packet_size/8-fluxoffoct;
	parm->flux->p=malloc(parm->flux->length);
	memcpy(parm->flux->p,parm->p+fluxoffoct,parm->flux->length);
	add_display(parm,D_INFO,-1,"Stock %d octets for the next packet",parm->flux->length);
	/*** -------------------------- ***/


	/*** back to the label containning the flux ***/
	parm->prot_index=parm->flux->prots_index;
	/*** -------------------------------------- ***/


	/*** ------------------------------- ***/
      } else {   
	if (prot->user_length_in_bits==-1)
	  add_display(parm,DE_PROTO,-1,"Packet too small. Please use a bigger snaplen (-s option).");
	else
	  add_display(parm,DE_PROTO,-1,"Packet too small. a flux is missing ?.");
      }
      return(END_DUE_TO_TOO_SMALL);
    }
    /*** ------------------------------ ***/

    /***          Reading         ***/
    soffoct=parm->offoct;
    soffbit=parm->offbit;

    if (f_of_var==F_SIGNED)
      ret_value=get_fixed_num_in_packet(parm,numbit_to_read,max_len,(void *)&svalue);
    if (f_of_var==F_UNSIGNED)
      ret_value=get_fixed_num_in_packet(parm,numbit_to_read,max_len,(void *)&uvalue);
    if (f_of_var==F_STRING)
      ret_value=get_fixed_char_in_packet(parm,numbit_to_read,max_len,(void *)&strvalue);
    if (f_of_var==F_BINAR)
      ret_value=get_fixed_char_in_packet(parm,numbit_to_read,max_len,(void *)&strvalue);
    if (ret_value==0) {

      bit_read+=numbit_to_read;

      var=add_new_variable(prot->pt->in.fmt.name,prot->pt->in.fmt.display,parm);
      if (f_of_var==F_SIGNED)
	modify_variable (var,rang,F_SIGNED,(void *)&svalue,0);
      if (f_of_var==F_UNSIGNED)
	modify_variable (var,rang,F_UNSIGNED,(void *)&uvalue,0);
      if (f_of_var==F_STRING) {
	   strvalue[numbit_to_read/8]='\0';
	   modify_variable (var,rang,F_STRING,(void *)strvalue,0);
	   free(strvalue);
      }
      if (f_of_var==F_BINAR)
	modify_variable (var,rang,F_BINAR,(void *)strvalue,numbit_to_read/8);
    } else {
      /* unable to read the value, not enough in packet */
      parm->offoct=soffoct;
      parm->offbit=soffbit;
      return(END_DUE_TO_TOO_SMALL);
    }  
  }  
    /***          ---------         ***/
  if ((var->printvalue[0]!='\0') && 
      (!(prot->pt->in.fmt.options & FMT_BLIND)))
      add_display(parm,D_VALUE,localindent,"%-10s : %s",prot->pt->in.fmt.name,var->printvalue);  

  if (prot->pt->in.fmt.nr==1){ /* packet will be reread */
    parm->offbit=soffbit;
    parm->offoct=soffoct;
  } else {
    /* update of  current */
    /* (*(prot->nb_bits_read))=parm->offoct*8+parm->offbit-prot->offset;  */
    (*(prot->nb_bits_read))+=bit_read;
  }

  return(NOT_END);
}

/*******
	display
	display_gestion
	return 0;
*******/ 
int display_gestion (params *parm, protocol *prot) {
  char *ds;          /* the display string modified, used in TYPE_DISPLAY */

  ds=display_str(parm, prot->pt->in.display.display_str);
  add_display(parm,D_SUFFIX,prot->indent," %s",ds);
  free(ds);
  return(0);
}

/*******
	stock
	stock_gestion
	return end_of_proto;
*******/ 
int stock_gestion (params *parm, protocol *prot) {
  SIG_DEC  offset=-2, total_len=-2, nbbits;
  int resp;

  /*** not at a start of an octet ***/
  if (parm->offbit) {
    add_display(parm,DE_CALC,prot->indent,"unable to stock packet, i'm not in the beggining of a octet\n");
    return(END_DUE_TO_ERROR);
  }
  /*** -------------------------- ***/
  
  /*** calculate the offset ***/
  switch (prot->pt->in.stock.type) {
  case STOCK_AT_THE_END:
    offset=-1;
    break;
  case STOCK_NUM:
    offset=prot->pt->in.stock.length;
    break;
  case STOCK_EVAL:
    offset=-2;
    resp=eval_decimal(&offset,prot->pt->in.stock.eval,
				  &prot->pt->in.stock.eval_compiled,parm);
    break;
  }
  /*** -------------------- ***/

  
  /*** calculate the total length of the packet  ***/
  switch (prot->pt->in.stock.type_totl) {
  case STOCK_AT_THE_END:  /* equal to undef */
    total_len=-1;
    break;
  case STOCK_NUM:
    total_len=prot->pt->in.stock.totl;
    break;
  case STOCK_EVAL:
    total_len=-2;
    resp=eval_decimal(&total_len,prot->pt->in.stock.eval_totl,
				  &prot->pt->in.stock.eval_totl_compiled,parm);
    break;
  }
  /*** ---------------------------------------  ***/
  
  
  if ((offset>-2)&&(total_len>-2)) {
    if (prot->user_length_in_bits>-1) 
      nbbits=(prot->user_length_in_bits-(*(prot->nb_bits_read)))/8;
    else
      nbbits=(prot->max_length_in_bits-(*(prot->nb_bits_read)))/8;
     
    stock(parm,prot->pt->in.stock,offset,nbbits,total_len);
  }
  else add_display(parm,DE_CALC,prot->indent,
		   "unable to calculate the offset for stock %d\n",offset);
  
  return(NOT_END);
}

/*******
	unstock
	unstock_gestion
	return end_of_proto;
*******/ 
int unstock_gestion (params *parm, protocol *prot) {
  int i;

  if (parm->offbit) {
    add_display(parm,DE_CALC,prot->indent,"unable to unstock packet, i'm not in the beggining of a octet\n");
    return(END_DUE_TO_ERROR);
  }

  i=unstock(parm,prot->pt->in.stock);
  if (i==-1) 
    return(END_DUE_TO_ERROR);
  else 
    prot->user_length_in_bits+=i;

  return(NOT_END);
}


/*******
	insert_flux
	return end_of_proto;
*******/ 
int insert_flux (params *parm, flux_chain *fl) {
  int i, j ;
  char *p;

  int offset;                        /* saving value of the */
  SIG_DEC nbitread;                  /* current proto       */
  

  offset=parm->prot[parm->prot_index].offset;
  nbitread=*(parm->prot[parm->prot_index].nb_bits_read);

  /*** adding protocol stack ***/ 
  for (i=0 ; i<= fl->prots_index_length ; i++) {

    /*** freing label_name and label_str ***/
    if (parm->prot[parm->prot_index+i].label_name)
      free(parm->prot[parm->prot_index+i].label_name);
    if (parm->prot[parm->prot_index+i].label_str)
      free(parm->prot[parm->prot_index+i].label_str);
    /*** ------------------------------- ***/

    /*** copying protocol ***/
    memcpy(&(parm->prot[parm->prot_index+i]),
	   &(fl->prots[i]),
	   sizeof(protocol));
    /*** ---------------- ***/
    

    /*** copy label_name and label_str ***/
    MY_STRDUP(parm->prot[parm->prot_index+i].label_name,
	      fl->prots[i].label_name,
	      strlen(fl->prots[i].label_name));
    if (fl->prots[i].label_str) {
      MY_STRDUP(parm->prot[parm->prot_index+i].label_str,
		fl->prots[i].label_str,
		strlen(fl->prots[i].label_str));
    /*** copy label_name and label_str ***/
      
      /*** display the label if usefull ***/
      /* i=0 is already displayed , so skip it */
      if (i>0)
	add_display(parm,D_LABEL,parm->prot[parm->prot_index+i].indent,"%s",parm->prot[parm->prot_index+i].label_str);
    } else parm->prot[parm->prot_index+i].label_str=NULL;
    /*** ----------------------------- ***/
    
    /*** copying protocols ***/
    for (j=0 ; j<=fl->prots[i].block_index ; j++)
      parm->prot[parm->prot_index+i].replay_block[j]=fl->prots[i].replay_block[j];      
    /*** ----------------- ***/

  }
  /*** (end) adding protocol stack ***/ 
  
  /*** insterting all variables ***/
  if (fl->loc_var_flux) {
    parm->vend->next=fl->loc_var_flux;
    fl->loc_var_flux->prev=parm->vend;
    while(parm->vend->next) {
      parm->vend=parm->vend->next;
    }
  }
  /*** ------------------------ ***/

      
  /*** re-calculating protocols offset  ***/
  parm->prot[parm->prot_index].offset=offset;
  (*(parm->prot[parm->prot_index].nb_bits_read))=nbitread;

  for (i=1 ; i<= fl->prots_index_length ; i++) {
    parm->prot[parm->prot_index+i].offset=
      (parm->offoct*8)-(*(parm->prot[parm->prot_index+i].nb_bits_read));
  }
  /*** (end) re-calculating protocols offset  ***/

  /*** re-calculating protocols size  ***/
  /*
  add_display(parm,D_DEBUG,-1,"before (%d) : max_length=%d, nb_read=%d = %d, ps=%d, off=%d poff=%d",
	      parm->prot_index,
	      parm->prot[parm->prot_index].max_length_in_bits,
	      parm->prot[parm->prot_index].nb_bits_read,
	      (*(parm->prot[parm->prot_index].nb_bits_read)),
	      parm->packet_size,
	      parm->offoct*8,
	      parm->prot[parm->prot_index].offset);
  */
  parm->prot[parm->prot_index].max_length_in_bits+=fl->length*8;
  /*
  for (i=1 ; i<= fl->prots_index_length ; i++) {
    if (parm->prot[parm->prot_index+i-1].user_length_in_bits>-1)
      parm->prot[parm->prot_index+i].max_length_in_bits=
	parm->prot[parm->prot_index+i-1].user_length_in_bits
	-(*(parm->prot[parm->prot_index+i-1].nb_bits_read))
	+parm->offoct*8-parm->prot[parm->prot_index+i].offset;
    else
      parm->prot[parm->prot_index+i].max_length_in_bits=
	parm->prot[parm->prot_index+i-1].max_length_in_bits
	-(*(parm->prot[parm->prot_index+i-1].nb_bits_read))
	+parm->offoct*8-parm->prot[parm->prot_index+i].offset;
  }
  */

  for (i=1 ; i<= fl->prots_index_length ; i++) {
    if (parm->prot[parm->prot_index+i-1].user_length_in_bits>-1)
      parm->prot[parm->prot_index+i].max_length_in_bits=
	parm->prot[parm->prot_index+i-1].user_length_in_bits
	-(*(parm->prot[parm->prot_index+i-1].nb_bits_read));
    else
      parm->prot[parm->prot_index+i].max_length_in_bits=
	parm->prot[parm->prot_index+i-1].max_length_in_bits
	-(*(parm->prot[parm->prot_index+i-1].nb_bits_read));
  }

  /*** (end) re-calculating protocols size  ***/

  /*
  for (i=0 ; i<= fl->prots_index_length ; i++) {
    add_display(parm,D_DEBUG,-1,"%d : max_length=%d, user_length=%d, nb_read=%d = %d, ps=%d, off=%d poff=%d",
		  parm->prot_index+i,
		  parm->prot[parm->prot_index+i].max_length_in_bits,
		  parm->prot[parm->prot_index+i].user_length_in_bits,
		  parm->prot[parm->prot_index+i].nb_bits_read,
		  (*(parm->prot[parm->prot_index+i].nb_bits_read)),
		  parm->packet_size,
		  parm->offoct*8,
		  parm->prot[parm->prot_index+i].offset);
  }
  */

  /*** insterting packet  ***/
  p=malloc(parm->packet_size/8+fl->length);
  memcpy(p,parm->p,parm->offoct);
  memcpy(p+parm->offoct,fl->p,fl->length);
  memcpy(p+parm->offoct+fl->length,parm->p+parm->offoct,parm->packet_size/8-parm->offoct);
  add_display(parm,D_INFO,-1,"Retrieving %d octets from the previous packet",fl->length);

  /*** (end) insterting packet  ***/





  if (parm->free_p) free(parm->p);
  parm->free_p=1;
  parm->p=(u_char *)p;
  parm->packet_size+=fl->length*8;
  free(fl->p);
  fl->p=NULL;
  fl->length=0;
  /*** -----------------  ***/



  /*** jump to the last protocol ***/

  parm->prot_index+=fl->prots_index_length;
  /*** ------------------------- ***/

  return(NOT_END);
}

/*******
	flux
	flux_gestion
	return end_of_proto;
*******/ 
int flux_gestion1 (params *parm, protocol *prot) {
  flux_chain *fl;
  int prot_index;

  prot_index=parm->prot_index;   /* saving the protocol index */

  /*** a flux exist ***/
  if ((fl=flux_exist(prot->pt->in.flux,parm))!=NULL) {

    parm->flux=fl;

    prot->flux=fl;
    prot->prev_flux=parm->prot_index;

    add_display(parm,D_DEBUG,-1,"insert a flux %d",fl);
    if (parm->offbit) {
      add_display(parm,DE_PROTO,-1,"unable to read from a flux, not at the beginning of an octet");
      return(1);
    }

    insert_flux(parm,fl);
  } 
  /*** ------------ ***/


  /*** creating a flux ***/
  else {
    fl=add_flux(prot->pt->in.flux,parm);
    add_display(parm,D_DEBUG,-1,"Add an new flux %d",fl);
    fl->p=NULL;
    fl->length=0;
    fl->prots_index=parm->prot_index;
    fl->prots_index_length=0;
    fl->loc_var_flux=parm->vend;

    prot->flux=fl;
    prot->prev_flux=prot_index;
  }
  /*** --------------- ***/


  /*** (re)initialising a flux ***/
  parm->free_flux=YES;   /* free the flux by default */  
  fl->first_time=YES;
  parm->flux=fl;
  /*** ----------------------- ***/


  return(NOT_END);
}



/*******
	free_flux
	free the flux. We are at the normal end of the proto
*******/ 
int free_flux (params *parm, protocol *prot) {
  flux_chain *fl, *fl1;
  int i;

  /* add_display(parm,D_DEBUG,-1,"free the flux %d",parm->flux); */
  /*** looking for the flux ***/
  fl=(*(parm->flux_start));
  if (fl==parm->flux) {
    *(parm->flux_start)=fl->next;
    fl1=fl;
  } else {
    while(fl->next!=parm->flux)
      fl=fl->next;

    fl1=fl->next;
    fl->next=fl1->next;
  }
  /*** --------------------- ***/
 

  /*** freeing ***/ 
  free(fl1->p);
  for (i=0 ; i<fl1->num ; i++) {
    free(fl1->key[i]);
  }
  
  for (i=0 ; i< fl1->prots_index_length ; i++) {
    free(fl1->prots[i].label_name);
    fl1->prots[i].label_name=NULL;
    free(fl1->prots[i].label_str);
    fl1->prots[i].label_str=NULL;
  }

  free(fl1);
  parm->flux=NULL;
  /*** ------- ***/  

  return(NOT_END);
}

/*******
	sum
	sum_gestion
	return 0;
*******/ 
int sum_gestion (params *parm, protocol *prot) {

  if (opt.sum==NO) return(NOT_END);

  if (parm->sum) free(parm->sum);
  parm->sum=display_str(parm,prot->pt->in.sum.str);
  return(NOT_END);
}

/*******
	calc
	calc_gestion
	return 0;
*******/ 
int calc_gestion (params *parm, protocol *prot) {
  variable_value *value;
  
  value=eval(prot->pt->in.calc.str,&prot->pt->in.calc.str_compiled,parm); 
  return(NOT_END);  

  /*
    never free the value, pointer are keep in mind
    in prot->pt->in.calc.str_compiled
  */
}

/*******
	return the value of the variable...
	return end_proto;
*******/ 
int return_value_gestion (params *parm, protocol *prot) {
  variable *var_to_modify=NULL;
  variable *val=NULL;

  /*** geting the variable ***/
  val=get_variable_by_name(prot->pt->in.ret.return_var, parm, STRICT_LABEL_PATH);
  if (val==NULL) return (END_DUE_TO_ERROR);
  /*** ------------------- ***/

  /*** adding the new variable ***/
  /* warning, modificaztion of parm->prot_index, to have the right
	protocol */
  parm->prot_index--;
  if (parm->prot_index<0) {
    parm->prot_index++;
    add_display(parm,DE_PROTO,-1,"protocol 0 cannot be a func !!");  
    return (END_DUE_TO_ERROR);    
  }
  var_to_modify=add_new_variable(prot->return_variable_name,FORMAT_D,parm);
  duplicate_variable(var_to_modify,val);
  parm->prot_index++;
  /*** (end) adding the new variable ***/

  /* add_display(parm,D_DEBUG,-1,"modifie %s -> %s\n",prot->pt->in.ret.return_var,
	prot->return_variable_name);*/

  /*** recalculate display fmt as a new variable ***/
  if (prot->return_variable_display) {
    memcpy(var_to_modify->dspfmt,prot->return_variable_display,strlen(prot->return_variable_display));
    modify_display(var_to_modify);
  }
  /*** (end) recalculate display fmt as a new variable ***/
  
  /*** display the value ***/
  if ((var_to_modify->printvalue[0]!='\0') && 
      (!(prot->return_variable_options & FMT_BLIND)))
      add_display(parm,D_VALUE,parm->prot[parm->prot_index-1].indent,"%-10s : %s",prot->return_variable_name,(var_to_modify->printvalue));  
  /*** ----------------- ***/


  return(END_DUE_TO_RETURN);  
}


/**************************************************************************/
/**************************************************************************/

/********
	 
 ********/
int decode_packet(params *parm) {
  variable *var;
  protocol *prot;
  int nbbits;    /* calculate the num of bits in the rest */

  /**** Adding the first ":_current_" variable ****/
  prot=&(parm->prot[0]);
  prot->nbbitsread=0;
  var=add_new_variable("_current_","%d",parm);
  modify_variable (var,0,F_SIGNED,(void *)&(prot->nbbitsread),0);  /* to initialise to 0 */
  parm->prot[parm->prot_index].nb_bits_read=(SIG_DEC *)(var->values[0].val);
  parm->prot[parm->prot_index].nb_bits_read=&prot->nbbitsread;
/**** ------------------------------- ****/


  /* for each protocol */
  while(parm->prot_index>=0) {
    
    prot=&(parm->prot[parm->prot_index]);

    while(prot->end_proto==NOT_END) {   /* moving in the confvalue struct */
      int valid_condition=1;             /* boolean to the condition */
      

      /**** is a or ? ***/
      if ((!(prot->to_do_or)) && (prot->pt->or)) valid_condition=0;
      if (!prot->pt->or) prot->to_do_or=0;
      /**** --------- ***/

      /**** the indentation ****/
      if (prot->pt->indent>0) prot->indent++;
      
      /**** testing the condition ****/
      if ((prot->pt->condition!=NULL) && (valid_condition==1)) {	  
	valid_condition=test(prot->pt->condition,
					 &prot->pt->condition_compiled,parm);
	/* add_display(parm,D_DEBUG,-1,"testing <%s> return %d et %d",prot->pt->condition,valid_condition,prot->pt->condition_compiled);   */
      }
      /**** --------------------- ****/
      

      /**** evaluation ****/
	 /* add_display(parm,D_DEBUG,-1,"type = %d (%s,%d)",prot->pt->type, prot->pt->file, prot->pt->num_line);  */
      if (valid_condition>0) {
	switch (prot->pt->type) {
	case TYPE_NOTDEF:          /* uhhh ?? */
	  break;
	case TYPE_BLOCK:           /* a block { or } */
	  prot->gotonext=block_gestion(parm,prot,1);
	  break;
	case TYPE_LABEL:
	  /* in case of loopfor, not redo the label itself */
	  prot->start_point=prot->pt->next;  
	  label_gestion(parm,prot);
	  break;
	case TYPE_GOTO:
	  prot->end_proto=goto_gestion(parm,prot);
	  break;
	case TYPE_LOOPFOR:
	  prot->end_proto=call_gestion(parm,prot,LOOP,NULL,NULL,0);
	  if (prot->end_proto==NOT_END) {
	    prot=&(parm->prot[parm->prot_index]);
	  }
	  break;
	case TYPE_CALL:
	  prot->end_proto=call_gestion(parm,prot,NOT_LOOP,NULL,NULL,0);
	  if (prot->end_proto==NOT_END) {
	    prot=&(parm->prot[parm->prot_index]);
	  }
	  break;
	case TYPE_FMT:
	  prot->end_proto=format_gestion(parm,prot);
	  if (prot->end_proto==NOT_END)
	    prot=&(parm->prot[parm->prot_index]);
	  break;
	case TYPE_GET:
	  prot->end_proto=call_gestion(parm,
							 prot,NOT_LOOP,
							 prot->pt->in.call.return_variable_name,
							 prot->pt->in.call.return_variable_display,
							 prot->pt->in.call.return_variable_options);
	  if (prot->end_proto==NOT_END)
	    prot=&(parm->prot[parm->prot_index]);
	  break;
	case TYPE_DISPLAY:
	  display_gestion(parm,prot);
	  break;
	case TYPE_CALC:
	  calc_gestion(parm,prot);
	  break;
	case TYPE_SUM:
	  sum_gestion(parm,prot);
	  break;
	case TYPE_END:
	  prot->end_proto=END_DUE_TO_END;
	  break;
	case TYPE_RETURN:
	  prot->end_proto=END_DUE_TO_RETURN;
	  if (prot->return_variable_name==NULL) {
	    add_display(parm,DE_PROTO,-1,"a return without a get.");
	    prot->end_proto=END_DUE_TO_ERROR;
	  } else {
	    prot->end_proto=return_value_gestion(parm,prot);
	  }
	  break;
	case TYPE_EXEC:
	  /* not yet implemented */
	  break;
	case TYPE_STOCK:
	  stock_gestion(parm,prot);
	  break;
	case TYPE_UNSTOCK:
	  prot->end_proto=unstock_gestion(parm,prot);
	  break;
	case TYPE_FLUX:
	  prot->end_proto=flux_gestion1(parm,prot);
	  if (prot->end_proto==NOT_END) {
	    prot=&(parm->prot[parm->prot_index]);
	  }
	  break;	

	default:
	  /* huuu ??? */
	  add_display(parm,DE_MEMORY,-1,"ERROR \n");
	  prot->end_proto=END_DUE_TO_ERROR;
	}
      } else { /* not a valid condition */
	/* add_display(parm,D_DEBUG,-1,"test (%s) not valid, %d",prot->pt->condition, prot->pt->type); */
	if (prot->pt->type==TYPE_BLOCK) {
	  prot->gotonext=block_gestion(parm,prot,0);
	}
      }   /* end test valid condition */
      
      /*** error and or ***/
      if ((prot->end_proto==END_DUE_TO_ERROR) &&
	  (prot->pt) && (prot->pt->next) &&
	  (prot->pt->next->or==YES)) {
	   prot->to_do_or=YES;
	   prot->end_proto=NOT_END;
      }
      /*** ------------ ***/

      /**** next description ****/
      if (prot->end_proto==NOT_END) {
	if (prot->gotonext) prot->pt=prot->pt->next;
	prot->gotonext=1;
	if (prot->pt==NULL) {
	  add_display(parm,DE_PROTO,prot->indent,"End of the chain in \"%s\".",prot->label_path);
	  prot->end_proto=END_DUE_TO_ERROR;
	}
      }
      /**** ---------------- ****/
      

    }  /* end while(prot->end_proto==NOT_END) */
  
    /*** end of protocol ***/


    /*** terminate a flux ***/
    if ((parm->flux) && 
	(parm->flux->prots_index==parm->prot_index) &&
	(parm->free_flux)) {
      free_flux(parm,prot);
    }
    /*** ---------------- ***/


    /*** calculate the size ***/
    nbbits=get_max_size(parm,prot);
    /*** (end) calculate the size ***/

    /* add_display(parm,D_DEBUG,-1,"end of proto(%d) <%s> user_length_in_bits=%d, max=%d, nb_bits_read = %d ",prot->end_proto, prot->label_path, prot->user_length_in_bits, prot->max_length_in_bits, (*(prot->nb_bits_read))); */

    /*** display the rest ***/
     if ((parm->prot_index==0)  ||                       /* the default */
    	   (prot->end_proto==END_DUE_TO_ERROR) ||           /* a bad end, du to some error */
    	   (prot->end_proto==END_DUE_TO_TOO_SMALL) ||       /* packet too small  */
    	   ((prot->to_loop==NOT_LOOP) && (prot->end_proto==END_DUE_TO_RETURN) && (prot->user_length_in_bits>-1) && (nbbits>0)) ||
    	   ((prot->to_loop==NOT_LOOP) && (prot->end_proto==END_DUE_TO_END) && (prot->user_length_in_bits>-1) && (nbbits>0))
	                                                  /* normal end before user length  */

	    ) { 

      display_rest(parm,prot->indent,nbbits/8);

      /*** calculate offbit and offoct ***/
      while(nbbits>8) { nbbits-=8 ; parm->offoct++; }
      parm->offbit+=nbbits;
      if (parm->offbit>=8) { parm->offbit-=8 ; parm->offoct++; }
      /*** --------------------------- ***/

      /*** re-calculate the protocol size ***/
      if (!parm->flux)
      (*(prot->nb_bits_read))+=nbbits;
      /*** ------------------------------ ***/
    
    }
    /*** (end) display the rest ***/

    /*** in a loopfor, continue ***/
    if ((nbbits>0) && (prot->to_loop==LOOP) && (prot->end_proto==END_DUE_TO_END)) {
      prot->pt=prot->start_point;
	 prot->end_proto=NOT_END;
    }
    /*** ---------------------- ***/
    /*** back to the previous protocol ***/
    else {
	 
      parm->prot_index--;
      if (parm->prot_index>=0)
	(*(parm->prot[parm->prot_index].nb_bits_read))+=(*(prot->nb_bits_read));
    }
    /*** ----------------------------- ***/

    /*** (end) end of protocol ***/

    
  } /* end while(parm->prot_index>=0) */
  return(0);
}

/********
	    analyse_packet
	    analyse the packet
	    called by pcap_loop...
*********/

void analyse_packet(u_char *user, const struct pcap_pkthdr *h, const u_char *p) {

  params parm;
  int i;
  variable *vp, *prev_vp; /* the variable list */

  /* initialising values */
  parm.aff_last=0;
  parm.vstart=NULL;
  parm.vend=NULL;
  parm.global_vstart=&global_variable_start;
  parm.stock_start=&stock_start;
  parm.flux_start=&flux_start;
  parm.flux=NULL;
  parm.p=p;
  parm.free_p=0;
  parm.packet_size=h->len*8;
  parm.offbit=0;
  parm.offoct=0;
  parm.get_variable=FOLLOW_LABEL_PATH;
  parm.sum=NULL;
  
  /* initialising the "first" protocol */
  parm.prot_index=0;
  parm.prot[0].label_path_length=0;
  parm.prot[0].label_path[0]='\0';
  parm.prot[0].label_name=NULL;
  parm.prot[0].label_str=NULL;
  parm.prot[0].add_to_label_path=1;
  parm.prot[0].offset=0;
  parm.prot[0].indent=0;
  parm.prot[0].to_loop=NOT_LOOP;
  parm.prot[0].to_do_or=0;
  parm.prot[0].user_length_in_bits=-1      ;
  parm.prot[0].max_length_in_bits=h->len*8;
  parm.prot[0].pt=c_start;
  parm.prot[0].start_point=c_start;
  parm.prot[0].block_index=0;
  parm.prot[0].gotonext=1;
  parm.prot[0].end_proto=NOT_END;
  parm.prot[0].flux=NULL;
  parm.prot[0].prev_flux=-1;
  parm.prot[0].return_variable_name=NULL;
  parm.prot[0].return_variable_display=NULL;


  /* dump into a file */
  /* it is not usefull to analyse the packet (doesn't want a display, and no filters) */
  if ((opt.w_file)&&
      (opt.quiet)&&
      (!opt.filter)&&
      (!opt.cond_start)&&
      (!opt.cond_stop)) {
    pcap_dump(user,h, p);
    return;
  }

  /* the datalink variable */
  vp=add_new_variable("_datalink_",FORMAT_D,&parm);
  modify_variable (vp,0,F_SIGNED,(void *)(&opt.first_label),0);  

  if  (opt.filter) test(opt.filter,&opt.filter_compiled,&parm);

  /* decode the packet */
  decode_packet(&parm);

  /* check if we have to display */
  parm.get_variable=RAMBO_SEARCH;
  /*** filter testing ***/
  if ((opt.cond_start!=NULL) 
      && (start_display_packet==NO) 
      && (test(opt.cond_start,&opt.cond_start_compiled,&parm))) start_display_packet=YES;

  if ((opt.filter==NULL) ||
	 ((opt.filter) && (test(opt.filter,&opt.filter_compiled,&parm)))) {
    if ((!opt.quiet) && (start_display_packet)) display(&parm,opt);
    if (opt.w_file) pcap_dump(user,h, p); /* dump the packet */
  }

  if ((opt.cond_stop!=NULL)
      && (start_display_packet)
      && (test(opt.cond_stop,&opt.cond_stop_compiled,&parm))) start_display_packet=NO;

  /*** -------------- ***/


  /* flushing memory (variable list) */
  /***********************************/
  vp=parm.vstart;
  while(vp!=NULL) {
    prev_vp=vp;
    vp=vp->next;
    if ((parm.flux) &&                       /* keep in mem flux variables */
	(prev_vp==parm.flux->loc_var_flux)) break;
    /*** freeing values ***/
    for (i=0 ; i<prev_vp->num ; i++) free(prev_vp->values[i].val);
    free(prev_vp); 
  }

  if (parm.free_p) free(parm.p);

}



/********
	    cleanup
	    fonction de fin propre.
*********/
void cleanup(int signo) {
  struct pcap_stat stat;
  if (pt != NULL && pcap_file(pt) == NULL) {
    fflush(stdout);
    if (pcap_stats(pt, &stat) < 0)
	 fprintf(stderr, "pcap_stats: %s\n",pcap_geterr(pt));
    else {
	 fprintf(stderr, "\n%d packets received by filter\n",stat.ps_recv);
	 fprintf(stderr, "%d packets dropped by kernel\n",stat.ps_drop);
    }
  }
  if (dump) pcap_dump_close(dump);
  pcap_close(pt);
  exit(0);
}


/********
	    Main
*********/
int main (int argc, char **argv) {
  int i;

  /*** for pcap ***/
  bpf_u_int32 localnet, netmask;
  struct bpf_program fcode;
  u_char *pcap_userdata=NULL;
  /*** -------- ***/

  char ebuf[MAX_LINE];
  char *f_to_read;   /* le fichier a lire */

  /* initialsing */
  c_start=malloc(sizeof(confvalue));
  c_end=c_start;
  c_end->next=NULL;
  c_end->type=TYPE_NOTDEF;
  c_end->condition=NULL;
  c_end->indent=0;
  c_end->or=0;
  c_end->num_line=0;

  tablefirst=malloc(sizeof(tablestruct));
  tablefirst->next=NULL;
  tablefirst->name[0]='\0';

  codefirst=malloc(sizeof(codestruct));
  codefirst->next=NULL;
  codefirst->name[0]='\0';

  global_variable_start=NULL;
  stock_start=NULL;
  flux_start=NULL;

  /* reading params */
  getopts(&opt,argc,argv);


  /* check syntax file and exit */
  if (opt.label) {
    if ((f_to_read=search_in_path(opt.nice_path,opt.label))==NULL)
	 fprintf(stderr,"no file %s in %s\n",opt.label,opt.nice_path);
    else {
      readconf(f_to_read,&c_end,codefirst,tablefirst);
      free(f_to_read);
    }
    exit(0);
  }


  /* start to display packet ? */
  if (opt.cond_start!=NULL) start_display_packet=NO;

  /***********************/
  /* pcap administrative */
  /***********************/

  /* open file for reading */
  if (opt.r_file !=NULL) {
    if ((pt=pcap_open_offline(opt.r_file,ebuf))==NULL) {
	 fprintf(stderr, "Error: %s\n",ebuf);
	 exit(1);
    }
    localnet = 0;
    netmask = 0;
    fprintf(stderr, "%s: listening from %s\n",argv[0], opt.r_file);
  } else {
    /* open a device */
    if (opt.device == NULL) {
	 if ((opt.device = pcap_lookupdev(ebuf))==NULL) {
	   fprintf(stderr, "Error: %s\n",ebuf);
	   exit(1);
	 }
    }
   if ((pt=pcap_open_live(opt.device,opt.snaplen,opt.pflag,1000,ebuf))==NULL) {
	 fprintf(stderr,"unable to open device %s for reading\n",opt.device);
	   fprintf(stderr, "Error: %s\n",ebuf);
	 exit(1);
    }
    i = pcap_snapshot(pt);
    if (opt.snaplen < i) {
	 fprintf(stderr,"snaplen raised from %d to %d", opt.snaplen, i);
	 opt.snaplen = i;
    }
    if (pcap_lookupnet(opt.device, &localnet, &netmask, ebuf) < 0) {
      localnet = 0;
      netmask = 0;
      fprintf(stderr,"pcap_lookupnet: %s", ebuf);
    }
    fprintf(stderr, "%s: listening on %s\n",argv[0], opt.device);
  }


  /* apply pcap filter (with optimizer) */
  if (pcap_compile(pt, &fcode, opt.p_filter, 1 , netmask) < 0) {
    fprintf(stderr,"%s\n", pcap_geterr(pt));
    exit(1);
  }
  if (pcap_setfilter(pt, &fcode) < 0) {
    fprintf(stderr,"%s\n", pcap_geterr(pt)); 
    exit(1);
  }


  /* dump into a file */
  if (opt.w_file) {
    dump=pcap_dump_open(pt, opt.w_file);
    if (dump == NULL) {
      fprintf(stderr,"%s", pcap_geterr(pt));
      exit(1);
    }
    pcap_userdata = (u_char *)dump;
  }

  /* getting the first label , used in table _start_, with variable _datalink_ */
  opt.first_label=pcap_datalink(pt);

  fprintf(stderr,"reading from conf file %s\n",opt.desc_file);

  /* reading the first conf file */
  if ((f_to_read=search_in_path(opt.nice_path,opt.desc_file))==NULL) {
    fprintf(stderr,"no file %s in %s\n",opt.desc_file,opt.nice_path);
    exit(1);
  }
  readconf(f_to_read,&c_end,codefirst,tablefirst);
  free(f_to_read);

  /* reading packet */
  signal(SIGTERM, cleanup);
  signal(SIGINT, cleanup);
  
  if (pcap_loop(pt, opt.count, analyse_packet, pcap_userdata) < 0) {
    fprintf(stderr, "%s: pcap_loop: %s\n",argv[0], pcap_geterr(pt));
    exit(1);
  }



  /* closing device */   
  if (dump) pcap_dump_close(dump);
  pcap_close(pt);
  return(0);
}
