/********
	    Packet filter by wallrich@loria.fr

	    structs and define descriptions
*********/

#ifndef NICEDUMP_H
#define NICEDUMP_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#ifdef HAVE_UNISTD_H 
#include <unistd.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <signal.h>

#ifdef HAVE_PCAP_H
#include <pcap.h>
#else
#ifdef HAVE_PCAP_PCAP_H
#include <pcap/pcap.h>
#endif
#endif

 
/* how to stock decimal */

#define SIG_DEC long int
#define SIG_DEC_DSP "%ld"

#define UNSIG_DEC unsigned long int
#define UNSIG_DEC_DSP "%lu"

/* how to stock chars */
#define UCHAR unsigned char



/* Les conditions */
/* numerical */
#define SUP ">"
#define INF "<"
#define EQ  "=="
#define NE  "!="
#define SE  ">="
#define IE  "<="
#define IN  "=~"
#define NOTIN  "!~"
#define AND "&&"
#define OR "||"
#define NOT "!"


#define F_SUP '>'
#define F_INF '<'
#define F_EQ  '2'
#define F_NE  '3'
#define F_SE  '4'
#define F_IE  '5'
#define F_IN  '~'
#define F_NI  'N'
#define F_ER  'E'

#define F_PLUS  '+'
#define F_MIN  '-'
#define F_MUL  '*'
#define F_DIV  '/'
#define F_AFF  '='
#define F_MOD  'M'
#define F_AND 'A'
#define F_OR 'O'
#define F_NOT '!'
#define F_FIN '#'

#define ORDER_OPERATORS { F_MOD, F_MUL, F_DIV, F_PLUS, F_MIN, F_AFF, F_SUP, F_INF, F_EQ, F_NE, F_SE,  F_IE, F_IN, F_NI, F_AND, F_OR , F_FIN }

/* maximums */
#define MAX_LINE 2000
#define MAX_VARIABLE_SIZE 20000

/* loop or not loop in call 
   ie :
        just a call ar a loopfor
*/
#define NOT_LOOP 0
#define LOOP 1

/* 
   for get_variable_value

*/
/* we don't know the rank, so let it chose */
#define NO_RANG_DEF -1
/* we WANT the printvalue (= variable as a string) */
#define STRING_FORCE -2


/*********************************
  STRUCTS
**********************************/

/******
       values of variables 
*******/
typedef struct values_type {
  void *val;                      /* the value itself */
  char format;                    /* its format (F_STRING, F_SIGNED, etc... */
  int length;                     /* its length in octet, or -1 */
} variable_value;


/******
	  variables chain
*******/
typedef struct variable_type {
  char name[MAX_LINE];         /* variable name       */
  int num;
  variable_value values[10];   /* values of variable */ 
  char printvalue[MAX_LINE];   /* the print value */
  char dspfmt[MAX_LINE];       /* the display format */
  struct variable_type *next;  /* la variable suivante */
  struct variable_type *prev;  /* la variable precedente */
} variable;


/******
	  The big data chain
	  used in descrition of protocols...
*******/

/* label structure */
struct label_struct {
  char *label_name;
  char *label_str;
};

/* block structure for { and } */
struct block_struct {
  char type_of_block;  /* START_BLOCK, END_BLOCK, START_BLOCK_WHILE, */
  struct confvalue_type *corresp;
};

/* call structure */
struct call_struct {
  char typecall;             /* LABEL or TABLE */ 
  union {
    char *label_name;
    char *table_name;
  } in;

  struct confvalue_type *quickjump;  /* to memorise the call */
  char to_search;            /* to know if it is usefull to search */
  char *variable_name;       /* in case of a table */
  char type;                 /* EVAL or FIXED  */
  int bits;

  char *eval;
  struct compiled_str_type *eval_compiled;

  char *return_variable_name; /* in case of TYPE_GET, the var name to return */    
  char *return_variable_display; /* in case of TYPE_GET, the variable display format */
  char return_variable_options;
};

/* goto structure */
struct go_to_struct {
  char typegoto;             /* LABEL or TABLE */ 
  union {
    char *label_name;
    char *table_name;
  } in;
  struct confvalue_type *quickjump;  /* to memorise the call */
  char to_search;            /* to know if it is usefull to search */
  char *variable_name;       /* in case of a table */
};

/* Display structure */
struct display_struct {
  char *display_str;
};

/* Exec structure */
struct exec_struct {
  char code_name[MAX_LINE];
};

/* calc structure */
struct calc_struct {
  char *str;
  struct compiled_str_type *str_compiled;
};

/* sum structure */
struct sum_struct {
  char *str;
};

/* return structure */
struct return_struct {
  char *return_var;
};

/* stock structure 
   used for stock AND unstock keyword
*/
struct stock_struct {
  char *name[10];
  int  num;            /* number of variables in the key */
  char type;           /*STOCK_AT_THE_END, STOCK_NUM or STOCK_EVAL */
  int  length;         /* fixed length */
  char *eval;          /* eval */
  struct compiled_str_type *eval_compiled;

  /* the total length */
  char type_totl;           /*STOCK_AT_THE_END, STOCK_NUM or STOCK_EVAL */
  int  totl;                /* fixed length */
  char *eval_totl;     /* eval */
  struct compiled_str_type *eval_totl_compiled;
};

/* 
   flux structure 
*/
struct flux_struct {
  char *name[10];
  int  num;            /* number of variables in the key */
};

/* fmt structure */
struct fmt_struct {
  char *name;
  char indent;             /* temporary indentation (boolean) */
  char nr;                 /* re-read after (boolean) */
  int  num;                /* number of variables */
  char options;            /* lists of options :
						BLIND (dont display the variable)
					  */
  char type[10];           /* EVAL FIXED or UNTIL */
  char *format[10];        /* the format */
  int  length[10];         /* fixed length, or length of until string */
  char *eval[10];          /* eval */
  struct compiled_str_type *eval_compiled[10];
  char *until[10];         /* until */
  char *display;           /* the display format */
};

typedef struct confvalue_type {
  char type;                    /* Le type de variable */
  char *condition;              /* the condition */
  struct compiled_str_type *condition_compiled;
  char or;                      /* if a or */
  char indent;                  /* definitive indentation */
  union {
    struct label_struct label;
    struct go_to_struct go_to;
    struct call_struct call;
    struct display_struct display;
    struct fmt_struct fmt;
    struct calc_struct calc;
    struct exec_struct exec;
    struct block_struct block;
    struct stock_struct stock;
    struct flux_struct flux;
    struct sum_struct sum;
    struct return_struct ret;
  } in;
  char *file;                   /* the file in which we read this action */
  int num_line;                 /* its line */
  struct confvalue_type *next;  /* le suivant */
} confvalue;


/****** 
	   Code structure
*******/
typedef struct code_type {
  char name[MAX_LINE];     /* the name of the code */
  char code[MAX_LINE*3];   /* the code itself */
  struct code_type *next;
} codestruct ; 

/****** 
	   Table structure
*******/
typedef struct table_in_struct {
  char *key;                           /* the key              */
  char *label;                         /* the value            */
  struct confvalue_type *quickjump;    /* to memorise the call */
  char to_search;            /* to know if it is usefull to search */
  struct table_in_struct *next;
} table_in;

typedef struct table_type {
  char name[MAX_LINE];     /* the name of the table */
  int table_size;
  struct table_in_struct *in;
  struct table_type *next;
} tablestruct ; 


/******
	  option structure
*******/
typedef struct option_type {
  char *cond_start;      /* -sc: la condition de depart */
  struct compiled_str_type *cond_start_compiled;

  char cond_start_ok;    /* boolean to say if the start cond is OK or not */
  char *cond_stop;       /* -st: la condition de fin    */
  struct compiled_str_type *cond_stop_compiled;

  char cond_stop_ok;     /* boolean to say if the stop cond is OK or not */
  char *filter;          /* -f: filter */
  struct compiled_str_type *filter_compiled;

  char *p_filter;        /* -pf: the pcap filter */
  char *device;          /* -i: the device */
  char pflag;            /* -p: promiscuous or not */
  char quiet;            /* -q: display or not */
  int  count;            /* -c: nb of packet to read */
  int  snaplen;          /* -s: size of each packet */
  char *desc_file;       /* the first file to read */
  int  first_label;      /* the first label to get */
  char *nice_path;       /* the path to read files */
  char *r_file;          /* -r: the input file */
  char *w_file;          /* -o: the output file */
  char *label;           /* -check: the file label to check       */
  char *disp_label;      /* -dl: the label from wich we start to display */
  char *display_proto;   /* -proto: filter on protocol */
  char display_rest;     /* -rest: display the rest or not  */
  char sum;              /* -sum: display summary or not  */
  char debug;            /* -d  : my debug  */
  FILE *dest;            /* print to stdout (buffered) or stderr (unbuffered) */
} options;


/**********
		 stock structure
		 used when stock packet before decoding
***********/
typedef struct stock_fragment_type {
  int offset;        /* where to place this fragment in the packet */
  int length;        /* the length of the fragment */
  u_char *p;         /* the fragment itself */
  struct stock_fragment_type *next;    /* the next fragment */
} stock_fragment;

typedef struct stock_chain_type {
  char *key[10];         /* keys */
  int num;
  int length;                /* the total length (in octet) of the packet */
  stock_fragment *frag;      /* a ptr to the first fragment */
  struct stock_chain_type *next;
} stock_chain;



/******
       protocol structure
       The list of value for a protocol
 ******/
typedef struct protocol_type {
  char *label_name;               /* name of the protocol */
  char *label_str;                /* name (display string) of the protocol */
  int max_length_in_bits;         /* the max number of the bits in this protocol, set by nicedump */
  int user_length_in_bits;        /* the number of the bits in this protocol, ask by user  -1 if not def) */
                                  /* used to know if we can modify length_int_bits in case of flux */
  SIG_DEC *nb_bits_read;          /* the number of bits already read in this protocol
				   */
  SIG_DEC nbbitsread;             /* the saving value of *nb_bits_read , used by flux */
  int offset;                     /* offset (in bits) of the start of this protocol
							  from the beginning of the packet */
  int indent;                     /* the identation for this protocol        */
  int to_loop;                    /* to loop or not */
  int to_do_or;                   /* to do the or command or not */
  int gotonext;                   /* I have to go to the next step in confvalue ? */
  int add_to_label_path;          /* I have to add to the label path in case of a label ? */
  confvalue *start_point;         /* The start of description in the config */
  confvalue *pt;                  /* where whe are in the desc */

  char label_path[MAX_LINE];      /* The label path of this protocol */
  int  label_path_length;	 
  
  /* block gestion */
  char replay_block[100];         /* to know if we have to replay 
							  the block when we arived at the } */
  int block_index;
  int end_proto;                  /* to know how we stop this protocol (see TYPE_OF_END) */

  /* futur used : flux on each protocol */
  struct flux_chain_type *flux;   /* this protocol as a flux */
  int prev_flux;                  /* the protocol where a flux exist */

  /* get functions */
  char *return_variable_name;
  char *return_variable_display;
  char return_variable_options;

} protocol;


/**********
		 flux structure
		 used to stock the rest of the previous paquet.
***********/
typedef struct flux_chain_type {
  char *key[10];                 /* keys */
  int num;
  int length;                    /* the length (in octet) of the rest of the prev packet */
  u_char *p;                     /* the rest of the prev packet */
  int offbit;                    /* the bit offset in the first octet (not used yet) */
  variable *loc_var_flux;        /* the locals variable list we stock  */
  variable *glob_var_flux;       /* the globals variable list we stock */
  int first_time;                /* to know if we re-use the flux      */
  protocol prots[100];           /* to stock all protocols             */
  int prots_index;               /* index of prot where flux start     */
  int prots_index_length;        /* num of protocols in the flux       */
  struct flux_chain_type *next;
} flux_chain;


/******
	  parameters structure
	  used to give less args to functions...
	  Its a big package of differents variables.
*******/
typedef struct params_type {
  variable *vstart;             /* The variable list, local to the packet */ 
  variable *vend;
  variable **global_vstart;     /* The global variable list */ 
  stock_chain **stock_start;    /* the stock table */
  flux_chain **flux_start;      /* the flux table */
  char aff[MAX_LINE][MAX_LINE]; /* the aff table */
  int indenttab[MAX_LINE];
  char type_aff[MAX_LINE];
  char *sum;
  int  aff_last;
  int packet_size;              /* the size of the packet itself */
  u_char *p;                    /* the packet itself */
  char free_p;                  /* bool to know if we have to free p or not */
  int offoct;                   /* offset in octet */
  int offbit;                   /* offset in bits in the last octet */
  char get_variable;            /* the way to get variables         */
  flux_chain *flux;             /* the packet is in a flux          */
  int free_flux;                /* boolean to say if we must free the flux */
  protocol prot[100];           /* the list of protocol */
  int prot_index;
} params;

/*********
		syntax structure
		used to understand the config file
*********/
struct syntax {
  char *sequence;     /* the command line */
  char action;
  char flush;
  char *error_message;
};



/*********
		syntax compiled_str
		compiled_str is the eval / test string in a "compiled" form.
*********/
typedef struct compiled_str_type {

  /* the original values, from evaluation_stage1 */
  /* thoses values never change                  */
  /* ------------------------------------------- */
  char *str;
  variable_value *preset_value;  /* the value */
  variable *preset_var;          /* the variable, in case of affectation */

  char preset_operator;          /* the operator with the next element */
  char preset_neg;               /* a negation */
  char preset_inc;               /* not used yet */
  int preset_test_result;        /* result of a test is case of test */

  struct compiled_str_type *preset_down;/* in case of (), the ptr to the sub */
  struct compiled_str_type *preset_next;
  struct compiled_str_type *preset_prev;

  /* values who are moving after each calcul */
  /* (evaluation_stage2 )                    */
  /* --------------------------------------- */
  variable_value *value;
  variable *var;

  char operator;
  char neg;
  char inc;
  int test_result;
  
  struct compiled_str_type *down;
  struct compiled_str_type *next;
  struct compiled_str_type *prev;
} compiled_str;

/*************************
	  DEFINES 
**************************/
/* for options */
#define DEFAULT_SNAPLEN 1600  /* too big, i know... */

/* type of confvalue */
#define TYPE_NOTDEF 0
#define TYPE_LABEL 1
#define TYPE_CALL 2
#define TYPE_LOOPFOR 3
#define TYPE_FMT 4
#define TYPE_DISPLAY 5
#define TYPE_CALC 6
#define TYPE_END 7
#define TYPE_EXEC 8
#define TYPE_GOTO 9
#define TYPE_BLOCK 10
#define TYPE_STOCK 11
#define TYPE_UNSTOCK 12
#define TYPE_FLUX 13
#define TYPE_SUM 14
#define TYPE_GET 15
#define TYPE_RETURN 16

/* type of fmt */
#define FMT_EVAL 1
#define FMT_FIXED 2
#define FMT_UNTIL 3

/* type of stock */
#define STOCK_AT_THE_END 0
#define STOCK_NUM 1
#define STOCK_EVAL 2

/* type of call / loopfor */
#define CALL_NONE 0
#define CALL_EVAL 1
#define CALL_FIXED 2

/* type of call / loopfor / goto / get */
#define CALL_LABEL 0
#define CALL_TABLE 1

/*********
		format of numbers 
**********/
#define FORMAT_D "%d"
#define FORMAT_U "%u"
#define FORMAT_S "%s"
#define FORMAT_B "%b"

#define F_STRING 'S'
#define F_SIGNED 's'
#define F_UNSIGNED 'u'
#define F_BINAR  'b'
#define F_UNDEF  'U'
#define F_ERROR 'E'


/* block type */
#define START_BLOCK '0'
#define START_BLOCK_WHILE '1'
#define END_BLOCK '2'


/* increments */
#define F_POST_INC '+'
#define F_POST_DEC '-'
#define F_PREV_INC '<'
#define F_PREV_DEC '>'

/* start file (not use) and path */
#define NICEDUMP_PATH ".:/etc/nicedump:/usr/etc/nicedump:"PRIMARY_PATH"\0"
#define NICEDUMP_FIRST_FILE "nicedump.ndc\0"
#define EXTEND ".ndc\0"
#define LENGTH_EXTEND 4

#define SIZETOREAD 1024

#define ADDITION 0
#define SOUSTRACTION 1
#define MULTIPLICATION 2
#define DIVISION 3

/***** stock defines ****/
#define ADD 1
#define DONT_ADD 0

/***** How get variable *****/
#define FOLLOW_LABEL_PATH '0'
#define STRICT_LABEL_PATH '1'
#define RAMBO_SEARCH '2'

/********************/
/* fmt options      */
/********************/
#define FMT_BLIND '\1'


/********************/
/* display messages */
/********************/
/* Error messages */
#define DE_SYNTAX 0
#define DE_CALC 1
#define DE_PROTO 2
#define DE_DEVICE 3
#define DE_MEMORY 4

/* normal messages */
#define D_VALUE  5
#define D_LABEL  6
#define D_REST   7
#define D_SUFFIX 8

/* other messages */
#define D_WARNING 9
#define D_INFO 10
#define D_DEBUG 11

/* TYPE_OF_END of protocol */
#define NOT_END 0
#define END_DUE_TO_ERROR 1
#define END_DUE_TO_END 2
#define END_DUE_TO_TOO_SMALL 3
#define END_DUE_TO_RETURN 4

#define YES 1
#define NO 0

/********************/
/* macros           */
/********************/
#define MEMCPY_STR(a,b,c) memcpy(a,b,c); a[c]='\0'
#define MY_STRDUP(a,b,c) a=malloc(c+1); MEMCPY_STR(a,b,c) 
#define MY_TYPEDUP(a,b,c) a=malloc(sizeof(c)); memcpy(a,b,sizeof(c)) 
#define MY_MALLOC malloc
#define MY_MAX(a,b) (a>b?a:b)
#define MY_MIN(a,b) (a<b?a:b)

#endif




