/********
	    Packet filter by wallrich@loria.fr

	    list of func used when reading a packet 
*********/

#include "nicedump.h"
#ifdef STDC_HEADERS
#include <stdarg.h>
#else
#include <varargs.h>
#endif

/********************************
 ******* EXTERN FUNCS  **********
 ********************************/

extern char *get_world(char *,int *,const int);
extern variable *get_variable_by_name(const char *, params *, char);
extern variable_value *get_variable_value(variable *, int);
extern int is_start_word(char);

/**************************
 ******* LABEL   **********
 **************************/
/********
	    get_label_ptr
	    get the label in funct of the name.
	    otherwise return NULL.
*********/
extern confvalue *get_label_ptr(char *label, confvalue *c_start) {
  confvalue *pt;

  pt=c_start;
  while(pt!=NULL) {
    if ((pt->type==TYPE_LABEL) 
	   && (strcmp(pt->in.label.label_name,label)==0)) return(pt);
    pt=pt->next;
  }

  return(NULL);
}

/**************************
 ******* DISPLAY **********
 **************************/
/***********
		  add_display
		  Add the string to the display struct
************/
#ifdef STDC_HEADERS
int add_display(params *parm, const char type_disp, int indent, const char *s, ...) {
  int retvalue;
  int rej;
  va_list ap;
  va_start(ap, s);
#else
int add_display(params *parm, const char type_disp, int indent, va_alist)
va_dcl
  {
    int retvalue;
    int rej;
    char *s;
	 
   va_start(ap);
   s = va_arg(ap, char *);
#endif

   /*** take the previous indentation ***/
   if (indent==-1) {
	if (parm->aff_last) indent=parm->indenttab[parm->aff_last-1];
	else indent=0;
   }
   /*** ----------------------------- ***/

   if (parm->aff_last>MAX_LINE) return(-1);

   parm->indenttab[parm->aff_last]=indent;
   parm->type_aff[parm->aff_last]=type_disp;   
   retvalue=vsnprintf(parm->aff[parm->aff_last],MAX_LINE,s,ap);
   parm->aff_last++;

   if ((retvalue<0) ||(retvalue>=MAX_LINE)) {
     rej=((retvalue<0)?0:retvalue-MAX_LINE+1);
     if (rej)
       add_display(parm,D_SUFFIX,-1," (truncated. %d octets lost)...",rej);
     else
       add_display(parm,D_SUFFIX,-1," (truncated)...");
   }


   /*
     vfprintf(stdout,s,ap);
     fprintf(stdout,"\n");
   */
   
   va_end(ap);
   return(0);
}


/*
  display the rest of the data not read int the protocol
  length is in octet :
*/
void display_rest(params *parm, int indent, int length) {
  int i;
  int l=0, b=0;
  
  char h[MAX_LINE];  /* the hexa version */
  int hlen=0;
  char a[MAX_LINE];  /* the ascii version */
  int alen=0;
  int block=0;


  for (i=parm->offoct; i<(parm->offoct+length) ; i++) {
    /* beginning of a line */
    if (l==0) { 
	 hlen=0;
	 alen=0;
	 l=1; 
	 b=0;
	 block=0;
    }
    
    /* middle of a line */
    if (l==1) {
	 sprintf(h+hlen,"%02x",parm->p[i]);
	 hlen+=2;
	 if ((parm->p[i]>=' ') && (parm->p[i]<='~')) a[alen++]=parm->p[i];
	 else a[alen++]='.';
	 b++;
	 if (b==2) {
	   block++;
	   b=0;
	   h[hlen++]=' ';
	   if (block==8) l=2;
	 }
    }

    /* end of a line */
    if (l==2) {
	 h[hlen]='\0';
	 a[alen]='\0';
	 add_display(parm,D_REST,indent,"%s] [ %s",h,a);
	 l=0;
    }
  }  /* end for */
  
  /* the last line, incomplete */
  if (l!=0) {
    /* completion with blanks */
    while(alen<16) {
	 h[hlen++]=' ';h[hlen++]=' ';
	 a[alen++]=' ';
	 b++;
	 if (b==2) {
	   b=0;
	   h[hlen++]=' ';
	 }
    }

    h[hlen]='\0';
    a[alen]='\0';
    add_display(parm,D_REST,indent,"%s] [ %s",h,a);
    l=0;
  }
  

}

/*******
	display value
	this proc change the %<variable_name> in its value
	(the printvalue). If then variable is suffixed with .[0-9]
	the value is adapded
********/
char *display_str(params *parm, char *str) {
  variable *p;
  char *s;
  char *varname;
  int i,j=0,k, len;
  int doit;
  int rang;
  variable_value *value;

  s=malloc(MAX_LINE);
  for (i=0;i<=strlen(str);i++) {
    doit=0;    
    if ((str[i]=='%') && (i>0) && (str[i-1]!='\\')) doit=1;
    if ((str[i]=='%') && (i==0)) doit=1;

    if (doit) {
      i++;  /* forget the '%' */
      /* look for the end of the variable */
      k=i;
      while((k<=strlen(str)) && ((is_start_word(str[k])) || (str[k]=='.'))) k++;
      MY_STRDUP(varname,str+i,k-i);
      i=k-1;

      /* look for the rang */
      rang=STRING_FORCE;
      len=strlen(varname);
      if ((str[i-1]=='.') && (str[i]>='0') && (str[i]<='9')) {
	rang=str[i]-'0';
	varname[len-2]='\0';
      }
      p=get_variable_by_name(varname,parm,FOLLOW_LABEL_PATH);
      free(varname);
      if (p) {
	value=get_variable_value(p,rang);
	if (value) {
	  if (value->format==F_STRING) snprintf(s+j,MAX_LINE-j,"%s",(char *)value->val);
	  if (value->format==F_SIGNED) snprintf(s+j,MAX_LINE-j,"%ld",*((SIG_DEC *)value->val));
	  if (value->format==F_UNSIGNED) snprintf(s+j,MAX_LINE-j,"%ld",*((UNSIG_DEC *)value->val));
	  while(s[j]!='\0') j++;  /* jump */
	}
      }
      /* i++; pass the the char after */
    } else { /* just copy the char */
      s[j]=str[i];
      j++;
    }
  }
  
  return(s);
}


/***********************************
 ***** READ IN THE PACKET **********
 ***********************************/


/******
	  get_*_*_in_packet
	  thoses functs read in the packet, and put the value
	  int var.
	  they modifie the offbit and offoct.
	  they return -1 if pbs.
*******/
/******
	  get_fixed_num_in_packet
*******/
int get_fixed_num_in_packet(params *parm, int nbits, int max_lenth_to_read, void *var) {
  
  int tor;              /* the num of bits to read in this octet */
  int toshift;
  UNSIG_DEC rep=0; 
  unsigned char mask;   /* the mask to read */

  /* too short !! */  
  /* unable to read the entire variable */
  if (max_lenth_to_read<nbits)  return(-1); 

  /* if whe don't start int the beginning of a octet */
  mask=255>>(parm->offbit);

  while(nbits>0) {
	 rep*=256;  /* shift in 10 base */

	 tor=(nbits<=(8-(parm->offbit)))?nbits:(8-(parm->offbit));
	 toshift=(8-tor-(parm->offbit));
	 mask=(255>>(parm->offbit))&(255<<toshift);

	 rep+=((parm->p[(parm->offoct)]&mask)>>toshift);
	 (parm->offbit)+=tor;
	 if ((parm->offbit)>=8) {
	   (parm->offbit)-=8;
	   (parm->offoct)++;
	 }
	 nbits-=tor;

  }

  memcpy(var,&rep,sizeof(UNSIG_DEC));
  return(0);
}

/******
	  get_fixed_char_in_packet
	  var is allocated in this func.

*******/
int get_fixed_char_in_packet(params *parm, int nbits, int max_lenth_to_read, char **var) {
  
  int oct;              /* the num of oct to read in this octet */


  /* too short !! */  
  /* unable to read the entire variable */
  if (max_lenth_to_read<nbits)  return(-1); 

  /* if whe don't start in the beginning of a octet ==> error */
  if ((parm->offbit)) return(-1);
  
  oct=nbits/8;
  (*var)=malloc(oct+1);
  memcpy((*var),parm->p+(parm->offoct),oct);
  (*var)[oct]='\0';
  (parm->offoct)+=oct;

  return(0);
}


/************
            get_numbit_to_read
            get the number of bit to read from the packet 
            until the string stop is reached
            if  the string stop is not found, value -1 is returned
*************/
int get_numbit_to_read(params *parm, const char *stop, int size_stop, int max_lenth_to_read) {

  int numbit=0;           /*the result returned*/
  int offoct;

  if (parm->offbit>0){
    add_display(parm,DE_PROTO,-1,"Unable to read a char not in the beggining of a octet...\n");
    return(-1);
  }
  offoct=parm->offoct;
  while(((numbit+8*size_stop)<=max_lenth_to_read) &&
	(memcmp(parm->p+offoct,stop,size_stop)!=0)){
    offoct++;
    numbit+=8;
  }
  numbit+=8*size_stop;          /* the string stop is included in reading */               

  if (numbit>max_lenth_to_read) {
    add_display(parm,DE_PROTO,-1,"<%s> not found int the packet size=%d\n",stop,max_lenth_to_read);
    return(-1);
  }
  return(numbit);
}

