/*
 * PcaPTure PaCket Sniffer ver 1.0
 * by Cthullhu 2000 
 * Member of pkcrew (www.pkcrew.org)
 * 
 * Usage :
 * pcature [-i interface] [-l logfile] [-v]
 *
 * -v Verbose Mode        
 * -i interface  (ppp0,eth0,lo ..... )
 * -l logfile    file to redirect logging 
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>

#include "./libpcap-0.4/pcap.h"

#include "sniff.h"

#define LOGFILE "logfile"
#define IPHDR   sizeof(struct iphdr)
#define TCPHDR  sizeof(struct tcphdr)

extern char *optarg;
extern int   optind;

FILE        *logfile;
char        *logfilename;
char        *logbuf;
char        *dev;
int          verbose=0;
int          change_log=0;
u_char      *buf;
u_short      ports[] = { 21, 23, 25, 110, 513, 514 };	/* Default Logging Ports */
u_short      eth_size;

struct pcap *capt_desc;
struct pcap_pkthdr pkhdr;

int
main (int argc, char **argv)
{
	int   cmd_parse ;
	char *port_buf;
	char  ebuf[255];

	signal (SIGINT, close_log_file);
	signal (SIGQUIT, close_log_file);
	signal (SIGKILL, close_log_file);

	if (argc < 2)
	  {
		  print_usage (argv[0]);
		  exit (0);
	  }

	while ((cmd_parse = getopt (argc, argv, "vl:i:")) != EOF)
	  {
		  switch (cmd_parse)
		    {
		    case 'i':
			    dev = optarg;
			    break;
		    case 'l':
			    change_log++;
			    logfilename = optarg;
			    break;
                    case 'v':
                            verbose++;
                            break;
		    case '?':
		    default:
			    print_usage (argv[0]);
			    exit (0);
		    }
	  }

        
        printf ("PcaPtUrE packet sniffer version 1.0 \n");
        printf ("by Cthulhu (pkcrew - www.pkcrew.org)\n");
        fflush(stdout);

	capt_desc = pcap_open_live (dev, 8024, 1, 1000, ebuf);
	if (capt_desc == NULL)
	  {
		  fprintf (stderr, "Sorry . Can't open device %s\n\n", dev);
		  exit (0);
	  }

	switch (pcap_datalink (capt_desc))
	  {
	  case DLT_NULL:
		  eth_size = 4;
		  break;
	  case DLT_EN10MB:
	  case DLT_EN3MB:
		  eth_size = 14;
		  break;
	  case DLT_PPP:
		  eth_size = 4;
		  break;
	  case DLT_SLIP:
		  eth_size = 16;
		  break;
	  case DLT_FDDI:
		  eth_size = 21;
		  break;
	  case DLT_RAW:
		  eth_size = 0;
		  break;
	  default:
		  fprintf (stderr, "\n Unknown device .... Aborting ..\n\n");
		  exit (0);
	  }

        if (change_log)
          {
                 if ((open_log_file () == ERR))
	           {
		          fprintf (stderr,
			           "\nCannot Open Log File ... Aborting ..\n\n");
	           }
          }

	while (1)
	  {
		  buf = (u_char *) pcap_next (capt_desc, &pkhdr);
                  fflush(stdout);
		  if ((buf != NULL) && ((pkhdr.len - eth_size) >= TCPHDR))
		    {
			    do_logging ();
		    }
	  }

	return (SUCCESS);
}


int
open_log_file (void)
{
	if ((logfile = (fopen ((change_log) ? logfilename : LOGFILE , "w"))))
		return (SUCCESS);
	return (ERR);
}

void
close_log_file (int sig)
{
	if (fclose (logfile) == EOF)
		fprintf (stderr,
			"\nSorry...Can't close log file ... Aborting !\n\n");

	exit (0);
}

int
do_logging (void)
{
	int istcp = 0, logging = 0, i;
	int sp;

	ip  = (struct iphdr *)  (buf + eth_size);
	tcp = (struct tcphdr *) (buf + sizeof (struct iphdr) + eth_size);

	if ((ip->protocol == IPPROTO_TCP))
	  {
		  /* if ((pkhdr.len - (eth_size + IPHDR)) < sizeof(struct tcphdr)) */
		  istcp++;
	  }

	if (istcp)
	  {
		  for (i = 0; ports[i] != (u_short) NULL; i++)
		    {

			    if (ports[i] == ntohs (tcp->th_sport)
			        || ports[i] == ntohs (tcp->th_dport))
			      {
				      logging++;
			      }
		    }

		  if (logging)
		    {
			    struct packet *pkt;
			    pkt =  malloc ((sizeof (struct packet) + sizeof (struct iphdr)
                                          + sizeof (struct tcphdr)));
			    logbuf = (char *) (buf + sizeof (struct iphdr) +
 				               sizeof (struct tcphdr) + eth_size);

			    pkt->sourceip = ip->saddr;
			    pkt->dstip = ip->daddr;
			    pkt->sourceport = tcp->th_sport;
			    pkt->dstport = tcp->th_dport;

		            print_data (pkt, logbuf, tcp);

			    free (pkt);
			    memset (buf, 0x0, sizeof (buf));
		    }
	  }
}

void
print_data (struct packet *pkt, char *logbuf, struct tcphdr *tcp)
{
	char dst[4012];
        char flag ;  
	int i;
	struct in_addr in;

	in.s_addr = pkt->sourceip;

	fprintf ((change_log) ? logfile : stdout , "\n.ooOO %s:%i  ----> ", inet_ntoa (in),
		ntohs (pkt->sourceport));

	in.s_addr = pkt->dstip;

	fprintf ((change_log) ? logfile : stdout , "%s:%i ", inet_ntoa (in), ntohs (pkt->dstport));

        /* More Interesting Flags */

        if      (tcp->th_flags & TH_SYN)
                  flag = 'S';
        else if (tcp->th_flags & TH_FIN)
                  flag = 'F';
        else if (tcp->th_flags & TH_RST)
                  flag = 'R';
        else if (tcp->th_flags & TH_ACK)
                  flag = 'A';
        else 
                  flag = '.';
          
	fprintf ((change_log) ? logfile : stdout , "%c seq %u win %u ack %u OOoo.\n ", 
                flag, tcp->th_seq,
		tcp->th_win, tcp->th_ack);

        if (verbose)
          {
                memset (dst, 0, sizeof (dst));

         	for (i = 0; i <	(pkhdr.len - (sizeof (struct iphdr) +
	          		 sizeof (struct tcphdr) + eth_size)); i++)
	         {
		        if      (isprint (logbuf[i]))
		        	   dst[i] = logbuf[i];
		        else if (dst[i] == '\n' || dst[i] == '\r')
		        	   dst[i] = '\n';
		        else
		                   dst[i] = '.';
	         }
         	i++;
        	dst[i] == '\0';

        	for (i = 0; i < strlen (dst); i++)
	        	fputc (dst[i], (change_log) ? logfile : stdout);
           } /* if (verbose) */

	fflush ((change_log) ? logfile : stdout);
}

void
print_usage (char *executable_name)
{
	printf ("\n %s [-i interface] [-l logfile] [-v]\n\n",
		executable_name);
}
