/* tcp-msglogger - tcpserver.c, by pheisar@ccl.pt 

   tcpserver(.c) saves the strings (packets) send by tcpclient(.c) in a file (FILENAME).
   the file contains the hostname of the sender and, of course, the string.

   usage:
   $ tcpserver &
*/

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define PORT 31337
#define FILENAME ".msgs"
#define USAGE "usage: %s &\n", argv[0]
#define MAXSIZE 1024

struct sockaddr_in s_addr, c_addr;

void main(int argc, char **argv)
{
	FILE *fd;
	int sfd, nsfd, n, c_len;	
	char mesg[MAXSIZE];

	if(argc > 2) {
		printf(USAGE);
		exit(1);
	}

	if( (fd = fopen(FILENAME, "a")) == NULL) {
		perror("fopen()");
		exit(1);
	}

	if( (sfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
		perror("socket()");
		exit(1);
	}

	memset((char *)&s_addr, 0, sizeof(s_addr));	
	s_addr.sin_family = AF_INET;
	s_addr.sin_port = htons(PORT);
	s_addr.sin_addr.s_addr = htonl(INADDR_ANY);

	if( (bind(sfd, (struct sockaddr *)&s_addr, sizeof(s_addr))) == -1) {
		perror("bind()");
		exit(1);
	}

	listen(sfd, 5);

	for( ; ; ) {
		c_len = sizeof(c_addr);
		if( (nsfd = accept(sfd, (struct sockaddr *)&c_addr, &c_len)) == -1) {
			perror("accept()");
			exit(1);
		}		
		if( (n = recv(nsfd, mesg, sizeof(mesg), 0)) < 0) {
			perror("recv()");
			exit(1);
		}

		fprintf(fd, "[%s] -> %s\n", inet_ntoa(c_addr.sin_addr), mesg);	 	 
		fflush(fd);
	}

	close(nsfd);
	close(sfd);
	fclose(fd);
}
		
