/* match.c
 * Some functions dealing with strings.
 */

#include "weedlog.h"

/* wl_psize(buf);
 * A little function to get the number of valid characters in a pointer.
 * Why? strlen would probly work as well but I was felt like reinvinting
 * the wheel for the hell of it.
 */
int wl_psize(const char *buf) {
  register const char *s;
  int count = 0;

  for(s = buf; *s; s++) {
    count++;
  }
  return count;
}

/* fixprogname(argv[0]);
 * This function strips the leading path off the filename given by argv[0].
 */
char *fixprogname(const char *fullname) {
  register const char *namestring = fullname;
  char *retname = (char*)NULL;
  int i = 0;

  if(!match(fullname, "*/*")) return (char*)fullname;
  if((i = wl_psize(fullname))) {
    char temp1[i], temp2[i];
    int n = 0;
    namestring += i;
    while(*namestring != '/') {
      if(temp1) snprintf(temp2, i, "%s", temp1);
      if(temp2) snprintf(temp1, i, "%c%s", *namestring, temp2);
      else snprintf(temp1, i, "%c", *namestring);
      namestring--;
    }
    n = wl_psize(temp1);
    if((retname = (char*)malloc(n)) < 0) {
      print_log("malloc: %s\n", strerror(errno));
      return (char*)NULL;
    }
    memcpy(retname, &temp1, n);
    return retname;
  } else retname = (char*)fullname;

  return retname;
}


/* checkisnum(string);
 * Checks to see whether the string is a number or not.
 */
int checkisnum(const char *string) {
  register const char *s = string;
  int firstloop = 1;

  while(*s) {
    if((*s != '0') && (*s != '1') && (*s != '2') && (*s != '3') && (*s != '4')
	&& (*s != '5') && (*s != '6') && (*s != '7') && (*s != '8')
	&& (*s != '9')) {
      if(firstloop) {
        if(*s != '-') {
	  return 0;
	}
	firstloop = 0;
      } else {
        return 0;
      }
    }
    s++;
  }
  return 1;
}

/* match(string, wild);
 * This function checks the string (string) against the (possibly?) wild
 * string (wild). The wild cards '*' and '?' are allowed. '*' matches any
 * number of characters. '?' matches against 1 character and cannot match
 * against less or more. Examples:
 * 'Heasdfallo' matchs h*llo, 'hellesallo' does not match h*llo
 * 'heheheeleleleeelllll!!!' matchs *h*ll*
 * 'he11o' matchs he??o, 'helo' does not match he??o
 * NOTE: This function is NOT case insensitive.
 *
 * Returns 1 on match.
 * Returns 0 on non-match.
 */
int match(const char *string, const char *wild) {
  register const char *s = string, *w = wild;
  int star = 0, n = 0;

#ifndef GOT_MATCH
#define GOT_MATCH 1
#endif
#ifndef MATCH_FAILED
#define MATCH_FAILED 0
#endif

  do {
top:
    if(*w == '\\') {	/* This is for escaping '*' & '?' in the wild string. */
      w++;		/* Go to the next char and see if it is is a '*' */
      if((*w == '*') || (*w == '?')) {	/* or a '?', otherwise ignore it's */
        if(tolower(*s) != tolower(*w))	/* exsistance. */
	  return MATCH_FAILED;
	w++;
	s++;
      }
    }
    if(*w == '*') {	/* Checks for multi-char wild card. */
      while(*w && (*w == '*')) { /* Clear out multiple stars because they are stupid, */
        w++;		/* '******' does exactly the same as '*'. */
      }
      if(!*w)		/* Last char is a '*' so we should asume the rest */
        return GOT_MATCH; /* is a match. */
      star = 1;
    }
    if(*w == '?') {	/* Checks for single char wild card. */
      while(*w && (*w == '?')) {
	w++;		/* If '?' jump to the next char to check against. */
	if(!star) {	/* Make sure there isn't a '*' and a '?' because a */
	  s++;		/* '*' preceeding '?' nullifies '?', otherwise move */
	}
      }
    }			/* the string we are checking forward 1 char. */
    if(star) {		/* This is set when a '*' is found. */
star:
      if(!*w)		/* Redundant '*' check: asume last char is wild. */
        return GOT_MATCH;
      if(!*s)		/* Would have returned 1 if last char was wild, so */
        return MATCH_FAILED; /* we asume a missmatch if nothing to check. */
      while(*s && (tolower(*s) != tolower(*w))) {
        s++;		/* '*' is any number of chars so go */
      }
      if(!*s && *w)	/* There is nothing left to check but we still have */
        return MATCH_FAILED; /* stuff to check against so assume a missmatch. */
      if(tolower(*s) == tolower(*w)) {
        int need = 1, got = 1;
        while(tolower(*w) == tolower(w[need])) {
	  need++;
	}
        while(tolower(s[got]) == tolower(*w)) {
	  got++;
	}
	if(got >= need) {
	  for(n = 0; n < need; n++) {
	    *w++;
	  }
	  for(n = 0; n < got; n++) {
	    *s++;
	  }
	  if(!(*s && *w)) {
	    return GOT_MATCH;
	  }
	} else {
	  for(n = 0; n < got; n++) {
	    *s++;
	  }
	  goto star;
	}
      }
      star = 0;
    }
    if(tolower(*s) != tolower(*w)) {
      if((*w == '*') || (*w == '?')) /* double check for wildcards */
        goto top;
      else return MATCH_FAILED;	/* We have a missmatch so return 0. */
    }
    while((*s && *w) && (tolower(*s) == tolower(*w))) {
      w++;
      s++;
      if(!(*s && *w))	/* If we are here the previous chars matched so if */
        return GOT_MATCH; /* there are none left we asume a match. */
    }
  }
  while(*s);

  if(!(*s && *w)) {
    return GOT_MATCH;
  }
  return MATCH_FAILED;
}
