/* parse_igmp.c
 * Stuff for dealing with igmp packets.
 */

#include "weedlog.h"

struct igmppacket {
  struct wl_iphdr {
#if __BYTE_ORDER == __LITTLE_ENDIAN
    u_char	ihl:4,
		version:4;
#elif __BYTE_ORDER == __BIG_ENDIAN
    u_char	version:4,
		ihl:4;
#else
#error	"Please fix <asm/byteorder.h>"
#endif
    u_char	tos;
    u_short	tot_len;
    u_short	id;
    u_short	frag_off;
    u_char	ttl;
    u_char	protocol;
    u_short	check;
    u_long	saddr;
    u_long	daddr;
  }ip;
  struct wl_igmphdr {
    u_char	type;
    u_char	code;
    u_short	csum;
    u_long	group;
  }igmp;
};

thread_function_t parse_igmp(void *argstruct) {
  int pktsize=0;
  char newbuf[512], suspip[24], psizebuf[24], tempflags[128], ipflags[128],
      saddress[82], daddress[88], pkttype[64], csums[64], miscstuff[200];
  struct parse_args *args = (struct parse_args *)argstruct;
  struct igmppacket *pkt = (struct igmppacket*)args->buffer;
  int verbose = args->verblev;
  time_t t;
  char *timestring = (char*)NULL;

  if(!pkt)
    goto pigmp_end;

  if((timestring = (char*)malloc(32)) < 0) {
    fprintf(stderr, "%s[%i]: malloc: %s\n", progname, wlpid, strerror(errno));
    goto pigmp_end;
  }
  if(dots) {
    if(dots > 1) {
      t = time(NULL);
      snprintf(timestring, 32, " (TS: %lu)", t);
    } else {
      struct tm *lt;
      t = time(NULL);
      lt = localtime(&t);
      if((strftime(timestring, 32, " (TS: %b %d %Y %T)", lt)) < 0) {
	fprintf(stderr, "%s[%i]: strftime: %s", progname, wlpid, strerror(errno));
      }
    }
  }

  if(pkt->ip.ihl != 5)
    snprintf(suspip, sizeof(suspip), " Suspicious IP Options: ");
  else
    snprintf(suspip, sizeof(suspip), " ");

  if((pktsize = ntohs(pkt->ip.tot_len)) > 1024) {
    if(floatsize) {
      snprintf(psizebuf, sizeof(psizebuf), "%4.02f KBytes", (float)pktsize/1024.0);
    } else {
      snprintf(psizebuf, sizeof(psizebuf), "%i KBytes", pktsize/1024);
    }
  } else snprintf(psizebuf, sizeof(psizebuf), "%i Bytes", pktsize);

  if(pkt->ip.saddr) {
    struct in_addr s;
    s.s_addr = pkt->ip.saddr;
    if(skipdns) {
      snprintf(saddress, sizeof(saddress), "%s", inet_ntoa(s));
    } else {
      snprintf(saddress, sizeof(saddress), "%s[%s]",
	  hostlookup(pkt->ip.saddr), inet_ntoa(s));
    }
  } else {
    snprintf(saddress, sizeof(saddress), "no source found");
  }
  if(pkt->ip.daddr) {
    struct in_addr d;
    d.s_addr = pkt->ip.daddr;
    if(skipdns) {
      snprintf(daddress, sizeof(daddress), " to %s,", inet_ntoa(d));
    } else {
      snprintf(daddress, sizeof(daddress), " to %s[%s],",
	  hostlookup(pkt->ip.daddr), inet_ntoa(d));
    }
  } else {
    snprintf(daddress, sizeof(daddress), ",");
  }

  if(verbose >= 3) {
    snprintf(ipflags, sizeof(ipflags), "  IP_FLAGS:");
    if(pkt->ip.version > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s VER=%i", ipflags,
	  pkt->ip.version);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.protocol > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s PROTO=%i", ipflags,
	  pkt->ip.protocol);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.frag_off > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s FRAG_OFF=%i", ipflags,
	  pkt->ip.frag_off);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.tos > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s TOS=%i", ipflags,
	  pkt->ip.tos);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.id > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s ID=%i", ipflags,
	  pkt->ip.id);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.ttl > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s TTL=%i", ipflags,
	  pkt->ip.ttl);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.check > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s CHECK=%i", ipflags,
	  pkt->ip.check);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
  } else snprintf(ipflags, sizeof(ipflags), " ");

  snprintf(pkttype, sizeof(pkttype), "Type %u (code %u)", pkt->igmp.type,
    pkt->igmp.code);

  if(pkt->igmp.group) {
    struct in_addr g;
    g.s_addr = pkt->igmp.group;
    snprintf(miscstuff, sizeof(miscstuff), "  Group: %s[%s] ",
	hostlookup(pkt->igmp.group), inet_ntoa(g));
/*    snprintf(miscstuff, sizeof(miscstuff), "  Group: %s ",
	inet_ntoa(g));*/
  } else snprintf(miscstuff, sizeof(miscstuff), " ");

  if(verbose >= 2)
    snprintf(csums, sizeof(csums), " Checksum: %u", pkt->igmp.csum);
  else snprintf(csums, sizeof(csums), " ");

  if(dots) {
    snprintf(newbuf, sizeof(newbuf), "IGMP:%s%s%s packet from %s%s Size: %s"
	"%s%s%s", timestring, suspip, pkttype, saddress, daddress, psizebuf,
	miscstuff, csums, ipflags);
  } else {
    snprintf(newbuf, sizeof(newbuf), "IGMP:%s%s packet from %s%s Size: %s"
	"%s%s%s", suspip, pkttype, saddress, daddress, psizebuf, miscstuff,
	csums, ipflags);
  }
  print_log("%s\n", newbuf);

pigmp_end:
  if(timestring) free(timestring);
  if(args->buffer) free(args->buffer);
  if(argstruct) free(argstruct);
#ifdef USETHREADS
  if(getpid() != wlpid) {
    if(debug >= 4) print_log("debug: Child thread == %u, exiting thread\n",
	getpid());
    deactivate_thread();
    pthread_exit(NULL);
  } else {
    if(debug >= 4)
      print_log("debug: Child (%i) and Parent (%i) have the same pid?!\n",
	  getpid(), wlpid);
  }
#endif

  return 0;
}
