#include "weedlog.h"

struct tcppacket {
  struct wl_iphdr {
#if __BYTE_ORDER == __LITTLE_ENDIAN
    u_char	ihl:4,
		version:4;
#elif __BYTE_ORDER == __BIG_ENDIAN
    u_char	version:4,
  		ihl:4;
#else
#error	"Please fix <asm/byteorder.h>"
#endif
    u_char	tos;
    u_short	tot_len;
    u_short	id;
    u_short	frag_off;
    u_char	ttl;
    u_char	protocol;
    u_short	check;
    u_long	saddr;
    u_long	daddr;
  }ip;
  struct wl_tcphdr {
#ifdef __FAVOR_BSD
    u_short	th_sport;		/* source port */
    u_short	th_dport;		/* destination port */
    u_long	th_seq;			/* sequence number */
    u_long	th_ack;			/* acknowledgement number */
#if __BYTE_ORDER == __LITTLE_ENDIAN
    u_char	th_x2:4,		/* (unused) */
		th_off:4;		/* data offset */
#elif __BYTE_ORDER == __BIG_ENDIAN
    u_char	th_off:4,		/* data offset */
		th_x2:4;		/* (unused) */
#else
#error	"Adjust your <asm/byteorder.h> defines"
#endif
    u_char	th_flags;
#define	TH_FIN	0x01
#define	TH_SYN	0x02
#define	TH_RST	0x04
#define	TH_PUSH	0x08
#define	TH_ACK	0x10
#define	TH_URG	0x20
    u_short	th_win;			/* window */
    u_short	th_sum;			/* checksum */
    u_short	th_urp;			/* urgent pointer */
#else
    u_short	source;
    u_short	dest;
    u_long	seq;
    u_long	ack_seq;
#if __BYTE_ORDER == __LITTLE_ENDIAN
    u_short	res1:4,
		doff:4,
		fin:1,
		syn:1,
		rst:1,
		psh:1,
		ack:1,
		urg:1,
		res2:2;
#elif __BYTE_ORDER == __BIG_ENDIAN
    u_short	doff:4,
		res1:4,
		res2:2,
		urg:1,
		ack:1,
		psh:1,
		rst:1,
		syn:1,
		fin:1;
#else
#error	"Adjust your <asm/byteorder.h> defines"
#endif	
    u_short	window;
    u_short	check;
    u_short	urg_ptr;
#endif
  }tcp;
};

thread_function_t parse_tcp(void *argstruct) {
  int synpkt=0;
  int pktsize=0;
  char newbuf[512], suspip[24], psizebuf[24], tempflags[128], ipflags[128],
      flags[128], saddress[82], daddress[88], dport[64], sport[64], issyn[20],
      seqsums[128];
  struct parse_args *args = (struct parse_args *)argstruct;
  struct tcppacket *pkt = (struct tcppacket*)args->buffer;
  int verbose = args->verblev;
  time_t t;
  char *timestring = (char*)NULL;

  if(!pkt)
    goto ptcp_end;

  if((timestring = (char*)malloc(32)) < 0) {
    fprintf(stderr, "%s[%i]: malloc: %s\n", progname, wlpid, strerror(errno));
    goto ptcp_end;
  }
  if(dots) {
    if(dots > 1) {
      t = time(NULL);
      snprintf(timestring, 32, " (TS: %lu)", t);
    } else {
      struct tm *lt;
      t = time(NULL);
      lt = localtime(&t);
      if((strftime(timestring, 32, " (TS: %b %d %Y %T)", lt)) < 0) {
	fprintf(stderr, "%s[%i]: strftime: %s", progname, wlpid, strerror(errno));
      }
    }
  }

  if(pkt->ip.ihl != 5)
    snprintf(suspip, sizeof(suspip), " Suspicious IP Options: ");
  else
    snprintf(suspip, sizeof(suspip), " ");

  if((pktsize = ntohs(pkt->ip.tot_len)) > 1024) {
    if(floatsize) {
      snprintf(psizebuf, sizeof(psizebuf), "%4.02f KBytes", (float)pktsize/1024.0);
    } else {
      snprintf(psizebuf, sizeof(psizebuf), "%i KBytes", pktsize/1024);
    }
  } else snprintf(psizebuf, sizeof(psizebuf), "%i Bytes", pktsize);

  if(pkt->ip.saddr) {
    struct in_addr s;
    s.s_addr = pkt->ip.saddr;
    if(skipdns) {
      snprintf(saddress, sizeof(saddress), "%s", inet_ntoa(s));
    } else {
      snprintf(saddress, sizeof(saddress), "%s[%s]",
	  hostlookup(pkt->ip.saddr), inet_ntoa(s));
    }
  } else {
    snprintf(saddress, sizeof(saddress), "no source found");
  }
  if(pkt->ip.daddr) {
    struct in_addr d;
    d.s_addr = pkt->ip.daddr;
    if(skipdns) {
      snprintf(daddress, sizeof(daddress), " to %s,", inet_ntoa(d));
    } else {
      snprintf(daddress, sizeof(daddress), " to %s[%s],",
	  hostlookup(pkt->ip.daddr), inet_ntoa(d));
    }
  } else {
    snprintf(daddress, sizeof(daddress), ",");
  }

  if(verbose >= 3) {
    snprintf(ipflags, sizeof(ipflags), "  IP_FLAGS:");
    if(pkt->ip.version > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s VER=%i", ipflags,
	  pkt->ip.version);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.protocol > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s PROTO=%i", ipflags,
	  pkt->ip.protocol);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.frag_off > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s FRAG_OFF=%i", ipflags,
	  pkt->ip.frag_off);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.tos > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s TOS=%i", ipflags,
	  pkt->ip.tos);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.id > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s ID=%i", ipflags,
	  pkt->ip.id);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.ttl > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s TTL=%i", ipflags,
	  pkt->ip.ttl);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.check > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s CHECK=%i", ipflags,
	  pkt->ip.check);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
  } else
    snprintf(ipflags, sizeof(ipflags), " ");

  snprintf(flags, sizeof(flags), "  TCP_Flags:");
#ifdef __FAVOR_BSD
  if(pkt->tcp.th_flags) {
    if(pkt->tcp.th_flags & TH_FIN) {
      snprintf(tempflags, sizeof(tempflags), "%s FIN", flags);
      snprintf(flags, sizeof(flags), "%s", tempflags);
    }
    if(pkt->tcp.th_flags & TH_SYN) {
      snprintf(tempflags, sizeof(tempflags), "%s SYN", flags);
      snprintf(flags, sizeof(flags), "%s", tempflags);
      synpkt = 1;
    }
    if(!(verbose > 0) && !(synpkt))
      goto ptcp_end;
    if(pkt->tcp.th_flags & TH_RST) {
      snprintf(tempflags, sizeof(tempflags), "%s RST", flags);
      snprintf(flags, sizeof(flags), "%s", tempflags);
    }
    if(pkt->tcp.th_flags & TH_PUSH) {
      snprintf(tempflags, sizeof(tempflags), "%s PUSH", flags);
      snprintf(flags, sizeof(flags), "%s", tempflags);
    }
    if(pkt->tcp.th_flags & TH_ACK) {
      snprintf(tempflags, sizeof(tempflags), "%s ACK", flags);
      snprintf(flags, sizeof(flags), "%s", tempflags);
    }
    if(pkt->tcp.th_flags & TH_URG) {
      snprintf(tempflags, sizeof(tempflags), "%s URG", flags);
      snprintf(flags, sizeof(flags), "%s", tempflags);
    }
  }
  if(pkt->tcp.th_off > 0) {
    snprintf(tempflags, sizeof(tempflags), "%s DOFF=%i", flags,
	pkt->tcp.th_off);
    snprintf(flags, sizeof(flags), "%s", tempflags);
  }
  if(verbose >= 2) {
    if(pkt->tcp.th_x2 > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s X2=%i", flags,
	  pkt->tcp.th_x2);
      snprintf(flags, sizeof(flags), "%s", tempflags);
    }
    if(pkt->tcp.th_win > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s WIN=%i", flags,
	  pkt->tcp.th_win);
      snprintf(flags, sizeof(flags), "%s", tempflags);
    }
    if(pkt->tcp.th_urp > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s URP=%i", flags,
	  pkt->tcp.th_urp);
      snprintf(flags, sizeof(flags), "%s", tempflags);
    }
  }

  if(skipport) {
    snprintf(dport, sizeof(dport), "%d", ntohs(pkt->tcp.th_dport));
    snprintf(sport, sizeof(sport), "%d", ntohs(pkt->tcp.th_sport));
  } else {
    snprintf(dport, sizeof(dport), "%d(%s)", ntohs(pkt->tcp.th_dport),
	servlookup(pkt->tcp.th_dport,6));
    snprintf(sport, sizeof(sport), "%d(%s)", ntohs(pkt->tcp.th_sport),
	servlookup(pkt->tcp.th_sport,6));
  }
  if(verbose >= 2)
    snprintf(seqsums, sizeof(seqsums), "  Checksum: %u  Seq: %lu  Ack Seq: %lu",
	pkt->tcp.th_sum, pkt->tcp.th_seq, pkt->tcp.th_ack);
  else
    snprintf(seqsums, sizeof(seqsums), " ");
#else /* ! __FAVOR_BSD */
  if(pkt->tcp.syn > 0) {
    snprintf(tempflags, sizeof(tempflags), "%s SYN", flags);
    snprintf(flags, sizeof(flags), "%s", tempflags);
    synpkt = 1;
  }
  if(!(verbose > 0) && !(synpkt))
    goto ptcp_end;
  if(pkt->tcp.ack > 0) {
    snprintf(tempflags, sizeof(tempflags), "%s ACK", flags);
    snprintf(flags, sizeof(flags), "%s", tempflags);
  }
  if(pkt->tcp.fin > 0) {
    snprintf(tempflags, sizeof(tempflags), "%s FIN", flags);
    snprintf(flags, sizeof(flags), "%s", tempflags);
  }
  if(pkt->tcp.rst > 0) {
    snprintf(tempflags, sizeof(tempflags), "%s RST", flags);
    snprintf(flags, sizeof(flags), "%s", tempflags);
  }
  if(pkt->tcp.psh > 0) {
    snprintf(tempflags, sizeof(tempflags), "%s PSH", flags);
    snprintf(flags, sizeof(flags), "%s", tempflags);
  }
  if(pkt->tcp.urg > 0) {
    snprintf(tempflags, sizeof(tempflags), "%s URG", flags);
    snprintf(flags, sizeof(flags), "%s", tempflags);
  }
  if(pkt->tcp.doff > 0) {
    snprintf(tempflags, sizeof(tempflags), "%s DOFF=%i", flags,
	pkt->tcp.doff);
    snprintf(flags, sizeof(flags), "%s", tempflags);
  }
  if(verbose >= 2) {
    if(pkt->tcp.res1 > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s RES1=%i", flags,
	  pkt->tcp.res1);
      snprintf(flags, sizeof(flags), "%s", tempflags);
    }
    if(pkt->tcp.res2 > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s RES2=%i", flags,
	  pkt->tcp.res2);
      snprintf(flags, sizeof(flags), "%s", tempflags);
    }
    if(pkt->tcp.window > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s WIN=%i", flags,
	  pkt->tcp.window);
      snprintf(flags, sizeof(flags), "%s", tempflags);
    }
    if(pkt->tcp.urg_ptr > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s URP=%i", flags,
	  pkt->tcp.urg_ptr);
      snprintf(flags, sizeof(flags), "%s", tempflags);
    }
  }

  if(skipport) {
    snprintf(dport, sizeof(dport), "%d", ntohs(pkt->tcp.dest));
    snprintf(sport, sizeof(sport), "%d", ntohs(pkt->tcp.source));
  } else {
    snprintf(dport, sizeof(dport), "%d(%s)", ntohs(pkt->tcp.dest),
	servlookup(pkt->tcp.dest,6));
    snprintf(sport, sizeof(sport), "%d(%s)", ntohs(pkt->tcp.source),
	servlookup(pkt->tcp.source,6));
  }
  if(verbose >= 2)
    snprintf(seqsums, sizeof(seqsums), "  Checksum: %u  Seq: %lu  Ack Seq: %lu",
	pkt->tcp.check, pkt->tcp.seq, pkt->tcp.ack_seq);
  else
    snprintf(seqsums, sizeof(seqsums), " ");
#endif /* __FAVOR_BSD */

  if(synpkt)
    snprintf(issyn, sizeof(issyn), "connection");
  else
    snprintf(issyn, sizeof(issyn), "packet");

  if(dots) {
    snprintf(newbuf, sizeof(newbuf), "TCP:%s%sPort %s %s from %s%s Size: %s  "
	"Source Port: %s%s%s%s", timestring, suspip, dport, issyn, saddress,
	daddress, psizebuf, sport, flags, seqsums, ipflags);
  } else {
    snprintf(newbuf, sizeof(newbuf), "TCP:%sPort %s %s from %s%s Size: %s  "
	"Source Port: %s%s%s%s", suspip, dport, issyn, saddress, daddress,
	psizebuf, sport, flags, seqsums, ipflags);
  }
  print_log("%s\n", newbuf);

ptcp_end:
  if(timestring) free(timestring);
  if(args->buffer) free(args->buffer);
  if(argstruct) free(argstruct);
#ifdef USETHREADS
  if(getpid() != wlpid) {
    if(debug >= 4) print_log("debug: Child thread == %u, exiting thread\n",
	getpid());
    deactivate_thread();
    pthread_exit(NULL);
  } else {
    if(debug >= 4)
      print_log("debug: Child (%i) and Parent (%i) have the same pid?!\n",
	  getpid(), wlpid);
  }
#endif

  return 0;
}
