/* sockets.h
 * Definitions and declarations for the code dealing with functions.
 */

#define SOCKTYPE_TCP 0x0001
#define SOCKTYPE_UDP 0x0002
#define SOCKTYPE_ICMP 0x0004
#define SOCKTYPE_IGMP 0x0008

#define SOCKTYPE_ALL (SOCKTYPE_TCP|SOCKTYPE_UDP|SOCKTYPE_ICMP|SOCKTYPE_IGMP)

#define TCPIsSet(x)	(x & SOCKTYPE_TCP)
#define UDPIsSet(x)	(x & SOCKTYPE_UDP)
#define ICMPIsSet(x)	(x & SOCKTYPE_ICMP)
#define IGMPIsSet(x)	(x & SOCKTYPE_IGMP)

#define SetTCP(x)	(x |= SOCKTYPE_TCP)
#define SetUDP(x)	(x |= SOCKTYPE_UDP)
#define SetICMP(x)	(x |= SOCKTYPE_ICMP)
#define SetIGMP(x)	(x |= SOCKTYPE_IGMP)

#define ClrTCP(x)	(x &= ~SOCKTYPE_TCP)
#define ClrUDP(x)	(x &= ~SOCKTYPE_UDP)
#define ClrICMP(x)	(x &= ~SOCKTYPE_ICMP)
#define ClrIGMP(x)	(x &= ~SOCKTYPE_IGMP)

#define GetProto(x,y)	if(ICMPIsSet(x)) y = 1; \
			else if(IGMPIsSet(x)) y = 2; \
			else if(TCPIsSet(x)) y = 6; \
			else if(UDPIsSet(x)) y = 17;

#define PACKET_BUF_SIZE 8192 /* 8 KiloBytes */

typedef struct wl_sockets {
  struct wl_sockets *next, *prev;
  int fd, type, verbose;
} wl_socket;

typedef struct wl_vl {
  int	global,	/* global verbose level (overruled by the below ones) */
	icmp,	/* icmp verbose level */
	igmp,	/* igmp verbose level */
	tcp,	/* tcp verbose level */
	udp;	/* udp verbose level */
} wl_verb;

wl_socket *socketlist;

extern void setup_sockets(int, wl_verb);
extern int addsocket(int, int, int);
extern int setsock_nonblock(int, int);
extern int check_for_packet();
