/* weedlog.c
 * Contains general stuff needed to run the program.
 */

#include "weedlog.h"
#include <pwd.h>

int socktypes=0, bequiet=0, daemon_mode=0, console_mode=0, gverb=0,
    tcpverb=0, icmpverb=0, udpverb=0, igmpverb=0, flmode=0;
#ifdef USETHREADS
int maxthreads = MAXTHREADS;
#endif

/* go_background();
 * Sends the program into the background (ie daemon mode).
 * Returns the new process id.
 */
int go_background() {
  int fs = -1;

  switch((fs = fork())) {
    case -1:
      return (-1);
    case 0:
      break;
    default:
      exit(0);
  }

  if(setsid() == -1)
    return(-1);

  (void)chdir("/");

  return fs;

}

int showhelp(char *name) {
  fprintf(stdout, "weedlogd version %s\n", VERSION);
  fprintf(stdout, "Option parts surrounded by [] are optional.\n");
  fprintf(stdout, "Option parts surrounded by <> are required.\n");
  fprintf(stdout, "Synopsis:\n");
  fprintf(stdout, "  %s [-F <file>] [-I] [-O[+]<l|s>] [-T] [-U] [-V<p> [#]] [-b] [-d [#]] [-f] [-h] [-n] [-p] [-t] [-v [#]] [-q]\n", name);
  fprintf(stdout, "    OR\n");
  fprintf(stdout, "  %s -[F][I][-O[+]<l|s>][T][U][V<p>][b][d][f][h][n][p][t][v][q] <file> [procotol verbose #] [debug #] [verbose #]\n", name);
  fprintf(stdout, "Command Line Options:\n");
  fprintf(stdout, " Protocols:\n");
  fprintf(stdout, "  -I\t\t: Watch for ICMP packets.\n");
  fprintf(stdout, "  -G\t\t: Watch for IGMP packets. (expirimental)\n");
  fprintf(stdout, "  -T\t\t: Watch for TCP packets.\n");
  fprintf(stdout, "  -U\t\t: Watch for UDP packets.\n");
  fprintf(stdout, " Output Options:\n");
  fprintf(stdout, "  -F <file>\t: Use <file> instead of syslog when in Background/Daemon mode.\n");
  fprintf(stdout, "  -b\t\t: Use Background/Daemon mode.\n");
  fprintf(stdout, "  -f\t\t: Use Foreground/Console mode.\n");
  fprintf(stdout, "  -O[+]<l|s>\t: Force output options, see the README file or man page.\n");
  fprintf(stdout, "  -n\t\t: Don't lookup domain names.\n");
  fprintf(stdout, "  -p\t\t: Don't lookup port names.\n");
  fprintf(stdout, "  -t\t\t: Use redundant timestamping.\n");
  fprintf(stdout, "\t\t- Use once for \'Month Day Year Time\' format.\n");
  fprintf(stdout, "\t\t- Use twice for raw timestamp format.\n");
  fprintf(stdout, " Verbose Options:\n");
  fprintf(stdout, "  -V<p> [#]\t: Set verbose level.\n");
  fprintf(stdout, "\t\t- \'<p>\' should be replaced by a letter signifying a protocol.\n");
  fprintf(stdout, "\t\t- Valid protocols are 'I', 'G', 'T' and 'U'.\n");
  fprintf(stdout, "  -d [#]\t: Use debug mode level #.\n");
  fprintf(stdout, "\t\t- Effective debug mode levels: 5\n");
  fprintf(stdout, "  -v [#]\t: Set verbose level.\n");
  fprintf(stdout, "\t\t- Effective verbose mode levels: 3\n");
  fprintf(stdout, "  READ THIS: Giving the options a numeric argument is optional. If you use the\n"
	"  compressed format (-dVTVIVUv...) and wish to give numeric arguments to some,\n"
	"  but not all, of these options then the ones you wish to give numeric arguments\n"
	"  to must come first or the others will steal their numbers.\n");
  fprintf(stdout, " Misc. Options:\n");
  fprintf(stdout, "  -h\t\t: This help message.\n");
#ifdef USETHREADS
  fprintf(stdout, "  -m <#>\t: Set the number of maximum simultanious threads. (default: 10)\n");
#endif
  fprintf(stdout, "  -q\t\t: Be quiet when starting up (good for starting at boot).\n");
  fprintf(stdout, "\nExamples:\n");
  fprintf(stdout, "  See the README file or manual page for examples.\n");
  fprintf(stdout, "\nNotes:\n");
  fprintf(stdout, "  - Please see the README file or man page for further explination of the\n"
      "output options.\n");
  exit(0);
  return 0;
}

/* parse_args(argc, argv);
 * Parses the command line arguments.
 */
int parse_args(int argc, char *argv[]) {
  register const char *s;
  int n = 1;

  while(argv[n]) {
    s = argv[n];
    if(*s == '-') {
      s++;
      while(*s) {
	if(*s == 'F') {
	  register const char *filepath;
	  char *mycwd, *fname;
	  n++;
	  filepath = argv[n];
	  if(!filepath) goto NEXT;
	  if((mycwd = (char*)malloc(PATH_MAX)) < 0) {
	    print_log("malloc: %s\n", strerror(errno));
	    goto NEXT;
	  }
	  if((fname = (char*)malloc(PATH_MAX)) < 0) {
	    print_log("malloc: %s\n", strerror(errno));
	    goto NEXT;
	  }
	  if(*filepath == '/') {
	    snprintf(fname, PATH_MAX, "%s", filepath);
	  } else {
	    getcwd(mycwd, PATH_MAX);
	    snprintf(fname, PATH_MAX, "%s/%s", mycwd, filepath);
	  }
	  if((outfile = open(fname, O_WRONLY|O_CREAT|O_APPEND, S_IRUSR|S_IWUSR|S_IRGRP)) < 0) {
	    print_log("error: could not open %s (%s)\n", fname, strerror(errno));
	  } else SetLFile(logmode);
	  if(fname)
	    free(fname);
	  if(mycwd)
	    free(mycwd);
	  goto NEXT;
	}
	if(*s == 'I') {
	  SetICMP(socktypes);
	  goto NEXT;
	}
	if(*s == 'G') {
	  SetIGMP(socktypes);
	  goto NEXT;
	}
	if(*s == 'O') {
	  s++;
	  if(*s == '+') {
	    s++;
/*
	    if(tolower(*s) == 'f') {
	      SetNFile(flmode);
	      goto NEXT;
	    }
*/
	    if(tolower(*s) == 'l') {
	      SetNSysL(flmode);
	      goto NEXT;
	    }
	    if(tolower(*s) == 's') {
	      SetNSOut(flmode);
	      goto NEXT;
	    }
	  }
/*
	  if(tolower(*s) == 'f') {
	    SetLFile(flmode);
	    goto NEXT;
	  }
*/
	  if(tolower(*s) == 'l') {
	    SetLSysL(flmode);
	    goto NEXT;
	  }
	  if(tolower(*s) == 's') {
	    SetLSOut(flmode);
	    goto NEXT;
	  }
	  goto NEXT;
	}
	if(*s == 'P') {
	  floatsize = 1;
	  goto NEXT;
	}
	if(*s == 'T') {
	  SetTCP(socktypes);
	  goto NEXT;
	}
	if(*s == 'U') {
	  SetUDP(socktypes);
	  goto NEXT;
	}
	if(*s == 'V') {
	  s++;
	  if(*s == 'I') {
	    if(argv[n+1]) {
	      if(checkisnum(argv[n+1])) {
		n++;
		icmpverb = atoi(argv[n]);
	      } else icmpverb++;
	    } else icmpverb++;
	  }
	  if(*s == 'G') {
	    if(argv[n+1]) {
	      if(checkisnum(argv[n+1])) {
		n++;
		igmpverb = atoi(argv[n]);
	      } else igmpverb++;
	    } else igmpverb++;
	  }
	  if(*s == 'T') {
	    if(argv[n+1]) {
	      if(checkisnum(argv[n+1])) {
		n++;
		tcpverb = atoi(argv[n]);
	      } else tcpverb++;
	    } else tcpverb++;
	  }
	  if(*s == 'U') {
	    if(argv[n+1]) {
	      if(checkisnum(argv[n+1])) {
		n++;
		udpverb = atoi(argv[n]);
	      } else udpverb++;
	    } else udpverb++;
	  }
	  goto NEXT;
	}
	if(*s == 'b') {
	  daemon_mode = 1;
	  SetLSysL(logmode);
	  goto NEXT;
	}
	if(*s == 'd') {
	  if(argv[n+1]) {
	    if(checkisnum(argv[n+1])) {
	      n++;
	      debug = atoi(argv[n]);
	    } else debug++;
	  } else debug++;
	  goto NEXT;
	}
	if(*s == 'f') {
	  console_mode = 1;
	  SetLSOut(logmode);
	  goto NEXT;
	}
	if(*s == 'h') {
	  showhelp(argv[0]);
	  goto NEXT;
	}
#ifdef USETHREADS
	if(*s == 'm') {
	  if(argv[n+1]) {
	    if(checkisnum(argv[n+1])) {
	      n++;
	      maxthreads = atoi(argv[n]);
	    } else maxthreads++;
	  } else maxthreads++;
	  goto NEXT;
	}
#endif
	if(*s == 'n') {
	  skipdns++;
	  goto NEXT;
	}
	if(*s == 'p') {
	  skipport++;
	  goto NEXT;
	}
	if(*s == 't') {
	  dots++;
	  goto NEXT;
	}
	if(*s == 'v') {
	  if(argv[n+1]) {
	    if(checkisnum(argv[n+1])) {
	      n++;
	      gverb = atoi(argv[n]);
	    } else gverb++;
	  } else gverb++;
	  goto NEXT;
	}
	if(*s == 'q') {
	  bequiet++;
	  goto NEXT;
	}
/*
	if(*s == '') {
	  goto NEXT;
	}
*/
NEXT:
	s++;
      }
    }
    n++;
  }

  if(daemon_mode == 1 && console_mode == 1)
    console_mode = 0;
  if(daemon_mode == 0 && console_mode == 0)
    daemon_mode = 1;
  return 0;
}

int main(int argc, char *argv[]) {
  int bg, i;
  uid_t uid = 0;
  char *uidname = (char*)NULL;
  time_t t;
  struct passwd *pwdinfo;
  wl_verb verblev;

  t = time(NULL);
  progname = fixprogname(argv[0]);
  wlpid = getpid();
  logmode = 0;
  skipdns = 0;
  skipport = 0;
  dots = 0;
  floatsize = 0;
#ifdef USETHREADS
  filelock = 0;
  stdlock = 0;
#endif

  parse_args(argc, argv);

  if(outfile) {
    if(!HasLFile(logmode)) SetLFile(logmode);
    if(!daemon_mode) SetLSOut(logmode);
    if(HasLSysL(logmode)) ClrLSysL(logmode);
  } else if(daemon_mode) {
    if(HasLFile(logmode)) ClrLFile(logmode);
    if(HasLSOut(logmode)) ClrLSOut(logmode);
    if(!HasLSysL(logmode)) SetLSysL(logmode);
  } else {
    if(HasLFile(logmode)) ClrLFile(logmode);
    if(HasLSysL(logmode)) ClrLSysL(logmode);
    if(!HasLSOut(logmode)) SetLSOut(logmode);
  }

  if(flmode) {
    if(HasLSysL(flmode) || HasNSysL(flmode)) {
      if(HasLSysL(flmode) && HasNSysL(flmode)) {
        if(daemon_mode)
	  if(!HasLSysL(logmode)) SetLSysL(logmode);
      } else if(HasLSysL(flmode)) {
        if(!HasLSysL(logmode)) SetLSysL(logmode);
      } else if(HasNSysL(flmode)) {
        if(HasLSysL(logmode)) ClrLSysL(logmode);
      }
    }
    if(HasLSOut(flmode) || HasNSOut(flmode)) {
      if(HasLSOut(flmode) && HasNSOut(flmode)) {
        if(console_mode)
	  if(!HasLSOut(logmode)) SetLSOut(logmode);
      } else if(HasLSOut(flmode)) {
        if(!HasLSOut(logmode)) SetLSOut(logmode);
      } else if(HasNSOut(flmode)) {
        if(HasLSOut(logmode)) ClrLSOut(logmode);
      }
    }
  }

  if(HasLSysL(logmode)) {
    char *syslogname;
    if((syslogname = (char*)malloc(PATH_MAX)) < 0) {
      print_log("malloc: %s\n", strerror(errno));
      return 1;
    }
    sprintf(syslogname, "%s[%i]", progname, wlpid);
    openlog(syslogname, 0, LOG_DAEMON);
    if(debug)
      print_log("Testing syslog status under name %s\n", syslogname);
  }

  uid = getuid();
  if((pwdinfo = getpwuid(uid))) {
    i = wl_psize(pwdinfo->pw_name);
    if((uidname = (char*)malloc(i)) < 0) {
      print_log("malloc: %s\n", strerror(errno));
      return 1;
    }
    memcpy(uidname, pwdinfo->pw_name, i);
  } else {
    if((uidname = (char*)malloc(16)) < 0) {
      print_log("malloc: %s\n", strerror(errno));
      return 1;
    }
    snprintf(uidname, 16, "unknown(%i)", uid);
  }

  if(!bequiet) {
    struct tm *lt;
    char *timestring;
    lt = localtime(&t);
    if((timestring = (char*)malloc(24)) < 0) {
      print_log("malloc: %s\n", strerror(errno));
      return 1;
    }
    if((strftime(timestring, 24, "%b %d %Y %T", lt)) < 0) {
      print_log("strftime: %s\n", strerror(errno));
      return 1;
    }
    print_log("init: %s started at %s by user %s (%i)\n",
	argv[0], timestring, uidname, uid);
    if(timestring) free(timestring);
  }
  if(uid != 0) {
#ifdef LOGNONROOT
    struct tm *lt;
    char *timestring, *fullargs = (char*)NULL;
    int i = 0, n = 0;
    lt = localtime(&t);
    if((timestring = (char*)malloc(24)) < 0) {
      print_log("malloc: %s\n", strerror(errno));
      return 1;
    }
    if((strftime(timestring, 24, "%b %d %Y %T", lt)) < 0) {
      print_log("strftime: %s\n", strerror(errno));
      return 1;
    }
    while(n < argc) {
      i += wl_psize(argv[n]);
      n++;
    }
    i += argc;
    if(i > 0) {
      char *argstmp, *tempstr;
      if((argstmp = (char*)malloc(i)) < 0) {
	print_log("malloc: %s\n", strerror(errno));
	return 1;
      }
      memset(argstmp, '\0', i);
      if((tempstr = (char*)malloc(i)) < 0) {
	print_log("malloc: %s\n", strerror(errno));
	return 1;
      }
      memset(tempstr, '\0', i);
      n = 0;
      while(n < argc) {
	if(n == 0) snprintf(argstmp, i, "%s", argv[n]);
	else {
	  snprintf(tempstr, i, "%s", argstmp);
	  snprintf(argstmp, i, "%s %s", tempstr, argv[n]);
	}
	n++;
      }
      if((fullargs = (char*)malloc(i)) < 0) {
	print_log("malloc: %s\n", strerror(errno));
	return 1;
      }
      memcpy(fullargs, argstmp, i);
      if(argstmp) free(argstmp);
      if(tempstr) free(tempstr);
    }
    syslog(LOG_NOTICE, "Non-root user %s (%i) started '%s' at %s",
	uidname, uid, fullargs, timestring);
    if(timestring) free(timestring);
    if(fullargs) free(fullargs);
#endif
    if(setuid(0) < 0) {
      print_log("init: Unable to set root privilidges for raw socket "
	  "operations. (%s)\n", strerror(errno));
      return 1;
    }
  }

  verblev.global = gverb;
  verblev.icmp = icmpverb;
  verblev.igmp = igmpverb;
  verblev.tcp = tcpverb;
  verblev.udp = udpverb;

  if(debug >= 2) {
    fprintf(stdout, "debug level: %i\nverbose levels:\n  global: %i\n  icmp:"
	" %i\n  tcp: %i\n  udp: %i\n", debug, verblev.global, verblev.icmp,
	verblev.tcp, verblev.udp);
  }

  if(daemon_mode) {
    if(!bequiet)
      printb_log(FALSE, "init: Entering background mode.\n");
    if((bg = go_background()) < 0) {
      print_log_err("error: fork() failed (%s), exiting.\n", strerror(errno));
      return 1;
    } else {
      wlpid = getpid();
      if(!bequiet)
	print_log("init: Background mode entered. PID changed to %i.\n", wlpid);
      if(HasLSysL(logmode)) {
	char *syslogname;
	if((syslogname = (char*)malloc(PATH_MAX)) < 0) {
	  print_log("malloc: %s\n", strerror(errno));
	  return 1;
	}
	sprintf(syslogname, "%s[%i]", progname, wlpid);
	closelog();
	openlog(syslogname, 0, LOG_DAEMON);
	SetLSysL(logmode);
	if(debug)
	  print_log("Testing syslog status under name %s\n", syslogname);
      }
    }
  } else wlpid = getpid();

#ifdef USETHREADS
  if(debug >= 4) print_log("Parent thread == %u\n", wlpid);
#endif

  setup_sockets(socktypes, verblev);
  if(!socketlist) {
    printb_log(FALSE, "error: No sockets to watch, exiting.\n");
    return 1;
  }

  for(;;) {
    if(debug >= 5)
      printb_log(FALSE, "debug: Checking for packets\n");
    check_for_packet();
    if(debug >= 5)
      printb_log(FALSE, "debug: Sleeping.\n");
    usleep(LOOPDELAY);
#ifdef USETHREADS
    while(activethreads >= maxthreads) usleep(1);
#endif
  }

  return 0;
}
